#define DRAWPRIM
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/ToggleB.h>
#include <stdlib.h>
#include <math.h>
#include "paint.h"

static int zoomXF, zoomYF, zoX, zoY;
static Drawable drawable;
static int drawZoom = FALSE;

static void sprayMotion(Widget w, XtPointer unused, XEvent * event);

/* structure for data about repeat type, etc. */
static struct
{
 int num, *repX, *repY, *distX, *distY, *offset, fract;
 Widget DAW, sizeXW, sizeYW, convertRPW, gridTBW;
 Boolean convertRP;
 unsigned char active, mode;
 XRectangle r[MAX_DIFF_REPEATS];
 repeatType rType[MAX_DIFF_REPEATS];
} repeatArea;

void setZoomParameters(int doZoom)
{
 if (doZoom)
  {
   zoomXF = zoomDraw.factorX;
   zoomYF = zoomDraw.factorY;
   zoX = zoomDraw.clipWin.x;
   zoY = zoomDraw.clipWin.y;
   drawZoom = TRUE;
  }
 else
  {
   zoomXF = zoomYF = 1;
   zoX = zoY = 0;
   drawZoom = FALSE;
  }
}

DDRAWPRIM void drawCross(Drawable d, GC gc, int x, int y, unsigned short l)
{
 XGCValues values;

 if (ApData.drawinfo.lineWidth != 1)
  {
   values.line_width = 0;
   XChangeGC(xinfo.dpy, gc, GCLineWidth, &values);
  }
 XDrawLine (xinfo.dpy, d, gc, x - l, y, x + l, y);
 XDrawLine (xinfo.dpy, d, gc, x, y - l, x, y + l);
 if (ApData.drawinfo.lineWidth != 1)
  {
   values.line_width = ApData.drawinfo.lineWidth;
   XChangeGC(xinfo.dpy, gc, GCLineWidth, &values);
  }
}

/* modify line width Callback */
DDRAWPRIM void line_widthCB(Widget w, XtPointer clientData, XtPointer callData)
{
 XGCValues values;

 XmScaleGetValue(w, &ApData.drawinfo.lineWidth);
 values.line_width = ApData.drawinfo.lineWidth;
 if (values.line_width == 1)
   values.line_width--;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth, &values);
}

DDRAWPRIM void drawLines(Drawable drawable, GC gc, Boolean fillRegion,
 XPoint *p, int pointsNo, int offsX, int offsY, Boolean contiguous)
{
 int i;

 movePts (p, pointsNo, -offsX, -offsY);
 if (contiguous)
  XDrawLines (xinfo.dpy, drawable, gc, p, pointsNo, CoordModeOrigin);
 else
  for (i = 1; i < pointsNo; i++)
   XDrawLine (xinfo.dpy, drawable, gc, p[i - 1].x, p[i - 1].y, p[i].x, p[i].y);
 if (fillRegion)
   XFillPolygon(xinfo.dpy, drawable, gc, p, pointsNo, Complex, CoordModeOrigin);
 movePts(p, pointsNo, offsX, offsY);
}

/* flip points pointed by pointer xp according to width & height in orR */
DDRAWPRIM void flipPts(XPoint *xp, int pointNo, repeatType rpt, XPoint *orR)
{
 int i;
 XPoint flipArea;

 memcpy (&flipArea, orR, sizeof (XPoint));
 flipArea.x--;
 flipArea.y--;
 switch (rpt)
  {
   case normal:
     break;
   case mirrorX:
     for (i = 0; i < pointNo; i++)
       xp[i].y = flipArea.y - xp[i].y;
     break;
   case mirrorY:
     for (i = 0; i < pointNo; i++)
       xp[i].x = flipArea.x - xp[i].x;
     break;
   case mirrorXY:
     for (i = 0; i < pointNo; i++)
       {
	 xp[i].x = flipArea.x - xp[i].x;
	 xp[i].y = flipArea.y - xp[i].y;
       }
     break;
  }
}

/* flip rectangle pointed by pointer xp according to width & height in orR */
/* first point is left top corner of rectangle, second are width and height */
static void flipRect(XPoint * xp, repeatType rpt, XPoint * orR, Boolean fill, int lineWidth)
{
 short fillDiff; /* correction by 1 due to the difference in how X server */
 /* draws filled and non filled rectangles */

 if ((!fill) && ((lineWidth % 2) == 1))
   fillDiff = 1;
 else
   fillDiff = 0;
 switch (rpt)
   {
   case normal:
     break;
   case mirrorX:
     xp[0].y = orR->y - xp[0].y - xp[1].y - fillDiff;
     break;
   case mirrorY:
     xp[0].x = orR->x - xp[0].x - xp[1].x - fillDiff;
     break;
   case mirrorXY:
     xp[0].x = orR->x - xp[0].x - xp[1].x - fillDiff;
     xp[0].y = orR->y - xp[0].y - xp[1].y - fillDiff;
     break;
   }
}

DDRAWPRIM void DrawLinesWrapped (Drawable drawable, GC gc, Boolean fillRegion,
 XPoint * p, int pointsNo, int offsX, int offsY, Boolean contiguous)
{
 int i;
 XPoint orR, newOr;

 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;

 for (i = 0; i < repeats.currNum; i++)
  if (repeats.draw[i])
   {
    newOr.x = repeats.r[i].x;
    newOr.y = repeats.r[i].y;
    movePts (p, pointsNo, -newOr.x, -newOr.y);
    flipPts (p, pointsNo, repeats.r[i].rType, &orR);
    drawLines (drawable, gc, fillRegion, p, pointsNo, offsX, offsY, contiguous);
    flipPts (p, pointsNo, repeats.r[i].rType, &orR);
    movePts (p, pointsNo, newOr.x, newOr.y);
   }
}

DDRAWPRIM void DrawRectangle(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY)
{
  if (DI.fill)
   XFillRectangle(xinfo.dpy, drawable, gc,
    p[0].x - offsX, p[0].y - offsY, p[1].x, p[1].y);
  else
   XDrawRectangle(xinfo.dpy, drawable, gc,
    p[0].x - offsX, p[0].y - offsY, p[1].x, p[1].y);
}

DDRAWPRIM void DrawRectangleWrapped (Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY)
{
 int i;
 XPoint orR, newOr;

 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;

 for (i = 0; i < repeats.currNum; i++)
  if (repeats.draw[i])
   {
    newOr.x = repeats.r[i].x;
    newOr.y = repeats.r[i].y;
    movePts (p, 1, -newOr.x, -newOr.y);
    flipRect (p, repeats.r[i].rType, &orR, DI.fill, DI.lineWidth);
    DrawRectangle (drawable, gc, DI, p, offsX, offsY);
    flipRect (p, repeats.r[i].rType, &orR, DI.fill, DI.lineWidth);
    movePts (p, 1, newOr.x, newOr.y);
   }
}

DDRAWPRIM void drawArc(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY)
{
 if (DI.fill)
  XFillArc(xinfo.dpy, drawable, gc,
	  p[0].x - offsX, p[0].y - offsY, p[1].x, p[1].y, p[2].x, p[2].y);
 else
  XDrawArc(xinfo.dpy, drawable, gc,
	  p[0].x - offsX, p[0].y - offsY, p[1].x, p[1].y, p[2].x, p[2].y);
}

DDRAWPRIM void DrawArcWrapped(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY)
{
 int i;
 XPoint c[3], orR, newOr;

 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;

 for (i = 0; i < repeats.currNum; i++)
  if (repeats.draw[i])
   {
    newOr.x = repeats.r[i].x;
    newOr.y = repeats.r[i].y;
    memcpy (&c[0], &p[0], sizeof (XPoint));
    memcpy (&c[2], &p[2], sizeof (XPoint));
    c[1].x = c[0].x + p[1].x;
    c[1].y = c[0].y + p[1].y;
    movePts (&c[0], 2, -newOr.x, -newOr.y);
    flipPts (&c[0], 2, repeats.r[i].rType, &orR);
    findMinMax (&c[0], 2, &c[0].x, &c[0].y, &c[1].x, &c[1].y);
    c[1].x -= c[0].x;
    c[1].y -= c[0].y;
    drawArc (drawable, gc, DI, &c[0], offsX, offsY);
   }
}

DDRAWPRIM void DrawElArcWrapped(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY)
{
 int i;
 XPoint c[3], orR, newOr;

 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;

 for (i = 0; i < repeats.currNum; i++)
  if (repeats.draw[i])
   {
    newOr.x = repeats.r[i].x;
    newOr.y = repeats.r[i].y;
    memcpy (&c[0], &p[0], sizeof (XPoint));
    memcpy (&c[2], &p[2], sizeof (XPoint));
    c[1].x = c[0].x + p[1].x;
    c[1].y = c[0].y + p[1].y;
    movePts (&c[0], 2, -newOr.x, -newOr.y);
    flipPts (&c[0], 2, repeats.r[i].rType, &orR);
    findMinMax (&c[0], 2, &c[0].x, &c[0].y, &c[1].x, &c[1].y);
    c[1].x -= c[0].x;
    c[1].y -= c[0].y;
    drawArc (drawable, gc, DI, &c[0], offsX, offsY);
   }
}

/* From: Parametrical Elliptical Arc Algorithm,
   Graphic Gems III pp 164 in 478 J. Van Aken in R. Simar
   Plot a series of points along a PI/2-radian arc of an ellipse.
   The arc is specified in terms of a control polygon (a triangle)
   with vertices P, Q and K.  The arc begins at P, ends at Q, and is
   completely contained within the control polygon.  The draw_point
   function plots a single pixel at display coordinates (x,y).
Entry:
  xP, yP, xQ, yQ, xK, yK -- coordinates of P, Q and K.  These
    are 32-bit fixed-point values with 16 bits of fraction.
  m -- nonnegative integer that controls spacing between points.
    The angular increment between points is 1/2^m radians.
Exit:
  The number of points plotted is 1 + floor((PI/2)*2^m). */

#define PIV2  102944		/* fixed point PI/2 */
#define TWOPI 411775		/* fixed point 2*PI */
#define HALF  32768		/* fixed point 1/2 */
typedef long FIX;		/* 32-bit fixed point, 16-bit fraction */

static int qtr_elips (long xP, long yP, long xQ, long yQ, long xK, long yK, long m, XPoint * pts);

/* draw parametric elliptical arc from points P, Q, K */
DDRAWPRIM int DrawElArc(int Px, int Py, int Qx, int Qy, int Kx, int Ky, XPoint * el_pts)
{
 long xP, yP, xQ, yQ, xK, yK, m;
 int xPQ, yPQ, n_pts, diff;
 unsigned int dist;

 /* popravi tocko K tako, da bo 2* dlje od daljice PQ tako da bo blizje vrhu elipsoide */
 xP = Px << 16;
 yP = Py << 16;
 xQ = Qx << 16;
 yQ = Qy << 16;
 xK = Kx;
 yK = Ky;
 xPQ = Px + (int) ((int) (Qx - Px) / 2);
 yPQ = Py + (int) ((int) (Qy - Py) / 2);

 xK += (int) (Kx - xPQ);
 yK += (int) (Ky - yPQ);

 diff = abs (Px - Kx);
 dist = diff * diff;
 diff = abs (Py - Ky);
 dist += diff * diff;

 diff = abs (Qx - Kx);
 dist += diff * diff;
 diff = abs (Qy - Ky);
 dist += diff * diff;

 n_pts = sqrt (dist) / 4;
 m = log10 (n_pts / (M_PI_2)) / log10 (2.0);
 if (m < 2)
   m = 2;
 else if (m > 12)
   m = 12; /* maximum */

 n_pts = qtr_elips (xP, yP, xQ, yQ, xK << 16, yK << 16, m, el_pts);
 el_pts[n_pts].x = (short) Qx;
 el_pts[n_pts++].y = (short) Qy;
 return(n_pts);
}

static int qtr_elips(long xP, long yP, long xQ, long yQ, long xK, long yK, long m, XPoint * pts)
{
 int i, x, y, n_pts = 0;
 FIX vx, ux, vy, uy, w, xJ, yJ;

 vx = xK - xQ;			/* displacements from center */
 ux = xK - xP;
 vy = yK - yQ;
 uy = yK - yP;
 xJ = xP - vx + HALF;		/* center of ellipse J */
 yJ = yP - vy + HALF;
 ux -= (w = ux >> (2 * m + 3));	/* cancel 2nd-order error */
 ux -= (w >>= (2 * m + 4));	/* cancel 4th-order error */
 ux -= w >> (2 * m + 3);	/* cancel 6th-order error */
 ux += vx >> (m + 1);		/* cancel 1st-order error */
 uy -= (w = uy >> (2 * m + 3));	/* cancel 2nd-order error */
 uy -= (w >>= (2 * m + 4));	/* cancel 4th-order error */
 uy -= w >> (2 * m + 3);	/* cancel 6th-order error */
 uy += vy >> (m + 1);		/* cancel 1st-order error */
 for (i = (PIV2 << m) >> 16; i >= 0; --i)
  {
   x = (xJ + vx) >> 16;
   y = (yJ + vy) >> 16;

   pts[n_pts].x = (unsigned short) x;
   pts[n_pts++].y = (unsigned short) y;

   ux -= vx >> m;
   vx += ux >> m;
   uy -= vy >> m;
   vy += uy >> m;
  }
 return (n_pts);
}

/* polygon drawing event handler */
DDRAWPRIM void polygonDraw(Widget w, XtPointer dataP, XEvent *event)
{
 static Boolean firstClick = True;
 static Window startWin = 0;

 switch (event->type)
   {
   case ButtonPress:
     switch (event->xbutton.button)
       {
       case Button1:
	if (firstClick)
	 {
	  XGCValues values;

	  finalDraw(ApData.pixy);
	  Select.rect_on = False;
	  disableUndo();
	  Undo.tool = ApData.tool = polyDraw;
	  memcpy (&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));
	  XCopyGC (xinfo.dpy, xinfo.gc,
		   GCFunction | GCLineStyle | GCLineWidth | GCForeground | GCBackground | GCJoinStyle,
		   Undo.gc);
          setZoomParameters(((int) dataP) !=0);
          drawable = (drawZoom) ? ApData.scrPix : XtWindow(w); /* ABC double buffer problem */
	  values.function = GXinvert;
	  XChangeGC (xinfo.dpy, Undo.gc, GCFunction, &values);
	  Undo.points[Undo.npoints].x = Undo.points[Undo.npoints + 1].x = (event->xbutton.x + zoX) / zoomXF;
	  Undo.points[Undo.npoints].y = Undo.points[Undo.npoints + 1].y = (event->xbutton.y + zoY) / zoomYF;
	  Undo.npoints++;
	  XDrawLine(xinfo.dpy, drawable, Undo.gc,
	     Undo.points[Undo.npoints - 1].x - zoX, Undo.points[Undo.npoints - 1].y - zoY,
	     Undo.points[Undo.npoints].x - zoX, Undo.points[Undo.npoints].y - zoY);
	  findMinMax (&Undo.points[Undo.npoints - 1], 2, &opBounds.x, &opBounds.y,
		      &opBounds.width, &opBounds.height);
	  opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	  opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	  opBounds.x -= ApData.drawinfo.lineWidth;
	  opBounds.y -= ApData.drawinfo.lineWidth;
	  refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	  startWin = event->xbutton.window;
	 }
	break;
       }
     break;
   case ButtonRelease:
     if (event->xbutton.window != startWin)
      break;
     if (firstClick)
      {
       if (Undo.npoints != 0)
	 firstClick = False;
      }
     else
       {
	 switch (event->xbutton.button)
	   {
	   case Button1:
	     firstClick = True;
	     XDrawLine (xinfo.dpy, drawable, Undo.gc,
	      Undo.points[Undo.npoints - 1].x - zoX / zoomXF,
	      Undo.points[Undo.npoints - 1].y - zoY / zoomYF,
	      Undo.points[Undo.npoints].x - zoX / zoomXF,
	      Undo.points[Undo.npoints].y - zoY / zoomYF);
	     XSetFunction (xinfo.dpy, Undo.gc, GXcopy);
	     Undo.points[Undo.npoints].x = Undo.points[Undo.npoints + 1].x = (event->xbutton.x + zoX) / zoomXF;
	     Undo.points[Undo.npoints].y = Undo.points[Undo.npoints + 1].y = (event->xbutton.y + zoY) / zoomYF;

	     XDrawLine (xinfo.dpy, drawable, Undo.gc,
	      Undo.points[Undo.npoints].x - zoX / zoomXF,
	      Undo.points[Undo.npoints].y - zoY / zoomYF,
	      Undo.points[Undo.npoints - 1].x - zoX / zoomXF,
	      Undo.points[Undo.npoints - 1].y - zoY / zoomYF);
	     if (drawZoom)
	      {
	       findMinMax(&Undo.points[Undo.npoints - 1], 2,
		&opBounds.x, &opBounds.y,
		&opBounds.width, &opBounds.height);
	       opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	       opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	       opBounds.x -= ApData.drawinfo.lineWidth;
	       opBounds.y -= ApData.drawinfo.lineWidth;
	       refreshZoomWin(opBounds.x, opBounds.y, opBounds.width,
		opBounds.height, False);
	      }
	     Undo.npoints++;

	     if (Undo.npoints == Undo.allocPoints)
	       {
		 LastDraw (ApData.pixy);
		 memcpy (&Undo.points[0], &Undo.points[Undo.npoints - 2],
			 2 * sizeof (XPoint));
		 Undo.npoints = 1;
	       }
	     XSetFunction (xinfo.dpy, Undo.gc, GXinvert);
	     firstClick = False;
	     break;
	   case Button3:
	     firstClick = True;
	     XDrawLine (xinfo.dpy, drawable, Undo.gc,
	      Undo.points[Undo.npoints - 1].x - zoX / zoomXF,
	      Undo.points[Undo.npoints - 1].y - zoY / zoomYF,
	      Undo.points[Undo.npoints].x - zoX / zoomXF,
	      Undo.points[Undo.npoints].y - zoY / zoomYF);
	     XSetFunction (xinfo.dpy, Undo.gc, GXcopy);

	     startWin = 0;

	     enableUndo (False);
	     ApData.pic_change = True;
	     if (Undo.npoints > 1)
	       {
		findMinMax(Undo.points, Undo.npoints, &opBounds.x, &opBounds.y,
			    &opBounds.width, &opBounds.height);
		opBounds.width -= opBounds.x;
		opBounds.height -= opBounds.y;
		opBounds.x -= ApData.drawinfo.lineWidth;
		opBounds.y -= ApData.drawinfo.lineWidth;
		opBounds.width += ApData.drawinfo.lineWidth * 2;
		opBounds.height += ApData.drawinfo.lineWidth * 2;

		segmentDrawOp();
		calcOpBounds();

		Undo.area.x = opBounds.x;
		Undo.area.y = opBounds.y;
		Undo.area.width = opBounds.width;
		Undo.area.height = opBounds.height;
		Undo.img = SubImage(ApData.img, opBounds.x, opBounds.y, opBounds.width, opBounds.height);
		Undo.imgAllocated = True;
		if (Undo.img == NULL)
		  disableUndo ();
		LastDrawOnAllWin(True);
	       }
	     else if (drawZoom)
	       refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	     break;
	   case Button2:
	     if (Undo.npoints > 1)
	       {
		 Boolean fillIt;

		 /* to prevent filling of the polygon before confirmation
		    of the last point */
		 fillIt = Undo.drawinfo.fill;
		 Undo.drawinfo.fill = False;
		 firstClick = True;
		 Undo.npoints--;
		 memcpy (&Undo.points[Undo.npoints], &Undo.points[Undo.npoints - 1],
			 sizeof (XPoint));
		 XSetFunction (xinfo.dpy, Undo.gc, GXcopy);

		 refreshDrawingArea(True, True);
		 drawLines(drawAreaWin, Undo.gc, False, &Undo.points[0], Undo.npoints, 0, 0, False);
		 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
		  zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
		 drawLines(ApData.scrPix, Undo.gc, False, &Undo.points[0],
	          Undo.npoints, zoomDraw.clipWin11.x,
		  zoomDraw.clipWin11.y, False); /* ABC double buffer problem */
		 refreshZoomWin(zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
		  zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, False);

		 XSetFunction (xinfo.dpy, Undo.gc, GXinvert);
		 Undo.points[Undo.npoints].x = (event->xbutton.x + zoX) / zoomXF;
		 Undo.points[Undo.npoints].y = (event->xbutton.y + zoY) / zoomYF;
		 XDrawLine (xinfo.dpy, drawable, Undo.gc,
			  Undo.points[Undo.npoints - 1].x - zoX / zoomXF,
			  Undo.points[Undo.npoints - 1].y - zoY / zoomYF,
			    Undo.points[Undo.npoints].x - zoX / zoomXF,
			    Undo.points[Undo.npoints].y - zoY / zoomYF);
		 if (drawZoom)
		   {
		     findMinMax (&Undo.points[Undo.npoints - 1], 2, &opBounds.x, &opBounds.y,
				 &opBounds.width, &opBounds.height);
		     opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
		     opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
		     opBounds.x -= ApData.drawinfo.lineWidth;
		     opBounds.y -= ApData.drawinfo.lineWidth;
		     refreshZoomWin (opBounds.x, opBounds.y, opBounds.width,
				     opBounds.height, False);
		   }
		 firstClick = False;
		 Undo.drawinfo.fill = fillIt;
	       }
	     break;
	   }
       }
     break;
   case MotionNotify:
     if ((!firstClick) && (event->xbutton.window == startWin))
      {
       XDrawLine (xinfo.dpy, drawable, Undo.gc,
	Undo.points[Undo.npoints - 1].x - zoX / zoomXF,
	Undo.points[Undo.npoints - 1].y - zoY / zoomYF,
	Undo.points[Undo.npoints].x - zoX / zoomXF,
	Undo.points[Undo.npoints].y - zoY / zoomYF);
       memcpy(&Undo.points[Undo.npoints + 1], &Undo.points[Undo.npoints], sizeof (XPoint));
       Undo.points[Undo.npoints].x = (event->xmotion.x + zoX) / zoomXF;
       Undo.points[Undo.npoints].y = (event->xmotion.y + zoY) / zoomYF;
       XDrawLine (xinfo.dpy, drawable, Undo.gc,
	Undo.points[Undo.npoints - 1].x - zoX / zoomXF,
	Undo.points[Undo.npoints - 1].y - zoY / zoomYF,
	Undo.points[Undo.npoints].x - zoX / zoomXF,
	Undo.points[Undo.npoints].y - zoY / zoomYF);
       if (drawZoom)
	{
	 findMinMax(&Undo.points[Undo.npoints - 1], 3, &opBounds.x, &opBounds.y,
	  &opBounds.width, &opBounds.height);
	 opBounds.width -= opBounds.x - ApData.drawinfo.lineWidth * 2;
	 opBounds.height -= opBounds.y - ApData.drawinfo.lineWidth * 2;
	 opBounds.x -= ApData.drawinfo.lineWidth;
	 opBounds.y -= ApData.drawinfo.lineWidth;
	 refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, False);
	}
      }
     break;
   }
}

DDRAWPRIM void sprayWrapped(Drawable drawable, GC gc, XPoint *p, int pointsNo, int offsX, int offsY)
{
 int i;
 XPoint orR, newOr;

 if (sprayData.mixMode)
  {
   unsigned char *data = ApData.img->data, col1, col2;
   int imgX = ApData.width, imgY = ApData.height, imgBPL = ApData.img->bytes_per_line;
   XPoint *xp;

   movePts(p, pointsNo, 0, 0);
   xp = Undo.points;
   for (i = 0; i < pointsNo; i++) /* to avoid negative molulo */
    {
     while (xp[i].x<0)
      xp[i].x+=imgX;
     while (xp[i].y<0)
      xp[i].y+=imgY;
    }
   for (i = 1; i < pointsNo; i++) /* all points must be within picture */
    { 
     col1 = data[(xp[i-1].y%imgY)*imgBPL + (xp[i-1].x%imgX)];
     col2 = data[(xp[i].y%imgY)*imgBPL + (xp[i].x%imgX)];
     if (col1!=col2)
      {
       data[(xp[i-1].y%imgY)*imgBPL + (xp[i-1].x%imgX)]= ApData.transfTbl[col2][col1];
       data[(xp[i].y%imgY)*imgBPL + (xp[i].x%imgX)]=ApData.transfTbl[col1][col2];
      }
    }
   for (i = 0; i < pointsNo; i++) /* and back to original */
    { xp[i].x+=imgX; xp[i].y+=imgY; }
  }
 else
  {
   orR.x = repeats.crop.width;
   orR.y = repeats.crop.height;
   movePts(p, pointsNo, -offsX, -offsY);
   for (i = 0; i < repeats.currNum; i++)
    if (repeats.draw[i])
     {
      newOr.x = repeats.r[i].x;
      newOr.y = repeats.r[i].y;
      movePts(p, pointsNo, -newOr.x, -newOr.y);
      flipPts(p, pointsNo, repeats.r[i].rType, &orR);

      XDrawPoints(xinfo.dpy, drawable, gc, &Undo.points[0], pointsNo, CoordModeOrigin);
      flipPts(p, pointsNo, repeats.r[i].rType, &orR);
      movePts(p, pointsNo, newOr.x, newOr.y);
     }
   movePts(p, pointsNo, offsX, offsY);
  }
}

static void sprayArea(XtPointer clientData, XtIntervalId *id)
{
 int i, deg, r;
 XPoint *xp;

 if (sprayData.delay == 0)
  return;
 if ((Undo.npoints + sprayData.sprayPts) >= Undo.allocPoints)
  {
   sprayData.callAgain = False;
   sprayMotion(sprayData.w, (XtPointer) 5, NULL);
   return;
  }

 XSetForeground(xinfo.dpy, xinfo.gc,
  foregroundDraw ? ApData.drawinfo.foregrndC : ApData.drawinfo.backgrndC);

 xp = &Undo.points[Undo.npoints];
 for (i = 0; i < sprayData.sprayPts; i++, xp++)
  {
   deg = rand () / (RAND_MAX / 360);
   r = rand () / (RAND_MAX / sprayData.radius);
   xp->x = cos (deg * M_PI / 180) * r + sprayData.center.x;
   xp->y = sin (deg * M_PI / 180) * r + sprayData.center.y;
  }
 Undo.npoints += sprayData.sprayPts;

 if (sprayData.mixMode)
  {
    {
     unsigned char *data = ApData.img->data, col1, col2;
     int imgX = ApData.width, imgY = ApData.height, imgBPL = ApData.img->bytes_per_line;

     xp = &Undo.points[Undo.npoints - sprayData.sprayPts];
     for (i = 1; i < sprayData.sprayPts; i++) /* all points must be within picture */
       if ( (xp[i-1].x>=0)&&(xp[i-1].y>=0)&&(xp[i-1].x<imgX)&&(xp[i-1].y<imgY) &&
            (xp[i].x>=0)&&(xp[i].y>=0)&&(xp[i].x<imgX)&&(xp[i].y<imgY) )
      { 
       col1 =  data[xp[i-1].y*imgBPL + xp[i-1].x];
       col2 =  data[xp[i].y*imgBPL + xp[i].x];
       if (col1!=col2)
        {
	 XSetForeground(xinfo.dpy, xinfo.gc, xinfo.pixelTC[xinfo.allocCells[col2].pixel]);
         if ((zoomXF == 1) && (zoomYF == 1))
	  XDrawPoint(xinfo.dpy, XtWindow(sprayData.w), xinfo.gc, xp[i-1].x, xp[i-1].y);
	 else
	  XFillRectangle(xinfo.dpy, XtWindow(sprayData.w), xinfo.gc,
	   xp[i-1].x * zoomXF - zoX, xp[i-1].y * zoomYF - zoY, zoomXF, zoomYF);
	 XSetForeground(xinfo.dpy, xinfo.gc, xinfo.pixelTC[xinfo.allocCells[col1].pixel]);
         if ((zoomXF == 1) && (zoomYF == 1))
	  XDrawPoint(xinfo.dpy, XtWindow(sprayData.w), xinfo.gc, xp[i].x, xp[i].y);
	 else
	  XFillRectangle(xinfo.dpy, XtWindow(sprayData.w), xinfo.gc,
	   xp[i].x * zoomXF - zoX, xp[i].y * zoomYF - zoY, zoomXF, zoomYF);
	}
      }
    
    }   
  }
 else
  {
   if ((zoomXF == 1) && (zoomYF == 1))
    XDrawPoints(xinfo.dpy, XtWindow (sprayData.w), xinfo.gc,
     &Undo.points[Undo.npoints - sprayData.sprayPts], sprayData.sprayPts, CoordModeOrigin);
   else
    {
     xp = &Undo.points[Undo.npoints - sprayData.sprayPts];
     for (i = 0; i < sprayData.sprayPts; i++, xp++)
       XFillRectangle(xinfo.dpy, XtWindow(sprayData.w), xinfo.gc,
	xp->x * zoomXF - zoX, xp->y * zoomYF - zoY, zoomXF, zoomYF);
    }
  }
 if (sprayData.callAgain)
  XtAddTimeOut(sprayData.delay, (XtTimerCallbackProc) *sprayArea, NULL);
 else
  sprayData.callAgain = True;
}

/* event handler for spray tool */
static void sprayMotion(Widget w, XtPointer unused, XEvent * event)
{
 if (((int) unused == 5) || (event->type == ButtonRelease))
  {
   int lnW = ApData.drawinfo.lineWidth;
   int oldFG = ColorMatrix.FrgColorInd;

   if (!foregroundDraw)
    ColorSelectConseq(Button1, ColorMatrix.BckColorInd);

   sprayData.delay = 0;
   setExclEventHandler(NULL, 0, True);
   enableUndo(True);
   /* find the affected area */
   findMinMax(&Undo.points[0], Undo.npoints, &opBounds.x, &opBounds.y,
    &opBounds.width, &opBounds.height);

   opBounds.width -= opBounds.x;
   opBounds.height -= opBounds.y;

   if (lnW == 0)
     lnW++;
   opBounds.x -= (lnW + 1) / 2;
   opBounds.y -= (lnW + 1) / 2;
   opBounds.width += lnW + 1;
   opBounds.height += lnW + 1;

   segmentDrawOp();
   calcOpBounds();
   Undo.area.x = opBounds.x;
   Undo.area.y = opBounds.y;
   Undo.area.width = opBounds.width;
   Undo.area.height = opBounds.height;
   Undo.img = SubImage (ApData.img, opBounds.x, opBounds.y, opBounds.width, opBounds.height);
   Undo.imgAllocated = True;
   if (Undo.img == NULL)
     disableUndo ();

   LastDrawOnAllWin(True);
   setExclEventHandler((XtEventHandler) HandleBtnPress, ButtonPressMask, True);
   if (!foregroundDraw)
    ColorSelectConseq(Button1, oldFG);
  }
 else
  {
   sprayData.center.x = (event->xbutton.x + zoX) / zoomXF;
   sprayData.center.y = (event->xbutton.y + zoY) / zoomYF;
   sprayData.callAgain = False;
   sprayArea(NULL, NULL);
  }
}

DDRAWPRIM void sprayBtnPress(Widget w, XtPointer dataP, XEvent * event)
{
 if (event->xbutton.button==Button3)
  return;

 setExclEventHandler (NULL, 0, True);
 Undo.tool = ApData.tool = spray;
 setZoomParameters((int) dataP > 1);

 sprayData.w = w;
 sprayData.center.x = (event->xbutton.x + zoX) / zoomXF;
 sprayData.center.y = (event->xbutton.y + zoY) / zoomYF;
 sprayData.delay = 200;
 sprayData.callAgain = True;

 XCopyGC(xinfo.dpy, xinfo.gc,
      GCFunction | GCForeground | GCBackground | GCLineWidth, Undo.gc);

 memcpy(&Undo.drawinfo, &ApData.drawinfo, sizeof(DrawInformation));

 foregroundDraw = (event->xbutton.button==Button1);
 XSetFunction(xinfo.dpy, xinfo.gc, GXcopy);

 if (ApData.drawinfo.lineWidth > 1)
  {
   XGCValues values;

   Undo.drawinfo.lineWidth = 1;
   values.line_width = 0;
   XChangeGC(xinfo.dpy, Undo.gc, GCLineWidth, &values);
  }
 sprayArea(NULL, NULL);
 setExclEventHandler((XtEventHandler) sprayMotion,
  PointerMotionMask | ButtonReleaseMask, True);
}

static void drawArrow(Window win, int x, int y, int width, int height, char arrowType)
{
 XPoint a[6];
 int aW, aH, aX, aY;

 aW = width - 6;
 aH = height - 6;
 aX = x + 3;
 aY = y + aH + 3;
 a[5].x = a[0].x = aX;
 a[5].y = a[1].y = a[0].y = aY;
 a[1].x = a[2].x = aX + aW;
 a[3].y = a[2].y = aY - aH / 2;
 a[4].x = a[3].x = aX + aW / 2;
 a[4].y = y + 3;
 switch (arrowType)
  {
   case 2:	/* mirror X */
     a[1].y = a[0].y;
     a[2].x = a[4].x;
     a[2].y = a[4].y;
     a[4].x = a[0].x;
     a[4].y = a[3].y;
     break;
   case 1:	/* mirror Y */
     a[0].y = a[4].y;
     a[2].y = a[3].y;
     a[4].y = a[1].y;
     a[1].y = a[0].y;
     break;
   case 3:	/* mirror X & Y */
     a[2].y = a[0].y;
     a[1].y = a[0].y = a[4].y;
     a[2].x = a[3].x;
     a[4].x = a[0].x;
     a[4].y = a[3].y;
     break;
   default:
     break;
  }
 a[5].x = a[0].x;
 a[5].y = a[1].y;
 XFillPolygon(xinfo.dpy, win, xinfo.gc, &a[0], 6, Convex, CoordModeOrigin);
}

/* find in which part of the repeat is point x, y */
DDRAWPRIM int pointInRepeat(int x, int y, unsigned int repeatNo)
{
 int i, res, maxRp;
 Rectangle r;

 r.width = ApData.width;
 r.height = ApData.height;
 maxRp = MIN (MAX_ALL_REPEATS, repeatNo);
 res = -1;
 for (i = 0; i < maxRp; i++)
   {
     r.x = repeats.r[i].x;
     r.y = repeats.r[i].y;
     if (pointInRect (x, y, &r))
       {
	 res = i;
	 break;
       }
   }
 return(res);
}

static void checkOutSizes(int *repX, int *repY)
{
 if (*repX>MAX_LINE_REPEAT)
  *repX=MAX_LINE_REPEAT;
 if (*repY>MAX_LINE_REPEAT)
  *repY=MAX_LINE_REPEAT;
}

/* expose handler for repeat drawing area */
static void expose_repeat_DA(Widget w)
{
 Dimension daW, daH;
 int rpW, rpH, i, j, xP, yP, offX, offY, repX, repY;
 XGCValues values;

 if (!widgetIsViewable (w))
   return;
 checkOutSizes(repeatArea.repX, repeatArea.repY);
 repX = *repeatArea.repX;
 repY = *repeatArea.repY;

 XtVaGetValues(w, XmNwidth, &daW, XmNheight, &daH, NULL);
 rpW = daW / (repX + 1);
 rpH = daH / (repY + 1);

 offX = offY = 0;
 if ((*repeatArea.offset) && (repeatArea.mode))
  {
   if ((repeatArea.mode == 1) && (*repeatArea.distX))
     offX = ((*repeatArea.offset * rpW) / *repeatArea.distX);
   else if (*repeatArea.distY)
     offY = ((*repeatArea.offset * rpH) / *repeatArea.distY);
  }
 if (ApData.drawinfo.lineWidth > 1)
   {
     values.line_width = 0;
     XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);
   }

 XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
 XFillRectangle (xinfo.dpy, XtWindow(w), xinfo.gc, 0, 0, daW, daW);

 XSetForeground (xinfo.dpy, xinfo.gc, xinfo.black);

 repeatArea.num = 0;
 switch (repeatArea.mode)
   {
   case 2:
     xP = rpW / 2;
     for (j = 0; j < repX; j++)
       {
	 yP = rpH / 2;
	 if (repeatArea.fract)
	   yP += (j % repeatArea.fract) * offY;
	 else
	   yP += (offY * j) % (rpH - 1);
	 for (i = 0; i < repY; i++)
	   {
	     XDrawRectangle (xinfo.dpy, XtWindow (w), xinfo.gc, xP, yP, rpW, rpH);
	     drawArrow (XtWindow (w), xP, yP, rpW, rpH,
			repeatArea.rType[repeatArea.num]);
	     repeatArea.r[repeatArea.num].x = xP;
	     repeatArea.r[repeatArea.num].y = yP;
	     repeatArea.r[repeatArea.num].width = rpW;
	     repeatArea.r[repeatArea.num].height = rpH;
	     yP += rpH;
	     repeatArea.num++;
	   }
	 xP += rpW;
	 if ((repeatArea.num + repX) > MAX_DIFF_REPEATS)
	   break;
       }
     break;
   default:
     yP = rpH / 2;
     for (j = 0; j < repY; j++)
       {
	 xP = rpW / 2;
	 if (repeatArea.fract)
	   xP += (j % repeatArea.fract) * offX;
	 else
	   xP += (offX * j) % (rpW - 1);
	 for (i = 0; i < repX; i++)
	   {
	     XDrawRectangle (xinfo.dpy, XtWindow (w), xinfo.gc, xP, yP, rpW, rpH);
	     drawArrow (XtWindow (w), xP, yP, rpW, rpH,
			repeatArea.rType[repeatArea.num]);
	     repeatArea.r[repeatArea.num].x = xP;
	     repeatArea.r[repeatArea.num].y = yP;
	     repeatArea.r[repeatArea.num].width = rpW;
	     repeatArea.r[repeatArea.num].height = rpH;
	     xP += rpW;
	     repeatArea.num++;
	   }
	 yP += rpH;
	 if ((repeatArea.num + repX) > MAX_DIFF_REPEATS)
	   break;
       }
     break;
   }
 values.line_width = 3;
 XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);
 XDrawRectangle (xinfo.dpy, XtWindow (w), xinfo.gc,
  repeatArea.r[repeatArea.active].x, repeatArea.r[repeatArea.active].y, rpW, rpH);

 values.line_width = ApData.drawinfo.lineWidth;
 if (values.line_width == 1)
   values.line_width--;
 XChangeGC (xinfo.dpy, xinfo.gc, GCLineWidth, &values);
 XSetForeground (xinfo.dpy, xinfo.gc, ApData.drawinfo.foregrndC);
}

/* process drawing area input callback */
static void processRepeatInput(Widget w, XtPointer data, XEvent * event)
{
 if (event->type == ButtonPress)
  {
   int i;

   for (i = 0; i < repeatArea.num; i++)
     if ((event->xbutton.x >= repeatArea.r[i].x) &&
	 (event->xbutton.y >= repeatArea.r[i].y) &&
	 (event->xbutton.x <= (repeatArea.r[i].x + repeatArea.r[i].width)) &&
	 (event->xbutton.y <= (repeatArea.r[i].y + repeatArea.r[i].height)))
       {
	 repeatArea.active = i;
	 break;
       }
   if (repeatArea.active==0)
    XBell(xinfo.dpy, 50);
   else
    repeatArea.rType[repeatArea.active] = (event->xbutton.button == Button1) ?
     ((++repeatArea.rType[repeatArea.active])%4) : 0;
   expose_repeat_DA(repeatArea.DAW);
  }
}

/* calculate picture size on repeat size, type etc. */
static int calcPictSize(char what)
{
  Widget w;
  int v, mismatch;
  char tmp[200];

  v = mismatch = 0;
  checkOutSizes (repeatArea.repX, repeatArea.repY);
  if (what == 0)
    {
      w = repeatArea.sizeXW;
      v = *repeatArea.distX * (*repeatArea.repX);
      if (repeatArea.mode == 1)
	mismatch = *repeatArea.distX;
    }
  else
    {
      w = repeatArea.sizeYW;
      v = *repeatArea.distY * (*repeatArea.repY);
      if (repeatArea.mode == 2)
	mismatch = *repeatArea.distY;
    }
  sprintf (&tmp[0], "%d", v);
  if (*repeatArea.offset)
    mismatch = mismatch % *repeatArea.offset;
  else
    mismatch = 0;
  if (mismatch)
    strcat (&tmp[0], "!!!");
  setLabelString(w, tmp);
  return (v);
}

/* set repeat structures, increase/decrease picture size, convert repeats into a new big image */
static void setRepeats()
{
  int i, j, x, y, num;
  Boolean newImgSize = False;
  Rectangle rectOld, rectNew;

  if ((*repeatArea.distX == 0) || (*repeatArea.distY == 0))
    return;
  Select.rect_on = False;
  x = y = repeats.num = num = 0;
  checkOutSizes (repeatArea.repX, repeatArea.repY);
  repeats.rows = *repeatArea.repY;
  repeats.cols = *repeatArea.repX;
  repeats.distX = *repeatArea.distX;
  repeats.distY = *repeatArea.distY;
  repeats.width = calcPictSize (0);
  repeats.height = calcPictSize (1);
  repeats.mode = repeatArea.mode;
  switch (repeatArea.mode)
    {
    case 0:			/* block */
      for (i = 0; i < *repeatArea.repY; i++)
	{
	  x = 0;
	  for (j = 0; j < *repeatArea.repX; j++)
	    {
	      repeats.r[repeats.num].x = x;
	      repeats.r[repeats.num].y = y;
	      repeats.r[repeats.num++].rType = repeatArea.rType[num++];
	      x += *repeatArea.distX;
	    }
	  y += *repeatArea.distY;
	}
      break;
    case 1:			/* brick */
      for (i = 0; i < *repeatArea.repY; i++)
	{
	  x = (i ** repeatArea.offset) % *repeatArea.distX;
	  for (j = 0; j < *repeatArea.repX; j++)
	    {
	      repeats.r[repeats.num].x = x;
	      repeats.r[repeats.num].y = y;
	      repeats.r[repeats.num++].rType = repeatArea.rType[num++];
	      x += *repeatArea.distX;
	    }
	  if (*repeatArea.offset)
	    {
	      repeats.r[repeats.num].x = (i ** repeatArea.offset) % *repeatArea.distX
		- *repeatArea.distX;
	      repeats.r[repeats.num].y = y;
	      repeats.r[repeats.num].rType = repeats.r[repeats.num - 1].rType;
	      if ((repeats.r[repeats.num].x + *repeatArea.distX) > 0)
		repeats.num++;
	    }
	  y += *repeatArea.distY;
	}
      break;
    case 2:			/* pillar */
      for (i = 0; i < *repeatArea.repX; i++)
	{
	  y = (i ** repeatArea.offset) % *repeatArea.distY;
	  for (j = 0; j < *repeatArea.repY; j++)
	    {
	      repeats.r[repeats.num].x = x;
	      repeats.r[repeats.num].y = y;
	      repeats.r[repeats.num++].rType = repeatArea.rType[num++];
	      y += *repeatArea.distY;
	    }
	  if (*repeatArea.offset)
	    {
	      repeats.r[repeats.num].y = (i ** repeatArea.offset) % *repeatArea.distY
		- *repeatArea.distY;
	      repeats.r[repeats.num].x = x;
	      repeats.r[repeats.num].rType = repeats.r[repeats.num - 1].rType;
	      if ((repeats.r[repeats.num].y + *repeatArea.distY) > 0)
		repeats.num++;
	    }
	  x += *repeatArea.distX;
	}
      break;
    }
  XtVaGetValues (repeatArea.convertRPW, XmNset, &repeatArea.convertRP, NULL);
  XtVaGetValues (repeatArea.gridTBW, XmNset, &repeats.drawGrid, NULL);
  /* !! if dist > picture resize picture !! */
  rectNew.x = rectNew.y = rectNew.width = rectNew.height = 0;
  rectOld.x = rectOld.y = 0;
  rectOld.width = (unsigned int) ApData.width;
  rectOld.height = (unsigned int) ApData.height;
  if (repeatArea.convertRP)
    {
      XImage *tmp;
      Pixmap tmpPix;
      int oldWidth, oldHeight, oldRNum;

      selectCB(mainW, "3", NULL); /* clear any existing motif to avoid side effects */
      newImgSize = True;
      oldWidth = ApData.width;
      oldHeight = ApData.height;
      oldRNum = repeats.currNum;	/* a patch to enable the next routine */
      repeats.currNum = repeats.num;
      tmp = GetRepeatImage (ApData.img, 0, 0, repeats.width, repeats.height);
      repeats.currNum = oldRNum;
      ApData.width = repeats.width;
      ApData.height = repeats.height;
      tmpPix = safeCreatePixmap (xinfo.dpy, xinfo.W,
				 repeats.width, repeats.height, 1);

      /*clear pixmap */
      XSetForeground(xinfo.dpy, xinfo.gc1bit, 0);
      XFillRectangle(xinfo.dpy, tmpPix, xinfo.gc1bit, 0, 0,
       repeats.width, repeats.height);
      XSetForeground(xinfo.dpy, xinfo.gc1bit, 1);
      if ((!checkXImage (tmp)) || (!checkPixmap (tmpPix)))
	{
	  if (tmp)
	    undoXImage (&tmp);
	  if (tmpPix)
	    XFreePixmap (xinfo.dpy, tmpPix);
	  repeats.num = repeats.cols = repeats.rows = 1;
	  repeats.mode = 0;
	  ApData.width = oldWidth;
	  ApData.height = oldHeight;
	  return;
	}
      freeMainImage (&ApData.img);
      ApData.pixy = tmpPix;
      repeats.num = repeats.cols = repeats.rows = 1;
      repeats.mode = 0;
      ApData.DAwidth = repeats.width;
      ApData.DAheight = repeats.height;
      ApData.img = tmp;
      XtVaSetValues(drawAreaWID, XmNheight, ApData.height, XmNwidth, ApData.width,
       XmNbackground, ApData.drawinfo.backgrndC, XmNforeground, ApData.drawinfo.foregrndC, NULL);
      ApData.pic_change = True;
    }
  else
    {
      int currInd, opposite, diffX, diffY;
      Boolean occupied;

      if ((ApData.width != *repeatArea.distX) ||
	  (ApData.height != *repeatArea.distY))
	{
	  int oldRW, oldRH;

	  oldRW = repeats.width;
	  oldRH = repeats.height;
	  newImgSize = True;
	  rectNew.width = *repeatArea.distX;
	  rectNew.height = *repeatArea.distY;
	  changeSizePicture (&rectOld, &rectNew, False);
	  repeats.width = oldRW;
	  repeats.height = oldRH;
	}
      /* add surrounding repeats for drawing outside visible area */
      currInd = repeats.num;
      for (i = 0; i < repeats.num; i++)
	{
	  if (repeats.r[i].x == 0)
	    {
	      opposite = pointInRepeat (repeats.width - 1, repeats.r[i].y,
					repeats.num);
	      if (opposite > -1)
		{
		  repeats.r[currInd].x = -*repeatArea.distX;
		  repeats.r[currInd].y = repeats.r[opposite].y;
		  repeats.r[currInd++].rType = repeats.r[opposite].rType;
		}
	    }
	  if (repeats.r[i].y == 0)
	    {
	      opposite = pointInRepeat(repeats.r[i].x, repeats.height - 1, repeats.num);
	      if (opposite > -1)
		{
		  repeats.r[currInd].y = -*repeatArea.distY;
		  repeats.r[currInd].x = repeats.r[opposite].x;
		  repeats.r[currInd++].rType = repeats.r[opposite].rType;
		}
	    }
	  if ((repeats.r[i].x + *repeatArea.distX) == repeats.width)
	    {
	      opposite = pointInRepeat (0, repeats.r[i].y, repeats.num);
	      if (opposite > -1)
		{
		  repeats.r[currInd].x = repeats.width;
		  repeats.r[currInd].y = repeats.r[opposite].y;
		  repeats.r[currInd++].rType = repeats.r[opposite].rType;
		}
	    }
	  if ((repeats.r[i].y + *repeatArea.distY) == repeats.height)
	    {
	      opposite = pointInRepeat (repeats.r[i].x, 0, repeats.num);
	      if (opposite > -1)
		{
		  repeats.r[currInd].y = repeats.height;
		  repeats.r[currInd].x = repeats.r[opposite].x;
		  repeats.r[currInd++].rType = repeats.r[opposite].rType;
		}
	    }
	  if ((repeats.r[i].x == 0) && (repeats.r[i].y == 0))
	    {
	      opposite = pointInRepeat(repeats.width - 1, repeats.height - 1, repeats.num);
	      occupied = pointInRepeat(repeats.width + 1, repeats.height + 1, currInd) > -1;
	      if ((opposite > -1) && (!occupied))
		  {
		    repeats.r[currInd].x = repeats.width;
		    repeats.r[currInd].y = repeats.height;
		    repeats.r[currInd++].rType = repeats.r[i].rType;
		  }
	    }
	  if (((repeats.r[i].x + *repeatArea.distX) == repeats.width) &&
	      (repeats.r[i].y <= 0))
	    {
	      opposite = pointInRepeat (0, repeats.height - 1, repeats.num);
	      occupied = pointInRepeat (-1, repeats.height + 1, currInd) > -1;
	      if ((opposite > -1) && (!occupied))
		if
		  ((repeats.r[opposite].x <= 0) &&
		   ((repeats.r[opposite].y + *repeatArea.distY) == repeats.height))
		  {
		    opposite = pointInRepeat (repeats.width - 1, 
		    			repeats.height - 1, repeats.num);
		    diffY = repeats.height - repeats.r[opposite].y;
		    repeats.r[currInd].x = -*repeatArea.distX;
		    repeats.r[currInd].y = (int) repeats.height - diffY;
		    repeats.r[currInd++].rType = repeats.r[i].rType;
		  }
	    }
	  if (((repeats.r[i].y + *repeatArea.distY) == repeats.height) &&
	      (repeats.r[i].x <= 0))
	    {
	      opposite = pointInRepeat (repeats.width - 1, 0, repeats.num);
	      occupied = pointInRepeat (repeats.width + 1, -1, currInd) > -1;
	      if ((opposite > -1) && (!occupied))
		if
		  ((repeats.r[opposite].x + *repeatArea.distX == repeats.width) &&
		   (repeats.r[opposite].y <= 0))
		  {
		    opposite = pointInRepeat (repeats.width - 1, 
		    			repeats.height - 1, repeats.num);
		    diffX = repeats.width - repeats.r[opposite].x;
		    repeats.r[currInd].x = (int) repeats.width - diffX;
		    repeats.r[currInd].y = -*repeatArea.distY;
		    repeats.r[currInd++].rType = repeats.r[i].rType;
		  }
	    }
	  if (((repeats.r[i].y + *repeatArea.distY) >= repeats.height) &&
	      ((repeats.r[i].x + *repeatArea.distX) >= repeats.width))
	    {
	      opposite = pointInRepeat (0, 0, repeats.num);
	      occupied = pointInRepeat (-1, -1, currInd) > -1;
	      if ((opposite > -1) && (!occupied))
		if
		  ((repeats.r[opposite].x == 0) &&
		   (repeats.r[opposite].y == 0))
		  {
		    repeats.r[currInd].x = -*repeatArea.distX;
		    repeats.r[currInd].y = -*repeatArea.distY;
		    repeats.r[currInd++].rType = repeats.r[i].rType;
		  }
	    }
	}
      repeats.num = currInd;
    }
  if (newImgSize)
   {
    XtUnmanageChild(zoomDialogW);
    ApData.aspectRatio = (float) (1.0 * ApData.width)/ApData.height;
    changeTitleCB(NULL, ApData.fname, NULL);
   }
  repeats.crop.x = repeats.crop.y = 0;
  repeats.crop.width = ApData.width;
  repeats.crop.height = ApData.height;
  zoomDraw.newImage = True;

  setRepeatViewMode();
  /* we do not want automatic switch to repeat view if we have just created */
  /* a new image based on this repeat. Too complicated code to cleanly eliminate */
  /* it so we switch in and out */
  if (repeatArea.convertRP)
   XmToggleButtonSetState (normalViewTBW, True, True);
  if ((newImgSize) || (ApData.viewMode == repeatView))
    refreshDrawingArea (True, True);
  zoomMapWindow();    
}


static void valChangedCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case 4:
    calcPictSize(0);
    break;
   case 5:
    calcPictSize(1);
    break;
   default:
    expose_repeat_DA(repeatArea.DAW);
    calcPictSize (0);
    calcPictSize (1);
    break;
  }
}

#define RP_INIT_DLG 1
#define RP_MAP 2
#define RP_INIT_REPX 3
#define RP_INIT_REPY 4
#define RP_INIT_DISTX 5
#define RP_INIT_DISTY 6
#define RP_INIT_SIZEX 8
#define RP_INIT_SIZEY 7
#define RP_ARM_BLOCK 9
#define RP_ARM_BRICK 10
#define RP_ARM_PILLAR 11
#define RP_INIT_DA 12
#define RP_EXPOSE_DA 13
#define RP_INIT_OFF 14
#define RP_INIT_OPT_MENU 15
#define RP_OK 16
#define RP_RESIZE_DA 22
#define RP_OPTMENU 23
#define RP_INIT_TB 28
#define RP_INIT_GRID 29

DDRAWPRIM void repeatCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);
 static Widget rDlgW, optMnW;
 static inpNumStruct *repXInd, *repYInd, *distXInd, *distYInd, *offInd;

 switch (prm)
   {
   case RP_INIT_DLG:
     rDlgW = w;
     break;
   case RP_INIT_REPX:
     repXInd = initInputText (w, 2, 1, MAX_LINE_REPEAT);
     repeatArea.repX = &repXInd->value;
     XtAddCallback (w, XmNvalueChangedCallback, valChangedCB, "1");
     break;
   case RP_INIT_REPY:
     repYInd = initInputText (w, 2, 1, MAX_LINE_REPEAT);
     repeatArea.repY = &repYInd->value;
     XtAddCallback (w, XmNvalueChangedCallback, valChangedCB, "2");
     break;
   case RP_INIT_DISTX:
     distXInd = initInputText (w, 0, 0, MAX_IMAGE);
     repeatArea.distX = &distXInd->value;
     XtAddCallback (w, XmNvalueChangedCallback, valChangedCB, "4");
     break;
   case RP_INIT_DISTY:
     distYInd = initInputText (w, 0, 0, MAX_IMAGE);
     repeatArea.distY = &distYInd->value;
     XtAddCallback (w, XmNvalueChangedCallback, valChangedCB, "5");
     break;
   case RP_INIT_SIZEX:
     repeatArea.sizeXW = w;
     setLabelString(w, " ");
     break;
   case RP_INIT_SIZEY:
     repeatArea.sizeYW = w;
     setLabelString(w, " ");
     break;
   case RP_ARM_BLOCK:
     if (repeatArea.mode == 0)
       return;
     repeatArea.mode = 0;
     setInpArrVal (offInd, 0);
     offInd->enableChange = False;
     XtVaSetValues (offInd->inpField, XmNsensitive, False, NULL);
     XtVaSetValues (optMnW, XmNsensitive, False, NULL);
     expose_repeat_DA (repeatArea.DAW);
     calcPictSize (0);
     calcPictSize (1);
     break;
   case RP_ARM_BRICK:
   case RP_ARM_PILLAR:
     if (repeatArea.mode == (prm - RP_ARM_BLOCK))
       return;
     repeatArea.mode = prm - RP_ARM_BLOCK;
     offInd->enableChange = True;
     if (prm == RP_ARM_BRICK)
       offInd->upperB = ApData.width;
     else
       offInd->upperB = ApData.height;

     if (repeatArea.fract)
      {
       if (prm == RP_ARM_BRICK)
	 setInpArrVal (offInd, distXInd->value / repeatArea.fract);
       else
	 setInpArrVal (offInd, distYInd->value / repeatArea.fract);
      }
     XtVaSetValues (offInd->inpField, XmNsensitive, repeatArea.fract == 0, NULL);
     XtVaSetValues (optMnW, XmNsensitive, True, NULL);
     expose_repeat_DA (repeatArea.DAW);
     calcPictSize (0);
     calcPictSize (1);
     break;
   case RP_MAP:
     repeatArea.active = 0;
     setInpArrVal(distXInd, ApData.width);
     setInpArrVal(distYInd, ApData.height);
     offInd->upperB = MAX(ApData.width, ApData.height);
     repeatArea.convertRP = False;
     XtVaSetValues (repeatArea.convertRPW, XmNset, False, NULL);
     checkOutSizes (repeatArea.repX, repeatArea.repY);
     setInpArrVal (repXInd, *repeatArea.repX);
     setInpArrVal (repYInd, *repeatArea.repY);

     if (repeats.currNum == 1)
      {
       int i;

       for (i = 0; i < MAX_DIFF_REPEATS; i++)
	 repeats.r[i].rType = normal;
      }
     break;
   case RP_INIT_DA:
     repeatArea.DAW = w;
     repeatArea.mode = 1;
     XtAddEventHandler(w, ButtonPressMask, FALSE, (XtEventHandler) processRepeatInput, NULL);
     break;
   case RP_RESIZE_DA:
     repeatArea.active = 0;
     expose_repeat_DA(w);
     break;
   case RP_EXPOSE_DA:
     expose_repeat_DA(w);
     break;
   case RP_INIT_OFF:
     offInd = initInputText(w, 0, 0, 100);
     repeatArea.offset = &offInd->value;
     XtAddCallback (w, XmNvalueChangedCallback, valChangedCB, "3");
     break;
   case RP_INIT_OPT_MENU:
     optMnW = w;
     break;
   case RP_OK:
     setRepeats();
     if (ApData.viewMode == repeatView)
       repeats.currNum = repeats.num;
     else
       repeats.currNum = 1;
     break;
   case RP_OPTMENU:
     {
       int value = 0;
       Boolean Bval = False;

       repeatArea.fract = getMenuIndexFromWidget(w)+1;
       if (repeatArea.fract==1)
	 {
	   repeatArea.fract = 0;
	   Bval = True;
	 }
       if (repeatArea.mode == 1)
	 if (repeatArea.fract)
	   value = distXInd->value / repeatArea.fract;
       if (repeatArea.mode == 2)
	 if (repeatArea.fract)
	   value = distYInd->value / repeatArea.fract;
       if (True)
	 {
	   offInd->enableChange = Bval;
	   XtVaSetValues (offInd->inpField, XmNsensitive, Bval, NULL);
	   setInpArrVal (offInd, value);
	 }
       calcPictSize(0);
       calcPictSize(1);
     }
     break;
   case RP_INIT_TB:
     repeatArea.convertRPW = w;
     break;
   case RP_INIT_GRID:
     repeatArea.gridTBW = w;
     break;
   }
}

DDRAWPRIM void repeatCopyArea(Drawable drawable, XImage *img, GC gc, int xs, int ys, int w, int h,
 int xd, int yd, Boolean specialPut, repeatType mode, unsigned int mirrorWidth, unsigned int mirrorHeight,
 int srcOffsX, int srcOffsY, Region region)
{
 XImage *tmp, *baseImg;
 XPoint orR, rect[2];

 baseImg = NULL;
 if ((specialPut) && (drawable != ApData.pixy))
   if ((baseImg = GetRepeatImage(ApData.img, xd + srcOffsX, yd + srcOffsY, w, h)) == NULL)
     return;
 switch (mode)
   {
   case normal:
     if (drawable == ApData.pixy)
       CopyAreaImgTP (img, ApData.img, region,
		      0, 0, xs, ys, w, h, xd, yd);
     else if (specialPut)
       putXImageTP(baseImg, img, drawable, gc, xs, ys, w, h, 0, 0, xd, yd, region);
     else
       MyPutImage(drawable, gc, img, xs, ys, xd, yd, w, h, region);
     break;
   case mirrorX:
   case mirrorY:
   case mirrorXY:
     if ((tmp = doXImage (8, w, h)) == NULL)
       return;
     orR.x = mirrorWidth;
     orR.y = mirrorHeight;
     rect[0].x = xs;
     rect[0].y = ys;
     rect[1].x = xs + w - 1;
     rect[1].y = ys + h - 1;
     flipPts (&rect[0], 2, mode, &orR);
     findMinMax (&rect[0], 2, &rect[0].x, &rect[0].y, &rect[1].x, &rect[1].y);
     rect[1].x -= rect[0].x - 1;
     rect[1].y -= rect[0].y - 1;
     mirrorAreaImg(tmp, img, rect[0].x, rect[0].y, rect[1].x, rect[1].y, mode);
     if (drawable == ApData.pixy)
       CopyAreaImgTP(tmp, ApData.img, region, 0, 0, 0, 0, w, h, xd, yd);
     else if (specialPut)
       putXImageTP(baseImg, tmp, drawable, gc, 0, 0, w, h, 0, 0, xd, yd, region);
     else
       MyPutImage(drawable, gc, tmp, 0, 0, xd, yd, w, h, region);

     undoXImage(&tmp);
     break;
   }
 if (baseImg)
   undoXImage(&baseImg);
}

/* really perform those operations that weren't done with LastDrawOnAllWin */
DDRAWPRIM void finalDraw(Drawable drawable)
{
 switch (Undo.tool)
  {
   case selectA:
   case selectPolygonA:
   case mirrorHor:
   case mirrorVer:
   case mirrorFlip:
   case moveA:
   case pasteA:
   case rotateA:
   case resizeA:
   case contourT:
   case pasteMulA:
   case replaceCol:
   case swapColor:
   case mapColorsA:
     LastDraw(drawable);
     break;
   default:
     break;
  }
}

/* calculate operation boundaries and store it into the opBounds structure */
DDRAWPRIM void calcOpBounds()
{
 int lnW, i, start, np;
 XPoint orR, minMax[4];
 Boolean firstOp;

 lnW = Undo.drawinfo.lineWidth;
 opBounds.x = opBounds.y = opBounds.width = opBounds.height = 0;
 firstOp = True;
 minMax[0].x = minMax[0].y = minMax[1].x = minMax[1].y = 0;
 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;
 switch (Undo.tool)
   {
   case lineDraw:
   case curveDraw:
   case spray:
   case arcDraw:
   case polyDraw:
     np = Undo.npoints;
     if (Undo.tool == arcDraw)
       {
	 np = Undo.points[np - 1].x;
	 start = Undo.npoints - np - 1;
       }
     else
       start = 0;
     for (i = 0; i < repeats.currNum; i++)
       if (repeats.draw[i])
	 {
	   movePts (&Undo.points[start], np, -repeats.r[i].x, -repeats.r[i].y);
	   flipPts (&Undo.points[start], np, repeats.r[i].rType, &orR);
	   findMinMax(&Undo.points[start], np, &minMax[2].x, &minMax[2].y, &minMax[3].x, &minMax[3].y);
	   if (firstOp)
	     {
	       firstOp = False;
	       memcpy (&minMax[0], &minMax[2], 2 * sizeof (XPoint));
	     }
	   else
	     findMinMax (&minMax[0], 4, &minMax[0].x, &minMax[0].y, &minMax[1].x, &minMax[1].y);
	   flipPts (&Undo.points[start], np, repeats.r[i].rType, &orR);
	   movePts (&Undo.points[start], np, repeats.r[i].x, repeats.r[i].y);
	 }
     break;
   case rectDraw:
   case squareDraw:
   case ellipseDraw:
   case circDraw:
   case cutA:
   case cutClearA:
   case copyA:
   case mirrorHor:
   case mirrorVer:
   case mirrorFlip:
   case moveA:
     if ((Undo.tool == circDraw) || (Undo.tool == ellipseDraw))
       start = 1;
     else
       start = 0;
     for (i = 0; i < repeats.currNum; i++)
       if (repeats.draw[i])
	{
	 memcpy (&minMax[2], &Undo.points[start], sizeof (XPoint));
	 minMax[3].x = minMax[2].x + Undo.points[start + 1].x;
	 minMax[3].y = minMax[2].y + Undo.points[start + 1].y;
	 movePts (&minMax[2], 2, -repeats.r[i].x, -repeats.r[i].y);
	 flipPts (&minMax[2], 2, repeats.r[i].rType, &orR);
	 if (firstOp)
	  {
	   firstOp = False;
	   findMinMax(&minMax[2], 2, &minMax[0].x, &minMax[0].y, &minMax[1].x, &minMax[1].y);
	  }
	 else
	  findMinMax(&minMax[0], 4, &minMax[0].x, &minMax[0].y, &minMax[1].x, &minMax[1].y);
	}
     break;
   case selectA:
   case selectPolygonA:
   case pasteA:
   case rotateA:
     break;
   default:
     break;
   }

 if (lnW == 0)
   lnW++;
 minMax[0].x -= (lnW + 1) / 2;
 minMax[0].y -= (lnW + 1) / 2;
 minMax[1].x += lnW + 1;
 minMax[1].y += lnW + 1;
 opBounds.x = MAX (minMax[0].x, 0);
 opBounds.y = MAX (minMax[0].y, 0);
 minMax[1].x = MAX (minMax[1].x, 0);
 minMax[1].y = MAX (minMax[1].y, 0);
 minMax[1].x = MIN (minMax[1].x, ApData.width);
 minMax[1].y = MIN (minMax[1].y, ApData.height);
 opBounds.width = minMax[1].x - opBounds.x;
 opBounds.height = minMax[1].y - opBounds.y;
}

static void findModifiedRepeats(Rectangle * opB, Boolean clearNotaffected)
{
 int i;
 Rectangle repB, resR;

 repB.width = repeats.crop.width;
 repB.height = repeats.crop.height;
 for (i = 0; i < repeats.currNum; i++)
  {
   repB.x = repeats.r[i].x;
   repB.y = repeats.r[i].y;
   if (intersectRect (opB, &repB, &resR))
     {
       repeats.draw[i] = True;
       if (repeats.firstDraw == -1)
	 repeats.firstDraw = i;
     }
   else if (clearNotaffected)
     repeats.draw[i] = False;
  }
}

/* find out which repeats are affected by drawing operation */
DDRAWPRIM void segmentDrawOp()
{
 Rectangle opB;
 XPoint orPt, rect[2];

 repeats.firstDraw = -1;
 rect[0].x = rect[1].x = opB.x = opBounds.x;
 rect[0].y = rect[1].y = opB.y = opBounds.y;
 rect[1].x = opB.width = opBounds.width;
 rect[1].y = opB.height = opBounds.height;
 orPt.x = repeats.width;
 orPt.y = repeats.height;
 findModifiedRepeats(&opB, True);

 if (repeats.firstDraw == -1)
   repeats.firstDraw = 0;
}

/* areaR is a rectangle consisting of several repeats */
/* boundR is a rectangle which may lay in areaR */
/* function finds out which repeats of the areaR rectangle */
/* are in inside boundR */
/* returns the result in res array - allocated by the caller */
/* and maximal boundaries in resR */
DDRAWPRIM int whichRepeats(Rectangle * areaR, Rectangle * boundR, Rectangle * resR)
{
 int i, len, Aind;
 XPoint r[2], maxB[4], orR;
 Rectangle partR, resRect;

 orR.x = repeats.crop.width;
 orR.y = repeats.crop.height;
 len = Aind = 0;

 maxB[0].x = maxB[0].y = maxB[1].x = maxB[1].y = 0;
 for (i = 0; i < repeats.currNum; i++)
  {
   r[0].x = r[1].x = boundR->x;
   r[0].y = r[1].y = boundR->y;
   r[1].x += boundR->width;
   r[1].y += boundR->height;
   flipPts (&r[0], 2, repeats.r[i].rType, &orR);
   movePts (&r[0], 2, repeats.r[i].x, repeats.r[i].y);
   findMinMax (&r[0], 2, &r[0].x, &r[0].y, &r[1].x, &r[1].y);
   partR.x = r[0].x;
   partR.y = r[0].y;
   partR.width = r[1].x - r[0].x;
   partR.height = r[1].y - r[0].y;
   if (intersectRect(&partR, areaR, &resRect))
    {
     len++;
     maxB[Aind].x = maxB[Aind + 1].x = resRect.x;
     maxB[Aind].y = maxB[Aind + 1].y = resRect.y;
     maxB[Aind + 1].x += resRect.width;
     maxB[Aind + 1].y += resRect.height;
     if (Aind == 0)
      Aind = 2;
     else
      findMinMax(&maxB[0], 4, &maxB[0].x, &maxB[0].y,
       &maxB[1].x, &maxB[1].y);
    }
  }
 resR->x = maxB[0].x;
 resR->y = maxB[0].y;
 resR->width = maxB[1].x - maxB[0].x;
 resR->height = maxB[1].y - maxB[0].y;
 return (len);
}

/* draw primitive using numerically entered parameters */
static void drawPrimitive()
{
 XPoint pts[3];

 pts[0].x = drawPrm[0]->value;
 pts[0].y = ApData.height - 1 - drawPrm[1]->value;
 pts[1].x = drawPrm[2]->value;
 pts[1].y = ApData.height - 1 - drawPrm[3]->value;

 if ((Undo.tool == ApData.tool) && (ApData.tool != none))
   undoCB (NULL, NULL, NULL);
 switch (ApData.tool)
   {
   case lineDraw:
     memcpy (&Undo.points[0], &pts[0], sizeof (XPoint) * 2);
     Undo.npoints = 2;
     findMinMax (&pts[0], 2, &opBounds.x, &opBounds.y,
		 &opBounds.width, &opBounds.height);
     opBounds.width -= opBounds.x;
     opBounds.height -= opBounds.y;
     break;
   case rectDraw:
   case squareDraw:
     findMinMax (&pts[0], 2, &opBounds.x, &opBounds.y,
		 &opBounds.width, &opBounds.height);
     Undo.points[1].x = opBounds.width -= opBounds.x;
     Undo.points[1].y = opBounds.height -= opBounds.y;
     Undo.points[0].x = opBounds.x;
     Undo.points[0].y = opBounds.y;
     Undo.npoints = 2;
     if (ApData.tool == squareDraw)
       {
	int dx, dy;

	dx = opBounds.width;
	dy = opBounds.height;
	if (dx >= dy)
	  dy = (int) (dx/densityRatio());
	else
	 {
	  dx = dy;
	  dy = (int) (dy/densityRatio());
	 }
	Undo.points[1].x = opBounds.width = dx;
	Undo.points[1].y = opBounds.height = dy;
	setInpArrVal(drawPrm[0], Undo.points[0].x);
	setInpArrVal(drawPrm[1], ApData.height - 1 - Undo.points[0].y);
	setInpArrVal(drawPrm[2], Undo.points[0].x + dx);
	setInpArrVal(drawPrm[3], ApData.height - 1 - (Undo.points[0].y + dy));
       }
     break;
   case ellipseDraw:
   case circDraw:
     findMinMax(&pts[0], 2, &opBounds.x, &opBounds.y,
	&opBounds.width, &opBounds.height);
     Undo.points[2].x = opBounds.width -= opBounds.x;
     Undo.points[2].y = opBounds.height -= opBounds.y;
     Undo.points[1].x = opBounds.x;
     Undo.points[1].y = opBounds.y;
     Undo.points[3].x = 0;
     Undo.points[3].y = 23040;
     Undo.npoints = 4;
     if (ApData.tool == circDraw)
      {
       int dx, dy;

       dx = opBounds.width;
       dy = opBounds.height;
       if (dx >= dy)
	 dy = (int) (dx/densityRatio());
       else
	 {
	   dx = dy;
	   dy = (int) (dy/densityRatio());
	 }
       Undo.points[2].x = opBounds.width = dx;
       Undo.points[2].y = opBounds.height = dy;
       setInpArrVal(drawPrm[0], Undo.points[1].x);
       setInpArrVal(drawPrm[1], ApData.height - 1 - Undo.points[1].y);
       setInpArrVal(drawPrm[2], Undo.points[1].x + dx);
       setInpArrVal(drawPrm[3], ApData.height - 1 - (Undo.points[1].y + dy));
      }
     break;
   default:
     return;
     break;
   }
 opBounds.x -= ApData.drawinfo.lineWidth;
 opBounds.y -= ApData.drawinfo.lineWidth;
 opBounds.width += ApData.drawinfo.lineWidth * 2;
 opBounds.height += ApData.drawinfo.lineWidth * 2;

 Undo.tool = ApData.tool;

 segmentDrawOp ();
 calcOpBounds ();
 enableUndo (True);

 Undo.area.x = opBounds.x;
 Undo.area.y = opBounds.y;
 Undo.area.width = opBounds.width;
 Undo.area.height = opBounds.height;
 Undo.img = SubImage(ApData.img, opBounds.x, opBounds.y, opBounds.width, opBounds.height);
 if (Undo.img == NULL)
   disableUndo ();
 else
   Undo.imgAllocated = True;
 XCopyGC(xinfo.dpy, xinfo.gc, GCFunction | GCLineStyle | GCLineWidth | GCForeground |
  GCBackground | GCJoinStyle | GCCapStyle, Undo.gc);
 memcpy (&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));

 LastDrawOnAllWin (True);
 ApData.pic_change = True;
}

DDRAWPRIM void initDrawPrmDlg(unsigned short prmNo, char *lblString, int value,
 int lowerB, int upperB)
{
 WidgetList wList = NULL;
 int numChildren = 0, i = 0;

 if (prmNo > 5)
   return;
 setInpArrVal (drawPrm[prmNo], value);
 drawPrm[prmNo]->lowerB = lowerB;
 drawPrm[prmNo]->upperB = upperB;
 XtVaGetValues(XtParent(drawPrm[prmNo]->inpField), XmNnumChildren, &numChildren, XmNchildren, &wList, NULL);
 while ((i < numChildren) && (!XmIsLabel(*wList)))
   { i++; wList++; }
 if (i != numChildren)
   setLabelString(*wList, lblString);
 XtManageChild(XtParent(drawPrm[prmNo]->inpField));
}

#define DP_MAP 1
#define DP_UNMAP 2
#define DP_OK 3
#define DP_CANCEL 4
#define DP_INIT_1 5
#define DP_INIT_2 6
#define DP_INIT_3 7
#define DP_INIT_4 8
#define DP_PREVIEW 9
#define DP_INIT_PREVIEW 10

/* callback for handling numerical input of drawing operations */
DDRAWPRIM void drawPrmCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);
 static Boolean drawnOnce;
 static Widget previewW;

 switch (prm)
  {
   case DP_INIT_1:
   case DP_INIT_2:
   case DP_INIT_3:
   case DP_INIT_4:
     {
      int ind;

      ind = prm - DP_INIT_1;
      drawPrm[ind] = initInputText (w, 0, 0, 13000);
      XtVaSetValues(drawPrm[ind]->inpField, XmNcolumns, 5, XmNmaxLength, 5, NULL);
     }
     break;
   case DP_MAP:
     drawnOnce = False;
     if ( (ApData.tool == spray) || (ApData.tool == mix) )
       XtUnmanageChild(previewW);
     else
       XtManageChild(previewW);
     break;
   case DP_UNMAP: /* unmanage all parameters */
     {
       WidgetList wList;
       int numChildren, i;

       XtVaGetValues (XtParent(XtParent(drawPrm[0]->inpField)),
	XmNnumChildren, &numChildren, XmNchildren, &wList, NULL);
       for (i = 0; i < numChildren; i++)
	 { XtUnmanageChild(*wList); wList++; }
     }
     break;
   case DP_PREVIEW:
     drawPrimitive();
     drawnOnce = True;
     break;
   case DP_OK:
     if ( (ApData.tool == spray) || (ApData.tool == mix) )
       {
	sprayData.sprayPts = drawPrm[0]->value;
	sprayData.radius = drawPrm[1]->value;
       }
     else if (!drawnOnce)
       drawPrimitive();
     break;
   case DP_CANCEL:
     if (drawnOnce)
       undoCB(NULL, NULL, NULL);
     break;
   case DP_INIT_PREVIEW:
     previewW = w;
     break;
   default:
     break;
  }
}
