/* loadimg.c */
#define LOADIMG

#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "paint.h"

static initConsumeProc initConsume;
static consumeBandProc consumeBand;
static endConsumeProc endConsume;
static tryLoadProc tryLoad;
static loadBandProc loadBand;
static endLoadProc endLoad;

static int nowLoading = FALSE;	/* so we dont't start loading again */
static char *loadTxt[MAX_LTXT];	/* localized strings for load interface */

/* load buffer parameters */
static unsigned char *data;	/* pointer to data */
static int bpl;			/* bytes per line */
static int bandh;		/* number of lines in one band */
static int progress;		/* advance in the data buffer or stay put */
static unsigned char cmap[256*3]; /* permanent cmap for colormap images */
static char loaderInfo[500];	/* tryload fills this in */

/* attributes of image being loaded */
static int loadType;		/* IMG_BW ... IMG_TRUE_COL */
static int xdpi;		/* horizontal dots per inch *1000 */
static int ydpi;		/* vertical dots per inch *1000 */
static int xpixels;
static int ypixels;

static int loadStatus;		/* the status of loading */

/* customize load interface */
static void initLoadInterface()
{
/******* BEGIN customization */
  loadTxt[LOAD_USER_ABORT] = getTxtI8N (OPERATION_CANCELLED_MSG);
  loadTxt[LOAD_NO_MEMORY] = getTxtI8N (NOT_ENOUGH_MEMORY_MSG);
  loadTxt[LOAD_IO_ERROR] = getTxtI8N (ERROR_IO_MSG);
  loadTxt[LOAD_INTERNAL_ERROR] = getTxtI8N (ERROR_INTERNAL_CALL_SUPPORT_MSG);
  loadTxt[LOAD_WRONG_IMG] = getTxtI8N (ERROR_WRONG_IMAGE_FOR_OPERATION_MSG);
  loadTxt[LOAD_FILETYPE_UNKNOWN] = getTxtI8N (NOT_A_BITMAP_FILE_MSG);
  loadTxt[LOAD_FILE_CORRUPT] = getTxtI8N (ERROR_FILE_CORRUPT_MSG);
  loadTxt[LTXT_WAIT] = getTxtI8N (WAIT_MSG);
  loadTxt[LTXT_LOADING_WAIT] = getTxtI8N (LOADING_WAIT_MSG);
  loadTxt[LTXT_CANCEL_LOAD] = getTxtI8N (CANCEL_LOADING_MSG);
  loadTxt[LOAD_NO_FREE_COLORS] = getTxtI8N (NOT_ENOUGH_COLORS_MSG);
/******* END customization */
  setDialogTitle(printingTW, loadTxt[LTXT_WAIT]);
  setLabelString(printCancelW, loadTxt[LTXT_CANCEL_LOAD]);
}

/* user wishes to cancel the load */
DLOADIMG void cancelLoadCB(Widget w, XtPointer clientData, XtPointer callData)
{
 loadStatus = LOAD_USER_ABORT;
}

static void processBand(int ypos)
{
 XmScaleSetValue (printProgressSW, (ypos * 100) / ypixels);
 if ((loadStatus != LOAD_OK) || (ypos >= ypixels))
  {
   XtUnmanageChild (printingTW);
   endLoad ();
   endConsume (loadStatus);
   if (loadStatus > LOAD_USER_ABORT)
     manageInfoDialog (loadTxt[loadStatus]);
   nowLoading = FALSE;
   return;
  }

 loadStatus = loadBand(data, min (bandh, ypixels - ypos), bpl);
 if (loadStatus == LOAD_OK)
  loadStatus = consumeBand(data, min (bandh, ypixels - ypos), bpl, ypos);

 ypos += min (bandh, ypixels - ypos);
 if (progress)
  data = &data[bpl * bandh];
 XtAddTimeOut(adaptiveTimeout(), (XtTimerCallbackProc) *processBand,
  (XtPointer) ypos);
}

/* loads the image from filename, using all available loaders, and consuming */
/* it with the passed functions; can specify desired image type, since sometimes */
/* already the loader can convert the image */
DLOADIMG int loadImage(char *filename, initConsumeProc iConsume,
 consumeBandProc cBand, endConsumeProc eConsume, int desiredLoadType)
{
 int i, numLoaders, numInterlace;
 char reverse;

 /* only one load job at a time */
 if (nowLoading)
  return(LOAD_INTERNAL_ERROR);

 numLoaders = returnLoaderFunctions(-1);
 initLoadInterface();
 initConsume = iConsume;
 consumeBand = cBand;
 endConsume = eConsume;
 loadStatus = LOAD_OK;

 if (access (filename, R_OK) != 0)
  loadStatus = LOAD_IO_ERROR;
 else
  for (i = 0; i < numLoaders; i++)
   {
    nowLoading = TRUE;
    returnLoaderFunctions(i);
    loadType = desiredLoadType;
    loadStatus = tryLoad(filename, &loadType,
    &xpixels, &ypixels, &xdpi, &ydpi, cmap, loaderInfo, &reverse, &numInterlace, TRUE);
    if (loadStatus == LOAD_OK)
     {
      i = numLoaders; /* to exit loop */
      XmScaleSetValue(printProgressSW, 0);
      loadStatus = initConsume(loadType, xpixels, ypixels, xdpi, ydpi, reverse,
       numInterlace, cmap, &data, &bpl, &bandh, &progress);
      if (loadStatus == LOAD_OK)
       {
	char waitInfo[200];

	if (xpixels*ypixels < 640000) /* do images below 800x800 in one pass */
	 processBand(0);
	else
	 {
	  sprintf(waitInfo, "%s\n%s\n%d x %d\n%s", loadTxt[LTXT_LOADING_WAIT],
	   baseName(filename), xpixels, ypixels, loaderInfo);
	  setLabelString(printingInfoLW, waitInfo);
	  XtAddTimeOut(adaptiveTimeout()*10, (XtTimerCallbackProc) *processBand, 0);
	  XtManageChild(printingTW);
	 }
       }
      else
       endLoad();
     }
   }
 if (loadStatus != LOAD_OK)
  {
   manageInfoDialog(loadTxt[loadStatus]);
   nowLoading = FALSE;
  }
 return(loadStatus);
}

/* tries to get image information before loading it; returns "" on fail */
DLOADIMG char *getImageInfo(char *filename, int *x, int *y)
{
 int i, numLoaders, numInterlace;
 char reverse;

 if (nowLoading)
  return("");
 numLoaders = returnLoaderFunctions(-1);
 initLoadInterface();
 if (access(filename, R_OK)==0)
  if (filename[strlen(filename)-1]!='/')
   for (i=0; i<numLoaders; i++)
    {
     returnLoaderFunctions(i);
     if (LOAD_OK == tryLoad(filename, &loadType, x, y, &xdpi, &ydpi,
      cmap, loaderInfo, &reverse, &numInterlace, FALSE))
      { endLoad(); return(loaderInfo); }
    }
 return("");
}

/* registers loader with system */
void registerLoader(tryLoadProc tLoad, loadBandProc lBand, endLoadProc eLoad)
{ tryLoad = tLoad; loadBand = lBand; endLoad = eLoad; }
