/*   ArahPaint
 *
 *   Copyright (C) 1993-2001  Arahne d.o.o.
 *   http://www.arahne.si
 *   E-mail: arahne@arahne.si
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <X11/cursorfont.h>
#include <stdio.h>
#include "../plib/print.h"
#include "../llib/loadimg.h"

#define MITSHM

#define SWAP(a,b) {a^=b; b^=a; a^=b;}
#define SGN(a)  (((a)<0) ? -1: +1)
#define EQUAL(str1,str2)	!strcmp((str1),(str2))

#ifndef MAX
#define MAX(a,b)                ((a)<(b) ? (b) : (a))
#endif
#ifndef MIN
#define MIN(a,b)		((a)<(b) ? (a) : (b))
#endif

#define MAX_CELLS      256 /* max number of colormap entries */
#define RED		0
#define GREEN		1
#define BLUE		2

#define MODEL_RGB	0
#define MODEL_HSL	1

#define INCH_MM		25.4

#define REAL_VIEW_SIZE	9000 // size of view area in real density view
#define IMG_BUF_Y	250  // height of screen buffer for refreshes; reused

/* maximum image size; will not present a problem until 2GB RAM will be common on desktop */
/* but many other things will need to be corrected in program before we can increase this */
/* like for example maximum window/drawable size in X Window ... */
#define MAX_IMAGE	32000


/* file destinations for print to file and for saving CAM files */
#define FILE_TO_A	0
#define FILE_TO_B	1
#define FILE_TO_TMP	2
#define FILE_TO_OTHER	3 /* full path is already specified in filename */

#ifndef MAIN
#define DMAIN extern
#else
#define DMAIN
#endif

DMAIN struct
{
 Display *dpy;		/* pointer to the display */
 Window W;		/* main window id; for creation of other stuff */
 int screenNo;		/* number of screen that we are using */
 Visual *visual;	/* default visual */
 int depth;		/* bits per pixel */
 GC gc;			/* default graphics context */
 GC gc1bit;		/* gc with depth 1 */
 Colormap cmap;		/* default colormap */
 XColor allocCells[MAX_CELLS];	/* allocated colors */
 unsigned long pixelTC[MAX_CELLS]; /* pixels for true color mode */
 unsigned char red[MAX_CELLS], green[MAX_CELLS], blue[MAX_CELLS];
 unsigned char redBitsNo, greenBitsNo, blueBitsNo; /* how many bits of RGB for every pixel */
 char redShift, greenShift, blueShift;/* position of RGB bits inside pixel; pozitive
  value means left shift, right otherwise */
 int dpyw, dpyh;		/* width and height of display */
 unsigned long black, white;	/* black and white pixel value on this colormap */
 Boolean trueColor;		/* indicator of current color mode */
 XVisualInfo *visualInfo;	/* information about the visual */
} xinfo;

typedef struct
{
  Widget inpField;
  int lowerB, upperB;
  int value;
  Boolean enableChange;
} inpNumStruct;

DMAIN struct
{
  short x, y, width, height;
} opBounds;

#define INP_NUM_COUNT 36
DMAIN inpNumStruct inpNumArr[INP_NUM_COUNT];

#ifdef MITSHM
#include <X11/extensions/XShm.h>
#endif
DMAIN struct
{
 Boolean present;	/* shared memory present */
 Boolean alloc;		/* shared memory allocated */
 XImage *img;		/* shared memory image */
#ifdef MITSHM
 XShmSegmentInfo Info;	/* shared memory segment info */
#endif
} Shm;

DMAIN Widget drawAreaWID, mainW, clipWin, scrollBX, scrollBY, zoomDialogW,
 warnDialogW, UndoBW, CutBW, PasteBW, CopyBW, DCircBW, DRectBW, DSquareBW,
 fillBW, unselectBW, DArcBW, DEllipseBW, colSelW, colSelDialogW, rotateDialogW,
 color_comp1W, color_comp2W, color_comp3W, scrollWinW, errorDialogW, fileSelectionW,
 helpDialogW, helpTextW, printDialogW, widthPrnLW, heightPrnLW, zoomDrawAreaW,
 posXYW, undoZoomBW, workingDialogW, infoDialogW, normalViewTBW, repeatViewTBW,
 DpolyDrawBW, zoomDrawW, drawPrmW, zoomScaleW, inchmmLW;
DMAIN Widget printingTW, printingInfoLW, printProgressSW, printCancelW,
  printerSelBW, printModeSelBW, printXdpiSelBW, printYdpiSelBW, printToFileIW,
  printToFileTBW, printToFileDstW, copiesW, curveDW;

DMAIN Window drawAreaWin;
DMAIN int initWidth, maxWidth;
DMAIN int initHeight, maxHeight;
DMAIN int screenWidthMM, screenHeightMM;
DMAIN int metricSystem; /* if false, use inches */
DMAIN int maxColors;		/* total number of colors */
DMAIN Boolean enableSHM;	/* enable shared memory use */
DMAIN char *pathCommon;		/* path of common directory */
DMAIN char *printCommand;	/* print command string */
DMAIN char *defaultPrinter;	/* default printer */
DMAIN char *copy2dosCommand;	/* command used on this OS to copy to/from MSD floppy */

DMAIN void exposeRepeat(Drawable drawable, int x, int y, int widthE, int heightE, 
 int destX, int destY, Boolean drawGrid, Boolean final_draw);

DMAIN XImage *GetRepeatImage(XImage *srcImg, int x, int y, int width, int height);
DMAIN void setRepeatViewMode(void);

/* init DrawingArea widget */
DMAIN void initDrawAreaCB(Widget w, XtPointer clientData, XtPointer callData);

typedef enum
{ normalView, repeatView, densityView } viewType;

#ifndef DRAW
#define DDRAW extern
#else
#define DDRAW
#endif

typedef struct
{
 int x, y;
 unsigned int width, height;
} Rectangle;

typedef struct
{
 enum
  { Up, Down, Left, Right } direction;
 int offs;
} ShiftData;

typedef enum
{
 none, selectA, selectPolygonA, curveDraw, lineDraw, shiftHor, 
 shiftVer, rectDraw, squareDraw, circDraw, arcDraw, ellipseDraw,
 polyDraw, mirrorHor, mirrorVer, mirrorFlip, fill, rotateByLine,
 cutA, cutClearA, copyA, pasteA, pasteMulA, moveA, rotateA, resizeA, contourT,
 swapColor, grayConv, makeGradation, drawText, replaceCol, spray, guessWeave,
 mapColorsA, mix
} Sel_tool;

typedef enum
{ solidFill, patternFill } FillTypes;
  
typedef struct
{
 int lineWidth, lineStyle;
 unsigned long foregrndC, backgrndC;
 Boolean fill;
 FillTypes fillType;
} DrawInformation;

typedef enum
{
 undef, move, resizeUp, resizeDown, resizeLeft, resizeRight,
 resizeDownLeft, resizeDownRight, resizeUpLeft, resizeUpRight
} SelectState;

typedef struct
{
 Boolean selected; /* region selected */
 Boolean rect_on;  /* bounding rectangle/polygon visible */
 XPoint *points;   /* array of XPoint where bounding points of rectangle polygon are stored */
 int npoints;	   /* number of points in points array */
 int allocPoints;  /* number of allocated points in points array */
 Boolean rectangular;/* is the selection is rectangular or not */
 Region reg;	   /* region for nonrectangular selection */
 GC gc, moveGC;	   /* GC for displaying boundaries and moving selection */
 int x, y, width, height;
 Boolean resLeft, resRight, resUp, resDown; /* for resizing op */
 Boolean preserveAspectR;	/* preserve aspect ratio for resizing op */
 float aspectR;	/* aspect ratio width: height */
 SelectState state; /* status of the mouse cursor */
 char fname[FILENAME_MAX];/* file name of the selection */
 XImage *img;
 Boolean imgAllocated;
 int colorNo;	/* number of colors in selection */
 Boolean colorCons;  /* colors of selection are consistent with main palette*/
 Boolean extraColors; /* if motif was loaded and extra colors were allocated */
 unsigned char red[MAX_CELLS], green[MAX_CELLS], blue[MAX_CELLS]; /* colors by RGB components */
 XColor allocCells[MAX_CELLS]; /* structure holding info about all colors */
 unsigned long pixelTC[MAX_CELLS]; /* pixels for true color mode */
} SelectInfo;

DDRAW SelectInfo Select;

typedef struct
{
 char fname[FILENAME_MAX];
 XPoint *points;
 int npoints, allocPoints;
 Boolean rectangular, colorCons;
 GC gc;
 int x, y, width, height;
 XImage *img;
 Boolean imgAllocated;
 int colorNo;
 unsigned char colors[MAX_CELLS];
 unsigned char red[MAX_CELLS], green[MAX_CELLS], blue[MAX_CELLS];
} MotifInfo;

DDRAW struct
{
 Sel_tool tool;
 XPoint *points;
 int npoints, allocPoints;
 GC gc;
 DrawInformation drawinfo;
 XImage *img;
 Boolean imgAllocated;
 Rectangle area;
} Undo;


DDRAW struct
{
 int x_offs, y_offs, row, col, width, height, gap;
 int FBx, FBy, FBwidth, FBheight;
 int BckColorInd, FrgColorInd;
 float aspectR, r_width, r_height;
 int mainWinWidth, mainWinHeight;
 Widget mainWinDAW;
} ColorMatrix;

DDRAW struct
{
 Sel_tool tool;		/* selected tool */
 DrawInformation drawinfo;
 Pixmap pixy;		/* main pixmap, color selection pixmap */
 int ColorNo;		/* number of colors*/
 int TranspColNo;	/* number of transparent colors */
 unsigned char TranspColors[MAX_CELLS];/* transparent color indexes */
 int ProtColNo;		/* number of protected colors */
 unsigned char ProtectColors[MAX_CELLS];/* protected color indexes */
 unsigned char transfTbl[MAX_CELLS][MAX_CELLS];
 Boolean specialPutImg;	/* if transp/prot colors then perform special type of PutImage */
 viewType viewMode;	/* current view mode */
 float aspectRatio;	/* width: height ratio*/
 double imageXdpi, imageYdpi; /* X and Y dpi of the image; it stays persistent after loading new image */
 unsigned int cursorShape, oldCursorShape;
 int width, height;	/* image width & height */
 int DAwidth, DAheight;	/* drawing area width & height */
 Rectangle clipWin;	/* clip window measures */
 char fname[FILENAME_MAX];/* file name of the edited picture */
 XPoint orig_xy;	/* point where the click occured */
 Boolean pic_change;	/* picture changed status flag */
 XImage *img;		/* backup image */
 Boolean imgAllocated;	/* status flag of backup image */
 int zoomX, zoomY;	/* current zoom factor*/
 char *imgPath;		/* image directory used in file selection */
 char *helpFile;	/* help messages file name */
 char *msgFile;		/* messages file name */
 char fontList[255];	/* font set in configuration file */
 XFontStruct *currFont;	/* currently loaded font */
 XtEventHandler CurrEventHandler;/* current event handler */
 EventMask eventMask;	/* mask of event handler */
 char *version;		/* program version */
 XImage *scrImg;	/* image for screen redraw */
 Pixmap scrPix;		/* pixmap for screen redraw; using SHM should be the same as scrImg */
} ApData;

/* disabled by default! until refreshes are corrected */
DDRAW int pixmapIsXimage; /* if we are on local machine with MIT-SHM the screen pixmap and Ximage are the same */
DDRAW int foregroundDraw; /* should we draw in foreground or background color */

DDRAW double densityRatio();
DDRAW double getImageXmm();
DDRAW double getImageYmm();

/* Generic callback for buttons in toolbar */
DDRAW void toolBarCB(Widget w, XtPointer clientData, XtPointer callData);

/* Undo Callback */
DDRAW void undoCB(Widget w, XtPointer clientData, XtPointer callData);

/* Draw the last picture element to the drawable */
DDRAW void LastDraw(Drawable drawable);

DDRAW void LastDrawOnAllWin(Boolean definitely);

/* position and refresh zoom window to the mouse position in the main window */
DDRAW void centerZoomWindowToMouse(XEvent *event);

/* Handle mouse button press inside drawing area w */
DDRAW void HandleBtnPress(Widget w, XtPointer unused, XEvent * event);

/*set /remove appropriate event handler according to selected tool*/
DDRAW void setToolEventHandler(Sel_tool tool, unsigned int curshape);
DDRAW void setExclEventHandler(XtEventHandler NewEH, EventMask eventMask, Boolean onZoomWin);

/* wrap around settings callback */
DDRAW void wrapAroundCB(Widget w, XtPointer clientData, XtPointer callData);

/* set aspect ratio callback - for drawing squares and circles properly*/
DDRAW void aspectRatioCB(Widget w, XtPointer clientData, XtPointer callData);

DDRAW void findMinMax(XPoint *pts, int numPts, short *minX, short *minY, short *maxX, short *maxY);

DDRAW void movePts(XPoint *xp, int pointNo, int dx, int dy);

DDRAW void refreshRepeatView(Drawable drawable, Rectangle *clipWinR,  Boolean wholeWindow, Boolean final_draw);

#ifndef DRAWPRIM
#define DDRAWPRIM extern
#else
#define DDRAWPRIM
#endif

typedef enum
{
 normal, mirrorX, mirrorY, mirrorXY
} repeatType;

typedef struct
{
 int x, y; /* offset to the normal coordinates of picture */
 repeatType rType;
} repeatStruct;

DDRAWPRIM  struct
{
 Widget w;
 int delay;
 XPoint center;
 Boolean callAgain;
 Boolean mixMode; /* if true, mix existing colors, else work as normal spray */
 int sprayPts, radius;
}
sprayData;

#define MAX_LINE_REPEAT 16
#define MAX_DIFF_REPEATS 256
#define MAX_ALL_REPEATS 3*MAX_DIFF_REPEATS+MAX_LINE_REPEAT
DDRAWPRIM struct
{
 int currNum; /* current number of different repeats */
 int num;     /* total number of different repeats */
 int rows, cols; /* number of rows & columns */
 Rectangle crop; /* crop of the whole image */
 unsigned short width, height;
 unsigned short distX, distY; /* difference between two consecutive repeats */
 unsigned char mode;
 int firstDraw;  /* repeat where drawing operation began */
 Boolean draw[MAX_ALL_REPEATS]; /* where the draw operation occurred */
 repeatStruct r[MAX_ALL_REPEATS]; /* position of each repeat */
 Boolean drawGrid;
} repeats;

DDRAWPRIM inpNumStruct *drawPrm[6];

DDRAWPRIM void drawCross(Drawable d, GC gc, int x, int y, unsigned short l);

DDRAWPRIM void line_widthCB(Widget w, XtPointer clientData, XtPointer callData);

DDRAWPRIM void drawLines(Drawable drawable, GC gc, Boolean fillRegion, 
 XPoint *p, int pointsNo, int offsX, int offsY, Boolean contiguous);

DDRAWPRIM void DrawLinesWrapped(Drawable drawable, GC gc, Boolean fillRegion, 
 XPoint *p, int pointsNo, int offsX, int offsY, Boolean contiguous);

DDRAWPRIM void DrawRectangleWrapped(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY);

DDRAWPRIM void DrawRectangle(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY);

DDRAWPRIM void drawArc(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY);

DDRAWPRIM void DrawArcWrapped(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY);

DDRAWPRIM int DrawElArc(int Px, int Py, int Qx, int Qy, int Kx, int Ky, XPoint * el_pts);

DDRAWPRIM void DrawElArcWrapped(Drawable drawable, GC gc, DrawInformation DI,
 XPoint *p, int offsX, int offsY);

DDRAWPRIM void polygonDraw(Widget w, XtPointer dataP, XEvent * event);

DDRAWPRIM void sprayBtnPress(Widget w, XtPointer dataP, XEvent * event);

DDRAWPRIM void repeatCB(Widget w, XtPointer clientData, XtPointer callData);

DDRAWPRIM int pointInRepeat(int x, int y, unsigned int repeatNo);

DDRAWPRIM void repeatCopyArea(Drawable drawable, XImage *img, GC gc, int xs, int ys,
 int w, int h, int xd, int yd, Boolean specialPut, repeatType mode, unsigned int mirrorWidth,
 unsigned int mirrorHeight, int srcOffsX, int srcOffsY, Region region);

DDRAWPRIM void finalDraw(Drawable drawable);

DDRAWPRIM void calcOpBounds(void);

DDRAWPRIM void segmentDrawOp(void);

DDRAWPRIM void flipPts(XPoint *xp, int pointNo, repeatType rpt, XPoint *orR);

DDRAWPRIM void sprayWrapped(Drawable drawable, GC gc, XPoint *p, int pointsNo, int offsX, int offsY);

DDRAWPRIM int whichRepeats(Rectangle *areaR, Rectangle *boundR, Rectangle *resR);

DDRAWPRIM void drawPrmCB (Widget w, XtPointer clientData, XtPointer callData);

DDRAWPRIM void initDrawPrmDlg(unsigned short, char *lblString, int value, int lowerB, int upperB);

#ifndef RESIZE
#define DRESIZE extern
#else
#define DRESIZE
#endif

DRESIZE void resizePicture(XImage * xsi, XImage * xdi, Rectangle * src, Rectangle * dst);

/* scales selected area to new size determined in Undo.points[newsizeP]*/
DRESIZE Boolean scaleSelectionArea(int newSizePtr);

/* Undo scale selection area*/
DRESIZE void undoScaleSelectionArea(void);

/* Initialize Change Size Dialog*/
DRESIZE void InitChSizeDlgCB(Widget w, XtPointer clientData, XtPointer callData);

/* Change size of picture*/
DRESIZE void ChSizePicCB(Widget w, XtPointer clientData, XtPointer callData);

DRESIZE void getRepeatPatternCB(Widget w, XtPointer clientData, XtPointer callData);

DRESIZE void changeSizePicture(Rectangle * oldSize, Rectangle * newSize, Boolean fit2size);


#ifndef MISC
#define DMISC extern
#else
#define DMISC
#endif

/* allocate one full screen buffer for screen redraw, use MIT_SHM if possible; in this */
/* case also create Pixmap, which can be used for XCopyArea instead of expensive XPutImage */
DMISC int allocateBuffers(int allocate);

/* copies the shared image (Pixmap) or Ximage to the Window */
DMISC void xPutBuffer(Window win, int srcx, int srcy, int dstx, int dsty, int wid, int hei);

/* painlessly sets XmString string to resource of widget */
DMISC void xtVaSetString(Widget w, String resource, char *setting);

/* sets the passed string to widget as titleString */
DMISC void setTitleString(Widget w, char *str);

/* sets the passed string to widget as labelString */
DMISC void setLabelString(Widget w, char *str);

DMISC char *xmStringToCstring(XmString motifString);

DMISC XImage *doXImage(int depth, int x, int y);

DMISC void undoXImage(XImage **theXImage);

DMISC int widgetIsViewable(Widget w);

DMISC void installErrorHandler(void);

DMISC void RGBtoHSL(int r, int g, int b, int *h, int *s, int *l);

DMISC void HSLtoRGB(int h, int s, int l, int *r, int *g, int *b);

DMISC void patchScrollBarPos(Widget hsb, Widget vsb, int x, int y);

DMISC void ChangeCursor(Window win, unsigned int shape, Boolean StoreOld);

DMISC Cursor getWindowCursor(Window win);

DMISC void RestoreCursor(Window win);

DMISC void getClipWinDims(Widget clipWinW, Widget scrollBarX, Widget scrollBarY,
 unsigned int maxWidth, unsigned int maxHeight, Rectangle *clipWinR);

DMISC void refreshWindow(Window win, XImage *src, int src_x,
 int src_y, int width, int height, int dest_x, int dest_y, Boolean final_draw);

DMISC void refreshDrawingArea(Boolean wholeWindow, Boolean final_draw);

DMISC Boolean fileExists(char *filename);

DMISC void manageWarningDialog(char *message, XtCallbackProc OKCallback, char *OKData, char *OKStr,
 XtCallbackProc CancelCallback, char *CancelData, char *CancelStr,
 XtCallbackProc HelpCallback, char *HelpData, char *HelpStr);

DMISC void manageErrorDialog(char *message, XtCallbackProc OKCallback, char *OKData, char *OKStr);

DMISC void manageInfoDialog(char *message);

DMISC void disableUndo(void);

DMISC void enableUndo(Boolean clearImg);

DMISC Boolean checkXImage(XImage *img);

DMISC Boolean checkPixmap(Pixmap pixy);

DMISC void customizeMessageBoxCB(Widget w, XtPointer clientData, XtPointer callData);

DMISC void mousePos(Widget w, XtPointer unused, XEvent *event);

DMISC void setButtonIconCB(Widget w, XtPointer clientData, XtPointer callData);

DMISC int ReadCFGFile(char *progName);

DMISC void checkKbdInputCB(Widget w, XtPointer clientData, XtPointer callData);

DMISC int getIntFromWidget(Widget w, int lowerB, int upperB);

DMISC int setIntToWidget(Widget w, int value, int lowerB, int upperB);

DMISC double getDoubleFromWidget(Widget w, double lowerB, double upperB, unsigned char decimals);
DMISC double setDoubleToWidget(Widget w, double value, double lowerB, double upperB, unsigned char decimals);

DMISC inpNumStruct *initInputText(Widget inpFieldW, int value, int lowerB, int upperB);

DMISC void setInpArrVal(inpNumStruct * ind, int value);

DMISC void incDecTextFieldCB(Widget w, XtPointer clientData, XtPointer callData);

DMISC Widget GetShellParent(Widget w);
DMISC Widget GetDialogParent(Widget w);

DMISC Boolean putXImageTP(XImage *baseImg, XImage * srcImg, Drawable drwbl, GC gc,
 int srcX, int srcY, int width, int height, int pixX, int pixY, int dstX, int dstY, Region region);
			   
DMISC int checkLocalAndMITSHM(void);
DMISC void clearSHM(void);
DMISC Pixmap CreatePixmap(Display *display, Drawable drawable, unsigned int width, unsigned int height,
 unsigned int depth);
DMISC Pixmap safeCreatePixmap(Display *display, Drawable drawable, int width, int height, int depth);
DMISC void FreePixmap (Display * display, Pixmap pixy);

DMISC Boolean intersectRect(Rectangle *first, Rectangle *second, Rectangle *result);

DMISC XImage *SubImage(XImage *src, int x, int y, unsigned int width, unsigned int height);

DMISC void highlightButton(Widget w);

DMISC Boolean pointInRect(int x, int y, Rectangle *r);

DMISC void CopyAreaImg(XImage *src, XImage *dest, Region clipRegion, int clipX, int clipY, int src_x,
 int src_y, unsigned int width, unsigned int height,  int dest_x, int dest_y);

DMISC void CopyAreaImgTP(XImage *src, XImage *dest, Region clipRegion, int clipX, 
 int clipY, int src_x, int src_y, unsigned int width, unsigned int height, int dest_x, int dest_y);

DMISC XImage *MyGetImage(XImage *img, int x, int y, unsigned int width,
 unsigned int height, viewType viewMode);

DMISC void freeMainImage(XImage **img);

DMISC XImage *createMainImage(unsigned int width, unsigned int height);

DMISC void draw2Image(XImage *img, Pixmap pixy, Boolean shared, int x, int y, unsigned int width,
 unsigned int height, unsigned char drawColorInd);

DMISC void MyPutImage(Drawable drawable, GC gc, XImage *image, int src_x, int src_y,
 int dest_x, int dest_y, unsigned int width, unsigned int height, Region region);

DMISC int adaptiveTimeout(void);

/* set icon image and name */
DMISC void setIcon(Widget shell, XImage *xim);

DMISC char *baseName(char *path);
DMISC char *dirName(char *path);

/* sets the passed string to widget as dialogTitle */
DMISC void setDialogTitle(Widget w, char *str);

/* returns the position of widget in the menu */
DMISC int getMenuIndexFromWidget(Widget w);

/* sets the state on toggle button widgets */
DMISC void setState(Widget w, int state);

/* makes the widget sensitive to input */
DMISC void setSensitive(Widget w, int state);


#ifndef MIRROR
#define DMIRROR extern
#else
#define DMIRROR
#endif

DMIRROR void mirrorPictureCB(Widget w, XtPointer clientData, XtPointer callData);
DMIRROR void mirrorImgHorizontal(XImage *img, unsigned int rows);
DMIRROR void mirrorImgVertical(XImage *img, unsigned int lines);
DMIRROR void mirrorImg(XImage *img, repeatType how);

DMIRROR void mirrorAreaImgHorizontal(XImage *dstImg, XImage *srcImg, int x, int y, 
 unsigned int width, unsigned int height);

DMIRROR void mirrorAreaImgVertical(XImage *dstImg, XImage *srcImg, int x, int y, 
 unsigned int width, unsigned int height);

DMIRROR void mirrorAreaImg(XImage *dstImg, XImage *srcImg, int x, int y, 
 unsigned int width, unsigned int height, repeatType how);

DMIRROR void ShiftPic(int w, int h);

DMIRROR void shiftPictureCB(Widget w, XtPointer clientData, XtPointer callData);

#ifndef FILL
#define DFILL extern
#else
#define DFILL
#endif

DFILL void fillArea(int x, int y, XImage *pattern);

#ifndef FILES
#define DFILES extern
#else
#define DFILES
#endif

#define MAX_HISTORY_LIST 5
#define HISTORY_LIST_FILE_NAME .apaint_history
DFILES struct
{
 Widget id[MAX_HISTORY_LIST];
 char *name[MAX_HISTORY_LIST];
 unsigned char num;
 char fileN[512];
}
historyList;

DFILES void NewFileCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void reallyExitCB(Widget w, XtPointer clientData, XtPointer callData);

/* return -1 if process can't be created, 0 on failure, 1 on success */
DFILES int executeCommand(char *command);

/* reads the floppy in MSDOS or Staubli format and copies the files to /tmp/floppy */
DFILES void readFloppyCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void loadPictButtonCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void loadPictureCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void saveDrawingCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void savePictureCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void printPictureCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void changeTitleCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES Pixmap loadBitmapFile(char *filename, Display *dpy, Window win, unsigned int *width, unsigned int *height,
 Pixel fore, Pixel back);

DFILES void historyListCB(Widget w, XtPointer clientData, XtPointer callData);

DFILES void initHistoryList(int first);

#ifndef SELECT
#define DSELECT extern
#else
#define DSELECT
#endif

DSELECT int EraseSqrSize;

/* selection button callback */
DSELECT void selectCB(Widget w, XtPointer clientData, XtPointer callData);

/* move selection rectangle around; if floating, move the whole block, if still in image */
/* move the selection boundaries */
DSELECT void moveSelection(int changeX, int changeY);

/* cut & copy button callback */
DSELECT void cutCopyCB(Widget w, XtPointer clientData, XtPointer callData);

/* paste move callback */
DSELECT void pastemoveCB(Widget w, XtPointer clientData, XtPointer callData);

/*check motions and actions inside selection*/
DSELECT void handleSelection(Widget w, XtPointer unused, XEvent * event);

/* Button Press event handler for selection moving & copying */
DSELECT void SelectBtnPress(Widget w, XtPointer unused, XEvent * event);

/* draw boundaries of the selection area*/
DSELECT void drawSelectionBoundaries(Window win, int zoomX, int zoomY, int offsX, int offsY);
DSELECT void drawSelectionBoundariesWrapped(Window win, int zoomX, int zoomY, int offsX, int offsY);

/*set boundary points of selection area*/
DSELECT void setBoundaryPtsR(int *First, int x, int y, int width, int height);

DSELECT void disableSelection(Boolean disableButtons);

DSELECT void multiplePasteCB(Widget w, XtPointer clientData, XtPointer callData);

DSELECT void moveBoundaryPts(int First, int Last, int dx, int dy);

DSELECT void pasteMotifWrapped(Drawable drawable);

#ifndef COLORSIM
#define DCOLORSIM extern
#else
#define DCOLORSIM
#endif

DCOLORSIM Pixel GrayBckGround;

/* allocate all available colors, set random values for them and release all unneccessary colors afterwards */
/* for servers which have unassigned color set to black - xor won't work */
DCOLORSIM int initAllAvailableColors(void);

DCOLORSIM int allocateMoreColors(unsigned int colsWanted, int firstEmpty);

DCOLORSIM void freeColors(int col2free);

DCOLORSIM void colorToolsCB(Widget w, XtPointer clientData, XtPointer callData);

/* Restore previously selected colors */
DCOLORSIM void restoreColorCB(Widget w, XtPointer clientData, XtPointer callData);

/* set number of active (used) colors */
DCOLORSIM void setColorNoCB(Widget w, XtPointer clientData, XtPointer callData);

/* create color selection area with colors colors */
DCOLORSIM void initColorSelectionArea(int color_no, int w, int h);

/* Act according to selected color in color selection area or drawing area */
DCOLORSIM void ColorSelectConseq(unsigned int MouseButton, unsigned long ColInd);

/* init color selection widget */
DCOLORSIM void initColSelCB(Widget w, XtPointer clientData, XtPointer callData);

DCOLORSIM void redrawColSelCB(Widget w, XtPointer clientData, XtPointer callData);

DCOLORSIM void changeColorKbd(Window win, int xPos, int yPos, short button);

DCOLORSIM void ColorSelectPress(Widget w, XtPointer clientData, XEvent * event);

DCOLORSIM void drawColorPalette(void);

#ifndef COLOR2
#define DCOLOR2 extern
#else
#define DCOLOR2
#endif

/* calculate color transformation table */
DCOLOR2 void calcColTransfTbl();

DCOLOR2 void transpProtColCB(Widget w, XtPointer clientData, XtPointer callData);

DCOLOR2 Boolean addColor2Array(unsigned char color, Boolean checkPassive, Boolean soundOn);

DCOLOR2 void cutOutColorBorder(void);

DCOLOR2 void calcTrueColorPixels(XColor *colCells, unsigned long *TCpixel, 
 unsigned char *red, unsigned char *green, unsigned char *blue);

DCOLOR2 unsigned long eight2MoreBits(int red, int green, int blue);

DCOLOR2 void convertImg2TrueColor(XImage *img8, XImage *imgTC, int x, int y,
 unsigned int width, unsigned int height, unsigned int magnX, unsigned int magnY);

DCOLOR2 void UDcolRemapCB(Widget w, XtPointer clientData, XtPointer callData);

DCOLOR2 Boolean addPivotColor(unsigned char color);

#ifndef QUANT
#define DQUANT extern
#else
#define DQUANT
#endif
DQUANT unsigned long getDistance(unsigned char *col1, unsigned char *col2);

DQUANT void initHist(unsigned long *colHist);

DQUANT void addToHist(unsigned long *colHist, unsigned char *data, int x, int y, int width, int height, int bpl);

DQUANT int countHistColors(unsigned long *colHist);

DQUANT void remapData(unsigned char *data, unsigned char *transTable, int length);

DQUANT void remapImgArea(XImage *img, unsigned char *transTable, int x, int y,
 unsigned int width, unsigned int height, Region region);

DQUANT void addToHistFromRegion(unsigned long *colHist, unsigned char *data, int x, int y, 
 unsigned int width, unsigned int height, unsigned int bpl, Region region, int dx, int dy);

DQUANT void quantize(XImage * srcDstImage, int requestNumCol, int *resultNumCol,
 unsigned char *srcCmap, XColor * allocCells, unsigned char *redCo, unsigned char *greenCo,
 unsigned char *blueCo, unsigned char *transTable, unsigned long *colHist);

DQUANT void quantizeOnCmap(XImage * img, int imgCol, unsigned char *redCo, unsigned char *greenCo,
 unsigned char *blueCo, unsigned char *trTbl);

#ifndef ROTATE
#define DROTATE extern
#else
#define DROTATE
#endif

/*rotate button callback */
DROTATE void rotateCB(Widget w, XtPointer clientData, XtPointer callData);

DROTATE void rotate90(int direction, XImage *src, XImage *dst);

DROTATE void rotatePictureByLine(void);

#ifndef CONTOUR
#define DCONTOUR extern
#else
#define DCONTOUR
#endif

DCONTOUR void doContourCB(Widget w, XtPointer clientData, XtPointer callData);

#ifndef HELP
#define DHELP extern
#else
#define DHELP
#endif

DHELP void helpCB(Widget w, XtPointer clientData, XtPointer callData);
DHELP Boolean initHelp(char *helpFName);
DHELP void endHelp(void);
DHELP void helpTools(Sel_tool tool);
DHELP int ReadMsgFile(char *msgFile);
DHELP void SetWidgetStrCB(Widget w, XtPointer clientData, XtPointer callData);
DHELP void SetWidgetStrDotsCB(Widget w, XtPointer clientData, XtPointer callData);
DHELP char *getTxtI8N(int ind);

#ifndef ZOOMDRAW
#define DZOOMDRAW extern
#else
#define DZOOMDRAW
#endif

DZOOMDRAW Widget scrollX, scrollY, posXYZm;
DZOOMDRAW int blockZoomRefreshes;

DZOOMDRAW struct
{
 int width, height;
 unsigned int factorX, factorY, thickX, thickY;
 Rectangle clipWin, clipWin11;
 int scrollW, scrollH;
 Boolean newImage;
 Boolean drawMesh;
} zoomDraw;

DZOOMDRAW void redrawZoomDAreaCB(Widget w, XtPointer clientData, XtPointer callData);

DZOOMDRAW void handleZoomDrawCB(Widget w, XtPointer clientData, XtPointer callData);

DZOOMDRAW void mousePosZoom(Widget w, XtPointer unused, XEvent *event);

DZOOMDRAW void exposeTkanina(int x, int y, int width, int height, Widget w);

DZOOMDRAW void setZoomFactorCB(Widget w, XtPointer clientData, XtPointer callData);

DZOOMDRAW void refreshZoomWin(short x, short y, unsigned short width, unsigned short height,
 Boolean refreshFromSource);
DZOOMDRAW void zoomMapWindow(void);

/* unified function to generate real view image for both screen and print in any depth */
DZOOMDRAW void renderRealImg(unsigned char *data, int bpl, int bytesPerPixel,
 int xpos, int ypos, int width, int height, int firstWeftPix, int xdpi, int ydpi);

#ifndef PRINT
#define DPRINT extern
#else
#define DPRINT
#endif

DPRINT int selectedPrintDst; /* index of selected destination, if print to file */

/* set the value of print to file and destination */
DPRINT void setPrintToFileCB(Widget w, XtPointer clientData, XtPointer callData);

/* customize print interface */
DPRINT void initPrtInterfaceCB(Widget w, XtPointer clientData, XtPointer callData);

/* user wishes to cancel the print */
DPRINT void cancelPrintCB(Widget w, XtPointer clientData, XtPointer callData);

/* returns localized string containing current printer settings */
DPRINT char *returnPrtSettingText(void);

/* get the printable limits of current printer */
DPRINT void getPrinterLimits(int *maxxpix, int *maxypix, int *prtxdpi,
 int *prtydpi, int *nonprtTop, int *nonprtBottom, int *nonprtLeft);

/* set page settings */
DPRINT void setPageSettings(int pxpixels, int pypixels,
 int pxmargin, int pymargin, int pcomplexImage);

/* application registers its routine that will be called by printer driver */
/* to get data, which will be printed */
DPRINT void registerGetBandProc(getNextBandProc getNextBandP);

/* prints the page according to previous settings */
DPRINT void printPage(getNextBandProc getNextBandP);

/* put the file in the print queue of the currently selected printer or copy it to */
/* the selected destination */
DPRINT int enqueuePrintFile(char *fn);

/* cleanup of temporary files */
DPRINT void removeFileFreeFilename(char *filename);

/* save an image in the passed format with print interface; return false if file not writeable */
DPRINT int saveImage(char *fname, XImage *xim, char *imgFormat, unsigned char *imgcmap,
 int freeWhenOver, int destination);

#ifndef LOADIMG
#define DLOADIMG extern
#else
#define DLOADIMG
#endif
DLOADIMG void cancelLoadCB(Widget w, XtPointer clientData, XtPointer callData);

/* loads the image from filename, using all available loaders, and consuming */
/* it with the passed functions; can specify desired image type, since sometimes */
/* already the loader can convert the image */
DLOADIMG int loadImage(char *filename, initConsumeProc iConsume,
 consumeBandProc cBand, endConsumeProc eConsume, int desiredLoadType);

#define ONLY_FIXED	0
#define PREF_FIXED	1
#define PREF_PROP	2

#ifndef FONTSEL
#define DFONTSEL extern
#else
#define DFONTSEL
#endif
DFONTSEL XFontStruct *FontSel(Widget parent, int showPropFonts, char *currFont);
DFONTSEL void moveString(Widget w, XtPointer prm, XEvent * event);
DFONTSEL void inpFontCB(Widget w, XtPointer clientData, XtPointer callData);

#ifndef TOOLS
#define DTOOLS extern
#else
#define DTOOLS
#endif
DTOOLS void toolsCB(Widget w, XtPointer clientData, XtPointer callData);

#define YES_MSG				1
#define NO_MSG				2
#define OK_MSG				3
#define EXIT_MSG			4
#define CANCEL_MSG			5
#define ABORT_MSG			6
#define NEW_PICTURE_MSG			10
#define LOAD_PICTURE_MSG		11
#define SAVE_PICTURE_MSG		12
#define DELETE_FILE_MSG			14
#define PROGRAM_EXIT_MSG		15
#define EDIT_MSG			25
#define UNDO_MSG			26
#define HELP_MSG			48
#define RED_MSG				69
#define GREEN_MSG			70
#define BLUE_MSG			71
#define HUE_MSG				72
#define SATURATION_MSG			73
#define LIGHT_MSG			74
#define NUMBER_OF_COLOR_USED_MSG	82
#define FONT_ISO_MSG			83
#define NOT_ENOUGH_COLORS_MSG		88
#define ZOOM_MSG			96

#define MOTIF_WIDTH_MSG			102
#define MOTIF_HEIGHT_MSG		103
#define WIDTH_REPEAT_NUMBER_MSG		105
#define HEIGHT_REPEAT_NUMBER_MSG	106
#define INVALID_FONT_MSG		107
#define FILE_SELECTION_OK_MSG		108
#define FILE_SELECTION_CANCEL_MSG	109
#define FILE_SELECTION_SEARCH_MSG	110
#define FILE_SELECTION_HELP_MSG		111
#define FILE_SELECTION_NO_MATCH_MSG	112
#define FILE_SELECTION_DIRECTORY_MSG	113
#define FILE_SELECTION_FILES_MSG	114
#define FILE_SELECTION_FILTER_MSG	115
#define FILE_SELECTION_SELECTION_MSG	116

#define MESSAGE_FILE_IN_PRINT_QUEUE_OK_MSG 123
#define PRINT_CANCELLED_MSG		124
#define ERROR_IO_MSG			125
#define ERROR_INTERNAL_CALL_SUPPORT_MSG	126
#define ERROR_PROCESS_CREATION_MSG	127
#define ERROR_WRONG_IMAGE_FOR_PRINT_MSG	128
#define WAIT_MSG			129
#define PRINTING_WAIT_MSG		130
#define CANCEL_PRINT_MSG		131
#define PRINTER_MSG			132
#define PRINT_MODE_MSG			133
#define PRINT_DENSITY_X_MSG		134
#define PRINT_DENSITY_Y_MSG		135
#define WARNING_PAPER_MSG		136
#define PRINT_TO_FILE_MSG		137
#define PRINT_TO_FILE_OK_MSG		138
#define PRINT_MODE_BLACK_AND_WHITE_MSG	139
#define PRINT_MODE_COLOR_DITHER_MSG	140
#define PRINT_MODE_COLOR_TRUE_MSG	141
#define PRINT_MODE_COLORMAP_MSG		152
#define SAVING_WAIT_MSG			153
#define CANCEL_SAVE_MSG			154
#define LOAD_MOTIF_MSG			155
#define SAVE_MOTIF_MSG			156
#define PREVIEW_MSG			158

#define OPERATION_CANCELLED_MSG		161
#define ERROR_WRONG_IMAGE_FOR_OPERATION_MSG 162
#define ERROR_FILE_CORRUPT_MSG		163
#define LOADING_WAIT_MSG		164
#define CANCEL_LOADING_MSG		165
#define INPUT_TEXT_MSG			166
#define SIZE_MSG			167
#define STYLE_MSG			168
#define FONT_MSG			169
#define FONT_SELECTOR_MSG		170
#define SHOW_PROP_FONT_MSG		171
#define SHOW_SIZE_IN_POINTS_MSG		172
#define NUMBER_OF_POINTS_MSG		183
#define RADIUS_MSG			184
#define NEW_MSG				185
#define SAVING_TO_FLOPPY_OK_MSG		204

#define NOT_ENOUGH_MEMORY_MSG		500
#define COLOR_NUM_LESS_THEN_USED_COLORS_MSG 501
#define NO_SELECTION_MADE_MSG		502
#define CONTOUR_COLOR_ALREADY_SET_MSG	503
#define PICTURE_CHANGED_SAVE_MSG	504
#define FILE_EXISTS_OVERWRITE_MSG	505
#define PIXMAP_CREATION_ERROR_MSG	507
#define NOT_A_BITMAP_FILE_MSG		508
#define UNABLE_TO_SAVE_PICTURE_MSG	509

#define POINT_OUT_OF_SELECTED_AREA_MSG	512
#define NO_HELP_FILE_MSG		513
#define NO_RELATED_HELP_TEXT_MSG	514
#define NO_APPROPRIATE_HELP_MSG		515
#define UNABLE_TO_RESIZE_SELECTED_AREA_MSG 516
#define VIEW_ONLY_IN_THIS_MODE_MODE_MSG	517
#define ONLY_RECTANGULAR_AREAS_MAY_BE_ROTATED_MSG 518
#define PROG_PATH_ERROR_MSG		519
#define RESIZE_PICTURE_FATAL_ERROR_MSG	520
#define SELECTION_EXISTS_REPLACE_MSG	521
