#define ROTATE
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "paint.h"

/* Scan line shearing algorithm by Alan W. Paeth from "Graphics Gems", Academic Press, 1990 */

static void rotateArea(float angle);
static Boolean initRotation();

static  Boolean wholePicture;

#define R_INIT		1
#define R_INIT_SCALE	2
#define R_INIT_CLOCK	4
#define R_DIRECTION	5
#define R_FIXED_VALUE	6
#define R_OK		7
#define R_CANCEL	8
#define R_ANGLE_SET	9

/*rotate button callback */
DROTATE void rotateCB(Widget w, XtPointer clientData, XtPointer callData)
{
 static int angle;
 static Widget clockWiseW, angleW;
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case R_INIT_SCALE:
    angleW = w;
    break;
   case R_INIT:
    if (Select.selected && !Select.rectangular)
      manageErrorDialog(getTxtI8N (ONLY_RECTANGULAR_AREAS_MAY_BE_ROTATED_MSG),
	NULL, NULL, getTxtI8N (OK_MSG));
    else if (Select.selected)
      if (initRotation())
       {
	highlightButton(w);
	XtManageChild(rotateDialogW);
       }
      else
       manageErrorDialog(getTxtI8N (NOT_ENOUGH_MEMORY_MSG),
	NULL, NULL, getTxtI8N (OK_MSG));
    else
     {
      ApData.tool = rotateByLine;
      setToolEventHandler(rotateByLine, 0);
      highlightButton(w);
     }
    break;
   case R_INIT_CLOCK:
    clockWiseW = w;
    break;
   case R_OK:
    XtUnmanageChild (rotateDialogW);
    disableUndo ();
    Undo.points[Undo.npoints].x = (short) Select.x;
    Undo.points[Undo.npoints++].y = (short) Select.y;
    Undo.points[Undo.npoints].x = (short) Select.width;
    Undo.points[Undo.npoints++].y = (short) Select.height;
    Undo.tool = moveA;
    ApData.pic_change = True;
    XtVaSetValues(CopyBW, XmNsensitive, False, NULL);
    XtVaSetValues(CutBW, XmNsensitive, True, NULL);
    XtVaSetValues(PasteBW, XmNsensitive, True, NULL);
    highlightButton(NULL);
    break;
   case R_ANGLE_SET:
    {
     float fangle;

     XmScaleGetValue(angleW, &angle);
     fangle = angle/10.0;
     if (!XmToggleButtonGetState(clockWiseW))
      fangle = -fangle;
     if (fabs (fangle) == 90.0)
      rotateArea(fangle);
     else
      rotateArea(fangle*M_PI/180);
    }
    break;
   case R_DIRECTION: rotateCB(w, "9", NULL); break;
   case R_FIXED_VALUE:
    {
     XmString angleStr;
     char *angleS;
     int newV = 0;

     XtVaGetValues(w, XmNlabelString, &angleStr, NULL);
     angleS = xmStringToCstring(angleStr);
     newV = atoi(angleS)*10; /* we have one decimal point */
     XtFree(angleS);
     XmStringFree(angleStr);
     XmScaleSetValue (angleW, newV);
     rotateCB (angleW, "9", NULL);
    }
    break;
   case R_CANCEL:
    {
     int i;

     drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);
     drawSelectionBoundariesWrapped(XtWindow (zoomDrawAreaW), zoomDraw.factorX,
      zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
     MyPutImage (drawAreaWin, xinfo.gc, ApData.img,
      Select.x, Select.y, Select.width, Select.height, Select.x, Select.y, NULL);
     refreshZoomWin (opBounds.x, opBounds.y, opBounds.width, opBounds.height, True);
     if (Select.imgAllocated)
       undoXImage (&Select.img);
     Select.x = Undo.points[0].x;
     Select.y = Undo.points[0].y;
     Select.width = Undo.points[1].x;
     Select.height = Undo.points[1].y;
     Select.npoints = Undo.points[2].x;
     Select.rectangular = Undo.points[2].y;

     Select.img = Undo.img;
     Undo.img = NULL;
     Undo.imgAllocated = False;
     Select.imgAllocated = True;

     if (wholePicture)
      {
       ApData.DAheight = ApData.height = Select.height;
       ApData.DAwidth = ApData.width = Select.width;

       changeTitleCB(NULL, ApData.fname, NULL);
       freeMainImage(&ApData.img);
       ApData.img = createMainImage(ApData.width, ApData.height);
       XtVaSetValues(drawAreaWID, XmNheight, ApData.height, XmNwidth, ApData.width, NULL);
      }
     for (i = 0; i < Select.npoints; i++)
      {
       Select.points[i].x = Undo.points[i + 3].x;
       Select.points[i].y = Undo.points[i + 3].y;
      }
     if (Select.reg)
      {
       XDestroyRegion(Select.reg);
       XSetClipMask(xinfo.dpy, Select.moveGC, None);
       Select.reg = NULL;
       XSetClipOrigin(xinfo.dpy, Select.moveGC, 0, 0);
      }
     LastDrawOnAllWin(False);
     if (!Select.rectangular)
      {
       Select.reg = XPolygonRegion(Select.points, Select.npoints, WindingRule);
       XSetRegion(xinfo.dpy, Select.moveGC, Select.reg);
      }
     disableUndo();
     Undo.tool = rotateA;
     Undo.npoints = 2;
     highlightButton(NULL);
    }
    refreshDrawingArea(True, True); /* just one more refresh to have it really correct */
    break;
  }
}

DROTATE void rotate90(int direction, XImage * src, XImage * dst)
{
 int x, y, w, h;

 w = (int) dst->height - 1;
 h = (int) dst->width - 1;
 if (direction > 0) /* clockwise 90 */
  for (y = 0; y < src->height; y++)
   for (x = 0; x < src->width; x++)
    dst->data[dst->bytes_per_line * x + y] = src->data[src->bytes_per_line * (h - y) + x];
 else
  for (y = 0; y < src->height; y++)
   for (x = 0; x < src->width; x++)
    dst->data[dst->bytes_per_line * (w - x) + y] = src->data[src->bytes_per_line * y + x];
}

static void rotateIt(float angle, XImage * src, XImage * dst)
{
 int i, j, ci, cj, oldi, oldj;
 double sin_a, cos_a, j_cj_cos_a, j_cj_sin_a, diffX, diffY;
 double *i_ci_sin_a, *i_ci_cos_a;

 sin_a = (double) sin ((double) angle);
 cos_a = (double) cos ((double) angle);
 diffX = 1.0 * dst->width;
 diffX -= src->width;
 diffY = 1.0 * dst->height;
 diffY -= src->height;
 diffX = diffX / 2;
 diffY = diffY / 2;
 ci = dst->width / 2;
 cj = dst->height / 2;
 i_ci_sin_a = (double *) calloc (dst->width, sizeof (double));
 i_ci_cos_a = (double *) calloc (dst->width, sizeof (double));
 for (i = 0; i < dst->width; i++)
  {
   i_ci_sin_a[i] = (double) ((i - ci) * sin_a);
   i_ci_cos_a[i] = (double) ((i - ci) * cos_a);
  }

 for (j = 0; j < dst->height; j++)
  {
   j_cj_cos_a = (j - cj) * cos_a;
   j_cj_sin_a = (j - cj) * sin_a;
   for (i = 0; i < dst->width; i++)
    {
     oldj = j_cj_cos_a - i_ci_sin_a[i] + cj - diffY;
     oldi = j_cj_sin_a + i_ci_cos_a[i] + ci - diffX;
     if ((oldi >= 0) && (oldj >= 0) && (oldi < src->width) && (oldj < src->height))
      dst->data[dst->bytes_per_line * j + i] =
       src->data[src->bytes_per_line * oldj + oldi];
    }
  }
 free ((char *) i_ci_sin_a);
 free ((char *) i_ci_cos_a);
}

static Boolean initRotation()
{
 int i;

 if (Select.selected)
  {
   if ((Undo.tool != selectA) && (Undo.tool != selectPolygonA) && (Undo.tool != moveA) &&
       (Undo.tool != mirrorHor) && (Undo.tool != mirrorVer) &&
       (Undo.tool != pasteA) && (Undo.tool != rotateA))
      finalDraw (ApData.pixy);
   disableUndo ();
   Undo.tool = ApData.tool = selectA;

   if (Select.imgAllocated)
    {
     Undo.img = Select.img;
     Select.imgAllocated = False;
     Select.img = NULL;
    }
   else
    Undo.img = MyGetImage(ApData.img, Select.x, Select.y, Select.width, Select.height, ApData.viewMode);

   if (!checkXImage (Undo.img))
    {
     undoXImage (&Undo.img);
     return (False);
    }
   Undo.imgAllocated = True;

   Undo.points[Undo.npoints].x = (short) Select.x;
   Undo.points[Undo.npoints++].y = (short) Select.y;
   Undo.points[Undo.npoints].x = (short) Select.width;
   Undo.points[Undo.npoints++].y = (short) Select.height;

   Undo.points[Undo.npoints].x = (short) Select.npoints;
   Undo.points[Undo.npoints++].y = (short) Select.rectangular;
   for (i = Undo.npoints; i < (Select.npoints + Undo.npoints); i++)
    {
     Undo.points[i].x = Select.points[i - Undo.npoints].x;
     Undo.points[i].y = Select.points[i - Undo.npoints].y;
    }
   Undo.npoints = i + 1;
   /* another copy of data for refresh */
   Undo.points[Undo.npoints].x = (short) Select.x;
   Undo.points[Undo.npoints++].y = (short) Select.y;
   Undo.points[Undo.npoints].x = (short) Select.width;
   Undo.points[Undo.npoints++].y = (short) Select.height;

   wholePicture = ((Select.height == ApData.height) &&
    (Select.y == 0) && (Select.x == 0) && (Select.width == ApData.width));
   return (True);
  }
 else
  {
   manageErrorDialog(getTxtI8N (NO_SELECTION_MADE_MSG), NULL, NULL, getTxtI8N (OK_MSG));
   return(False);
  }
}

/* rotate area algorithm */
static void rotateArea(float angle)
{
 XImage *imgR;
 int w1=0, w2=0, h1=0, h2=0, width, height;

 if (Select.selected)
   {
     Undo.tool = ApData.tool = pasteA;

     imgR = NULL;
     Select.x = Undo.points[0].x;
     Select.y = Undo.points[0].y;
     Select.width = Undo.points[1].x;
     Select.height = Undo.points[1].y;

     if (abs (angle) != 90.0)
       {
	 height = (int) (fabs (cos ((double) angle) * Select.height) +
			 fabs (sin ((double) angle) * Select.width) + 1.5);
	 h1 = (int) (cos ((double) angle) * Select.height + 0.5);
	 h2 = (int) (sin ((double) angle) * Select.width + 0.5);

	 width = (int) (fabs (sin ((double) angle) * Select.height) +
			fabs (cos ((double) angle) * Select.width) + 1.5);
	 w1 = (int) (sin ((double) angle) * Select.height + 0.5);
	 w2 = (int) (cos ((double) angle) * Select.width + 0.5);
       }
     else
       { width = Select.height; height = Select.width; }

     imgR = doXImage (8, width, height);
     if (checkXImage(imgR))
      {
       memset(imgR->data, xinfo.allocCells[ColorMatrix.BckColorInd].pixel,
	imgR->bytes_per_line * imgR->height);

       if (fabs (angle) != 90.0)
	 rotateIt((float) angle, Undo.img, imgR);
       else
	 rotate90(SGN (angle), Undo.img, imgR);

       drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);

       /*refresh background of the selection area */
       MyPutImage(drawAreaWin, xinfo.gc, ApData.img, 
	Undo.points[Undo.npoints - 2].x, Undo.points[Undo.npoints - 2].y,
	Undo.points[Undo.npoints - 2].x, Undo.points[Undo.npoints - 2].y,
	Undo.points[Undo.npoints - 1].x, Undo.points[Undo.npoints - 1].y, NULL);

       opBounds.x = Undo.points[Undo.npoints - 2].x;
       opBounds.y = Undo.points[Undo.npoints - 2].y;
       opBounds.width = Undo.points[Undo.npoints - 1].x;
       opBounds.height = Undo.points[Undo.npoints - 1].y;

       refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, True);
       drawSelectionBoundariesWrapped (XtWindow (zoomDrawAreaW), zoomDraw.factorX,
	zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);

       if (wholePicture && ((width != ApData.width) || (height != ApData.height)))
	{
	 freeMainImage (&ApData.img);
	 ApData.img = createMainImage (width, height);
	 if (!checkXImage(ApData.img))
	  {
	   ApData.img = createMainImage (ApData.width, ApData.height);
	   width = ApData.width;
	   height = ApData.height;
	   if (ApData.img == NULL)
	    {
	     char *msg, *msg1;

	     msg1 = getTxtI8N(RESIZE_PICTURE_FATAL_ERROR_MSG);
	     msg = malloc(strlen (msg1));
	     msg = strcpy(msg, msg1);
	     manageErrorDialog(msg, reallyExitCB, "1", getTxtI8N(PROGRAM_EXIT_MSG));
	     return;
	    }
	  }
	 else
	  {
	   repeats.height = ApData.DAheight = repeats.crop.height = ApData.height = height;
	   repeats.width = ApData.DAwidth = repeats.crop.width = ApData.width = width;
           changeTitleCB(NULL, ApData.fname, NULL);
	   XtVaSetValues(drawAreaWID, XmNheight, ApData.height, XmNwidth, ApData.width, NULL);
 	  }
	}

       Select.width = width;
       Select.height = height;
       Select.npoints = 0;
       if (fabs (angle) != 90.0)
	{
	 int n = Select.npoints;
	 XPoint *p = &Select.points[Select.npoints];

	 if (angle > 0.00)
	  {
	   p[n].x = Select.x + w1;
	   p[n++].y = (unsigned short) Select.y;
	  }
	 else
	  {
	   p[n].x = (unsigned short) Select.x;
	   p[n++].y = Select.y - h2;
	  }
	 if (fabs (angle) > M_PI / 2) /* angle > 90 */
	  {
	   p[n - 1].x -= w2;
	   p[n - 1].y -= h1;
	  }
	 p[n].x = p[n - 1].x + w2;
	 p[n].y = p[n - 1].y + h2;
	 n++;
	 p[n].x = p[n - 1].x - w1;
	 p[n].y = p[n - 1].y + h1;
	 n++;
	 p[n].x = p[n - 1].x - w2;
	 p[n].y = p[n - 1].y - h2;
	 n++;
	 p[n].x = p[n - 1].x + w1;
	 p[n].y = p[n - 1].y - h1;
	 n++;
	 /* hold the offset of the 1st point from x,y */
	 p[n].x = (unsigned short) Select.x;
	 p[n].y = (unsigned short) Select.y;
	 Select.npoints = n;
	 Select.rectangular = False;
	 if (Select.reg)
	  {
	   XDestroyRegion (Select.reg);
	   XSetClipMask (xinfo.dpy, Select.moveGC, None);
	  }
	 Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
	 XSetRegion (xinfo.dpy, Select.moveGC, Select.reg);
	}
       else
	{
	 setBoundaryPtsR(&Select.npoints, Select.x, Select.y, Select.width, Select.height);
	 if (Select.reg)
	  {
	   XDestroyRegion (Select.reg);
	   Select.reg = NULL;
	   XSetClipMask (xinfo.dpy, Select.moveGC, None);
	  }
	 Select.rectangular = True;
	}

       if (Select.imgAllocated)
	 undoXImage (&Select.img);

       Select.imgAllocated = True;
       Select.img = imgR;
       opBounds.x = Undo.points[Undo.npoints - 2].x = (short) Select.x;
       opBounds.y = Undo.points[Undo.npoints - 2].y = (short) Select.y;
       opBounds.width = Undo.points[Undo.npoints - 1].x = (short) Select.width;
       opBounds.height = Undo.points[Undo.npoints - 1].y = (short) Select.height;

       if (wholePicture)
	{
	 refreshDrawingArea (True, True);
	 zoomDraw.newImage = True;
	 zoomMapWindow();
	}
       LastDrawOnAllWin(False);
       imgR = NULL;
       RestoreCursor(XtWindow(mainW));
      }
     else
      { undoXImage(&Undo.img); undoXImage(&imgR); }
   }
 else
  manageErrorDialog(getTxtI8N(NO_SELECTION_MADE_MSG), NULL, NULL, getTxtI8N(OK_MSG));
}

DROTATE void rotatePictureByLine()
{
 float angle;
 int dx, dy;

 dx = Undo.points[0].x - Undo.points[1].x;
 dy = Undo.points[0].y - Undo.points[1].y;
 if ((abs (dx) == abs(dy)) || (dx == 0) || (dy == 0)) return;

 angle = atan2(dx, dy);
 angle = (angle*180)/M_PI;

 if ((angle<-45) && (angle>-135))
  angle+=90.0;
 if ((angle>45) && (angle<135))
  angle-=90.0;
 if ((angle>135) && (angle<225))
  angle-=180.0;
 Select.x = Select.y = 0;
 Select.width = ApData.width;
 Select.height = ApData.height;
 Select.selected = Select.rect_on = True;
 initRotation();
 rotateArea(angle*M_PI/180);
 highlightButton(NULL);
 setButtonIconCB(unselectBW, "unselect.xbm", NULL); 
}
