/* stylus.c
Ver. 1.1b common commands are grouped in epson.c file
Ver. 1.1c tried and failed support for 1440 dpi mode :-(
Ver. 1.1d changed ink control algorithm and dot size for 720 dpi
Ver. 1.1e support for Epson Photo
Ver. 1.1f support for Epson Stylus 3000; Stylus 850 only in old 720 dpi mode
Ver. 1.2  unifed all epson drivers back together, finalized Photo driver:
          - implemented smart switch from light to dark cyan/magenta 
	  - use CMY to make gray on light grays, than gradually switch to black
	  - increased total ink limit to 110000 to avoid ink limit convergence problems
	  - added support for 360 dpi photo printing
	  added support for Stylus Photo 700, Photo EX and Photo 5000
	  - they are all supported only in StylusPhoto mode - 720x720 dpi MICRO_DOT
	    SUPER_MICRO_DOT is only available in software microweaving mode (180x360)
	    print speed is a dog as it is - 18 minutes for full page 720x720 dpi
	    it would be 36 minutes fot 1440 dpi - not acceptable
	  put first 720 dpi mode on all printers, to save a few mouse clicks
	  put first COLORMAP_DITHER mode, to save a few mouse clicks
Ver. 1.3 support for Stylus Photo 750 and 1200, only 720 and 360 dpi modes, one dot size
         without the advertised variable dot size, since it is undocumented 
Ver. 1.4 Support for Photo 870/1270, single dot size (very big!!!), only 720 dpi

TO-DO    TRUE_COLOR printing */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "print.h"
#include "epson.h"

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* THIS DEFINE SHOULD BE ENABLED ONLY FOR PRINTING LINEARIZATION STRIPS */
/* it disables all linearization corrections and ink limit functions    */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*#define LINEARIZATION*/

#define OVERLAP	50

/* the three supported classes of printers; in variable printerClass */
#define CLASS_PRO	0
#define CLASS_800	1
#define CLASS_PHOTO	2
#define CLASS_PHOTO_2	3
#define CLASS_PHOTO_3	4

#define A4LEFT_MARGIN 3
#define A4TOP_MARGIN 3
#define A4RIGHT_MARGIN 4
#define A4BOTTOM_MARGIN 14
#define A4PRINT_AREA 279
#define A4WIDTH 210
#define A4HEIGHT 297
#define A4HEIGHT_720 8420
#define A4TOP_MARGIN_720 78
#define A4BOTTOM_MARGIN_720 8052
#define A4HEIGHT_360 4210
#define A4TOP_MARGIN_360 40
#define A4BOTTOM_MARGIN_360 4026

#define A3LEFT_MARGIN 3
#define A3TOP_MARGIN 3
#define A3RIGHT_MARGIN 4
#define A3BOTTOM_MARGIN 14
#define A3PRINT_AREA 403
#define A3WIDTH 297
#define A3HEIGHT 420

#define A3EXTLEFT_MARGIN 3
#define A3EXTTOP_MARGIN 3
#define A3EXTRIGHT_MARGIN 4
#define A3EXTBOTTOM_MARGIN 14
#define A3EXTPRINT_AREA 464
#define A3EXTWIDTH 330
#define A3EXTHEIGHT 482
#define A3EXTHEIGHT_720 13692
#define A3EXTTOP_MARGIN_720 78
#define A3EXTBOTTOM_MARGIN_720 13324
#define A3EXTHEIGHT_360 6846
#define A3EXTTOP_MARGIN_360 40
#define A3EXTBOTTOM_MARGIN_360 6647

#define A2LEFT_MARGIN_1520 3
#define A2LEFT_MARGIN_3000 5
#define A2TOP_MARGIN 3
#define A2RIGHT_MARGIN_1520 72
#define A2RIGHT_MARGIN_3000 5
#define A2BOTTOM_MARGIN 14
#define A2PRINT_AREA 577
#define A2WIDTH 420
#define A2HEIGHT 594
#define A2HEIGHT_720 16838
#define A2TOP_MARGIN_720 78
#define A2BOTTOM_MARGIN_720 16470
#define A2HEIGHT_360 8419
#define A2TOP_MARGIN_360 40
#define A2BOTTOM_MARGIN_360 8235

static int printerClass, inks6=FALSE;
static unsigned char *CMYKBlck;
static unsigned char *outBlock;
static int outBlockSize;
static unsigned char *dithMK, *dithM_C, *dithM_M, *dithM_Y, *dithM_ML, *dithM_CL;
static unsigned int blackDots[512], magentaDots[512], cyanDots[512], yellowDots[512],
  magentaLightDots[512], cyanLightDots[512];
static unsigned char convTbl[256][4], convTblPhoto[256][6];
static unsigned char dithColOffsX[6], dithColOffsY[6];
static unsigned char *RLLBuf;
static unsigned char mag2DarkM[256], cyan2DarkC[256], cyan2LightC[256], mag2LightM[256];

static struct
 {
  unsigned char *buf;
  int bufSize;
  FILE *fout;
  char *fname;
  int modeType;
  int currBandType;
  int leftM, topM, width, height;
  int Hdpi, Vdpi, HdpiMode, VdpiMode;
  int vertPos;
  Boolean firstLine;
  unsigned char unit;
  int inkNo;
 }
Epson;


/* print data in epson stylus format to the file prnFile */
static int print2fileColor(unsigned char *lnP, int width, int fullWidth, int rowNo,
 int vertShift, int leftMargin, FILE * prnFile)
{
 int c, j, i;
 unsigned char *p, *dataP, colSeq[6] = {2, 1, 0, 3, 4, 5};

 p = Epson.buf;
 for (i = 0; i < Epson.inkNo; i++)
  {
   c = colSeq[i];
   /* check if line is made of 0's */
   dataP = lnP + width * c;
   j = 0;
   while ((j < width) && (!(*dataP++))) j++;
   if (j < width)
    {
     dataP = lnP + width*c;
     if (leftMargin > 0)
      p = SetRelXPos(leftMargin, p);
     if (printerClass!=CLASS_PHOTO_3)
      {
       if (Epson.inkNo == 6)
	p = SetColorCCMMYK(c, p);
       else
	p = SetColorCMYK(c, p);
       p = PrintRasterGraphics(COMPRESSION, Epson.HdpiMode, Epson.VdpiMode, rowNo, fullWidth, p);
       /* compress line of data */
       p = compressLine(dataP, p, RLLBuf, (fullWidth + 7) / 8);
       *p++ = 0xd; /* CR */
      }
     else
      {
       p = TransferRasterImage(c, COMPRESSION, 1 /* BPP */, (fullWidth + 7) / 8, rowNo, p);
       /* compress line of data */
/*     memcpy(p, dataP, (fullWidth + 7) / 8);
       p+=((fullWidth + 7) / 8);*/
       p = compressLine(dataP, p, RLLBuf, (fullWidth + 7) / 8);
       *p++ = 0xd; /* CR */
      }
    }
  }
 if (vertShift > 0)
  p = SetRelYPos(vertShift, p);
 fwrite (Epson.buf, 1, p - Epson.buf, prnFile);
 p = Epson.buf;
 return(ferror(prnFile) ? PRT_IO_ERROR : PRT_OK);
}

/*calculate the conversion table from RGB to CMYK */
static void RGB2CMYK(unsigned char *cmap)
{
 unsigned char *colorP = cmap, *p;
 int i;
 float blackProb;
 unsigned long cyan, magenta, yellow, black, s, inkLimit, f, f_new;

 for (i = 0; i < 256; i++)
  {
   p = convTbl[i];
   if (printerClass==CLASS_800)
    inkLimit = (Epson.Hdpi==720) ? 95000 : 75000;
   else /* CLASS_PRO */
    inkLimit = (Epson.Hdpi==720) ? 0x10000 : 105000;

   cyan = (unsigned char) 255 - *colorP++;
   magenta = (unsigned char) 255 - *colorP++;
   yellow = (unsigned char) 255 - *colorP++;

   black = cyan > magenta ? magenta : cyan;
   black = black > yellow ? yellow : black;
   if (black != 255)
    {
     cyan = (cyan - black) * 255 / (255 - black);
     magenta = (magenta - black) * 255 / (255 - black);
     yellow = (yellow - black) * 255 / (255 - black);
     blackProb = (blackDots[black] * 1.0) / blackDots[255];
     s = cyanDots[cyan] + magentaDots[magenta] + yellowDots[yellow]
       + blackDots[black];
     s -= blackProb * cyanDots[cyan];
     s -= blackProb * magentaDots[magenta];
     s -= blackProb * yellowDots[yellow];
     if (s > inkLimit)
      {
       int ap, dotDiff = 0, interval, maxDiff, oldSgn = 1, newSgn;

       f = inkLimit;
       blackProb = (blackDots[(black * f) / s] * 1.0) / blackDots[255];
       f_new = cyanDots[(cyan * f) / s] + magentaDots[(magenta * f) / s]
	 + yellowDots[(yellow * f) / s] + blackDots[(black * f) / s] -
	 blackProb * (cyanDots[(cyan * f) / s] + magentaDots[(magenta * f) / s]
		      + yellowDots[(yellow * f) / s]);
       interval = (int) (f - f_new);
       ap = (int) f - interval;
       maxDiff = f / 100;
       do
	{
	 ap += oldSgn * interval;
	 blackProb = (blackDots[(black * ap) / s] * 1.0) / blackDots[255];
	 f_new = cyanDots[(cyan * ap) / s] + magentaDots[(magenta * ap) / s]
	   + yellowDots[(yellow * ap) / s] + blackDots[(black * ap) / s] -
	   blackProb * (cyanDots[(cyan * ap) / s] + magentaDots[(magenta * ap) / s]
			+ yellowDots[(yellow * ap) / s]);
	 dotDiff = (int) (f - f_new);
	 newSgn = SGN (dotDiff);
	 if (oldSgn != newSgn)
	   interval /= 2;
	 oldSgn = newSgn;
	}
       while ((abs (dotDiff) > maxDiff) && (interval > 1));
       cyan = (cyan * ap) / s;
       cyan = (cyan > 255) ? 255 : cyan;
       magenta = (magenta * ap) / s;
       magenta = (magenta > 255) ? 255 : magenta;
       yellow = (yellow * ap) / s;
       yellow = (yellow > 255) ? 255 : yellow;
       black = (black * ap) / s;
       black = (black > 255) ? 255 : black;
      }
    }
   else
    cyan = magenta = yellow = 0;
   *p++ = (unsigned char) cyan;
   *p++ = (unsigned char) magenta;
   *p++ = (unsigned char) yellow;
   *p = (unsigned char) black;
  }
}

/*calculate the conversion table from RGB to C CL M ML Y K */
static void RGB2CCMMYK(unsigned char *cmap)
{
 unsigned char *colorP = cmap, *p;
 int i;
 float blackProb, cyanDProb, magentaDProb;
 unsigned long cyan, cyanD, cyanL, magenta, magentaD, magentaL, yellow, black, s, inkLimit, f, f_new;

 for (i = 0; i < 256; i++)
  {
   p = convTblPhoto[i];
   inkLimit = (Epson.Hdpi==720) ? 110000 : 75000;
   if (printerClass == CLASS_PHOTO_3)
    inkLimit = 75000;
   cyan = (unsigned char) 255 - *colorP++;
   magenta = (unsigned char) 255 - *colorP++;
   yellow = (unsigned char) 255 - *colorP++;
#ifdef LINEARIZATION
/* set to "=cyan" the one that you want to print, others should be 0 */
   cyanD=0;
   magentaD=cyan;
   yellow=0;
   black=0;
   cyanL=0;
   magentaL=0;
#else
   black = cyan > magenta ? magenta : cyan;
   black = black > yellow ? yellow : black;
   if (black != 255)
    {
     // make a colored black if less than 100, add black to full level until 200 
     black = (black>100) ? ((black>200) ? black : (black-100)*2) : 0;
     cyan = (cyan - black) * 255 / (255 - black);
     magenta = (magenta - black) * 255 / (255 - black);
     yellow = (yellow - black) * 255 / (255 - black);
     cyanD = cyan2DarkC[cyan];
     cyanL = cyan2LightC[cyan];
     magentaD = mag2DarkM[magenta];
     magentaL = mag2LightM[magenta];
     blackProb = (blackDots[black]*1.0)/blackDots[255];
     cyanDProb = (cyanDots[cyanD]*1.0)/cyanDots[255];
     magentaDProb = (magentaDots[magentaD]*1.0)/magentaDots[255];
     s = cyanDots[cyanD] + cyanLightDots[cyanL] + magentaDots[magentaD] +
         magentaLightDots[magentaL] + yellowDots[yellow] + blackDots[black] -
         blackProb*(cyanDots[cyanD] + cyanLightDots[cyanL] + magentaDots[magentaD] +
	  magentaLightDots[magentaL] + yellowDots[yellow]) -
        (1.0-blackProb)*(cyanLightDots[cyanL]*cyanDProb + magentaLightDots[magentaL]*magentaDProb);
     if (s > inkLimit)
      {
       int ap, dotDiff = 0, interval, maxDiff, oldSgn = 1, newSgn;

       f = inkLimit;
       blackProb = (blackDots[(black*f)/s]*1.0)/blackDots[255];
       cyanDProb = (cyanDots[(cyanD*f)/s]*1.0)/cyanDots[255];
       magentaDProb = (magentaDots[(magentaD*f)/s]*1.0)/magentaDots[255];
       f_new = cyanDots[(cyanD*f)/s] + cyanLightDots[(cyanL*f)/s] + magentaDots[(magentaD*f)/s] +
	 magentaLightDots[(magentaL*f)/s] + yellowDots[(yellow*f)/s] + blackDots[(black*f)/s] -
	 blackProb*(cyanDots[(cyanD*f)/s] + cyanLightDots[(cyanL*f)/s] + magentaDots[(magentaD*f)/s] +
	  magentaLightDots[(magentaL*f)/s] + yellowDots[(yellow*f)/s]) -
	 (1.0-blackProb)*(cyanLightDots[(cyanL*f)/s]*cyanDProb + magentaLightDots[(magentaL*f)/s]*magentaDProb);
       interval = (int) (f - f_new);
       ap = (int) f - interval;
       maxDiff = f / 100;
       do
	{
	 ap += oldSgn * interval;
	 blackProb = (blackDots[(black*ap)/s]*1.0)/blackDots[255];
	 cyanDProb = (cyanDots[(cyanD*ap)/s]*1.0)/cyanDots[255];
	 magentaDProb = (magentaDots[(magentaD*ap)/s]*1.0)/magentaDots[255];
	 f_new = cyanDots[(cyanD*ap)/s] + cyanLightDots[(cyanL*ap)/s] + magentaDots[(magentaD*ap)/s] +
	   magentaLightDots[(magentaL*ap)/s] + yellowDots[(yellow*ap)/s] + blackDots[(black*ap)/s] -
	   blackProb*(cyanDots[(cyanD*ap)/s] + cyanLightDots[(cyanL*ap)/s] + magentaDots[(magentaD*ap)/s] +
	    magentaLightDots[(magentaL*ap)/s] + yellowDots[(yellow*ap)/s]) -
	   (1.0-blackProb)*(cyanLightDots[(cyanL*ap)/s]*cyanDProb + magentaLightDots[(magentaL*ap)/s]*magentaDProb);
	 dotDiff = (int) (f - f_new);
	 newSgn = SGN (dotDiff);
	 if (oldSgn != newSgn)
	  interval /= 2;
	 oldSgn = newSgn;
	}
       while ((abs (dotDiff) > maxDiff) && (interval > 1));
       cyanD = (cyanD*ap)/s;
       cyanD = (cyanD > 255) ? 255 : cyanD;
       cyanL = (cyanL*ap)/s;
       cyanL = (cyanL > 255) ? 255 : cyanL;
       magentaD = (magentaD*ap)/s;
       magentaD = (magentaD > 255) ? 255 : magentaD;
       magentaL = (magentaL*ap)/s;
       magentaL = (magentaL > 255) ? 255 : magentaL;
       yellow = (yellow*ap)/s;
       yellow = (yellow > 255) ? 255 : yellow;
       black = (black*ap)/s;
       black = (black > 255) ? 255 : black;
      }
    }
   else
    cyan = cyanD = magenta = magentaD = yellow = cyanL = magentaL = 0;
#endif
   *p++ = (unsigned char) cyanD;
   *p++ = (unsigned char) magentaD;
   *p++ = (unsigned char) yellow;
   *p++ = (unsigned char) black;
   *p++ = (unsigned char) cyanL;
   *p = (unsigned char) magentaL;
  }
}

/* calculate the conversion table from RGB to C CL M ML Y K */
/* version for second generation photo: 750 & 1200 */
/* without the total ink limit blurb, since dot size is adequate for the first time! */
static void RGB2CCMMYK2(unsigned char *cmap)
{
 unsigned char *colorP = cmap, *p;
 int i;
 unsigned long cyan, cyanD, cyanL, magenta, magentaD, magentaL, yellow, black;

 for (i = 0; i < 256; i++)
  {
   p = convTblPhoto[i];
   cyan = (unsigned char) 255 - *colorP++;
   magenta = (unsigned char) 255 - *colorP++;
   yellow = (unsigned char) 255 - *colorP++;
#ifdef LINEARIZATION
/* set to "=cyan" the one that you want to print, others should be 0 */
   cyanD=0;
   magentaD=0;
   yellow=0;
   black=cyan;
   cyanL=0;
   magentaL=0;
#else
   black = cyan > magenta ? magenta : cyan;
   black = black > yellow ? yellow : black;
   if (black != 255)
    {
     // make a colored black if less than 170, then gradually add black to maximum
     black = (black>170) ? (black-170)*3 : 0;
     cyan = (cyan - black) * 255 / (255 - black);
     magenta = (magenta - black) * 255 / (255 - black);
     yellow = (yellow - black) * 255 / (255 - black);
     cyanD = cyan2DarkC[cyan];
     cyanL = cyan2LightC[cyan];
     magentaD = mag2DarkM[magenta];
     magentaL = mag2LightM[magenta];
    }
   else
    cyan = cyanD = magenta = magentaD = yellow = cyanL = magentaL = 0;
#endif
   *p++ = (unsigned char) cyanD;
   *p++ = (unsigned char) magentaD;
   *p++ = (unsigned char) yellow;
   *p++ = (unsigned char) black;
   *p++ = (unsigned char) cyanL;
   *p = (unsigned char) magentaL;
  }
}

/* this is wrong; need to sync it with normal palette based CMYK, with buffering of old calculations */
static void TrueColor2CMYK (unsigned char *img, unsigned char *cmykBuf, int width,
 unsigned long gm764, unsigned long gm510, unsigned long gm255)
{
 int i;
 unsigned char *data = img, *c, *y, *m, *k;
 unsigned long cyan, magenta, yellow, black, s, f, shiftS;
 unsigned long diff764_255 = gm764 - gm255;

 c = cmykBuf;
 m = cmykBuf + width;
 y = cmykBuf + 2 * width;
 k = cmykBuf + 3 * width;
 for (i = 0; i < width; i++)
  {
   cyan = (unsigned char) 255 - *data++;
   magenta = (unsigned char) 255 - *data++;
   yellow = (unsigned char) 255 - *data++;
   black = cyan > magenta ? magenta : cyan;
   black = black > yellow ? yellow : black;
   if (black != 255)
    {
     cyan = (cyan - black) * 255 / (255 - black);
     magenta = (magenta - black) * 255 / (255 - black);
     yellow = (yellow - black) * 255 / (255 - black);
     s = cyanDots[cyan] + magentaDots[magenta] + yellowDots[yellow]
       + blackDots[black];
     if (s > 0x10000)
      {
       f = ((diff764_255 * (s - 0x10000)) >> 10) + (gm255 << 6);
       shiftS = s << 6;
       *c = (cyan * f) / shiftS;
       *m = (magenta * f) / shiftS;
       *y = (yellow * f) / shiftS;
       *k = (black * f) / shiftS;
      }
    }
   else
     { *c = *m = *y = 0; *k = 255; }
   c++; m++; y++; k++;
  }
}

/* this is wrong; need to sync it with normal palette based CMYK, with buffering of old calculations */
/* print color picture using Wienna Dusan dithering algorithm */
static int processImageBlockTrueColor(char *img, int bpl, int width, int firstRow, int rowNo, FILE *prnFile)
{
 int result = PRT_OK, y = 0, x, col = 0, outWidth;
 unsigned char *data, *outB, *dithY, *dithMArr[4];
 unsigned long gm764=0, gm510=0, gm255=0;

 outWidth = (width + 7) / 8;
 switch (Epson.Hdpi)
  {
   case 720:
     gm764 = 190710;		/*0x30000 * 0.97 */
     gm510 = 127140;		/*0x20000 * 0.97 */
     break;
   case 360:
     gm764 = 0x30000;		/*0x30000 * 1 */
     gm510 = 0x20000;
     break;
  }
 gm255 = 0x10000;
 dithMArr[0] = dithM_C;
 dithMArr[1] = dithM_M;
 dithMArr[2] = dithM_Y;
 dithMArr[3] = dithMK;
 while ((result == PRT_OK) && (y < rowNo))
  {
   TrueColor2CMYK((unsigned char *) &img[y * bpl], CMYKBlck, width, gm764, gm510, gm255);
   /*clear output buffer */
   memset (outBlock, 0, outWidth * 4);
   for (col = 0; col < 4; col++)
    {
     outB = outBlock + col * outWidth;
     data = CMYKBlck + col * width;
     dithY = dithMArr[col] + (((y + firstRow + dithColOffsY[col]) & 255) << 8);
     for (x = 0; x < width; x++)
      {
       if (*data != 0)
	if (((int) *data + (int) dithY[(dithColOffsX[col] + x) & 0xff]) > 255)
	 *outB |= (128 >> (x & 7));
       if ((x & 7) == 7)
	outB++;
       data++;
      }
    }
   result = print2fileColor(outBlock, outWidth, width, 1, 1, Epson.leftM, prnFile);
   y++;
  }
 return (result);
}

static int processImageBlockCmap(char *img, int bpl, int width, int firstRow, int rowNo,
 unsigned char *cmap, FILE * prnFile)
{
 int result = PRT_OK, y = 0, x, outWidth;
 unsigned char *data, *outC, *outM, *outY, *outK, *outCL, *outML, *dithCY,
  *dithMY, *dithKY, *dithYY, *dithMLY, *dithCLY, *imgP;

 if (inks6)
  {
   memset(convTblPhoto, 0, 256 * 6);
   if ( (printerClass == CLASS_PHOTO_2)/* || (printerClass == CLASS_PHOTO_3)*/ )
    RGB2CCMMYK2(cmap);
   else
    RGB2CCMMYK(cmap);
  }
 else
  { memset (convTbl, 0, 256 * 4); RGB2CMYK(cmap); }

 outWidth = (width+7)/8;
 while ((result == PRT_OK) && (y < rowNo))
  {
   memset (outBlock, 0, outWidth * Epson.inkNo); /*clear output buffer */
   outC = outBlock;
   outM = outC + outWidth;
   outY = outM + outWidth;
   outK = outY + outWidth;
   if (inks6)
    { outCL = outK + outWidth; outML = outCL + outWidth; }
   else
    outCL = outML = NULL;
   imgP = (unsigned char *) &img[bpl * y];

   dithCY = &dithM_C[((y + firstRow + dithColOffsY[0]) & 255) << 8];
   dithMY = &dithM_M[((y + firstRow + dithColOffsY[1]) & 255) << 8];
   dithYY = &dithM_Y[((y + firstRow + dithColOffsY[2]) & 255) << 8];
   dithKY = &dithMK[((y + firstRow + dithColOffsY[3]) & 255) << 8];
   if (inks6)
    {
     dithCLY = &dithM_CL[((y + firstRow + dithColOffsY[4]) & 255) << 8];
     dithMLY = &dithM_ML[((y + firstRow + dithColOffsY[5]) & 255) << 8];
    }
   else
     dithCLY = dithMLY = NULL;
   for (x = 0; x < width; x++)
    {
     data = (inks6) ? convTblPhoto[*imgP] : convTbl[*imgP];

     data += BLACKOFFSET;
     if (((int) *data + (int) dithKY[(dithColOffsX[3] + x) & 0xff]) > 255)
      /* if black dot down, don't use the other 3 colors */
      *outK |= (128 >> (x & 7));
     else
      {
       data--;
       if (*data)
	if (((int) *data + (int) dithYY[(dithColOffsX[2] + x) & 0xff]) > 255)
	 *outY |= (128 >> (x & 7));
       data--;
       if (*data)
	if (((int) *data + (int) dithMY[(dithColOffsX[1] + x) & 0xff]) > 255)
         *outM |= (128 >> (x & 7));
       data--;
       if (*data)
	if (((int) *data + (int) dithCY[(dithColOffsX[0] + x) & 0xff]) > 255)
	 *outC |= (128 >> (x & 7));
       if (inks6)
	{
	 data += CYANLIGHTOFFSET;
	 if (*data)
	  if (!(*outC & (128 >> (x & 7)))) /* put cyan light only if not dark cyan */
	   if (((int) *data + (int) dithCLY[(dithColOffsX[4] + x) & 0xff]) > 255)
	    *outCL |= (128 >> (x & 7));
	 data++;
	 if (*data)
	  if (!(*outM & (128 >> (x & 7)))) /* put cyan light only if not dark magenta */
	   if (((int) *data + (int) dithMLY[(dithColOffsX[5] + x) & 0xff]) > 255)
	    *outML |= (128 >> (x & 7));
	}
      }
     imgP++;
     if ((x & 7) == 7)
      {
       outC++; outM++; outY++; outK++;
       if (inks6)
	{ outCL++; outML++; }
      }
    }
   result = print2fileColor(outBlock, outWidth, width, 1, 1, Epson.leftM, prnFile);
   y++;
  }
 return(result);
}

static int processImageBlockBW(char *img, int bpl, int width8, int firstRow, int rowNo,
 int leftMargin, FILE * prnFile)
{
 int ln;
 unsigned char *p, *d;
 int width, newWidth;

 if (width8 > Epson.width)
  { width = (Epson.width+7)/8; newWidth = Epson.width; }
 else
  { width = (width8+7)/8; newWidth = width8; }

 for (ln = 0; ln < rowNo; ln++)
  {
   p = Epson.buf;
   if (leftMargin > 0)
     p = SetRelXPos (leftMargin, p);
   if (printerClass!=CLASS_PHOTO_3)
    {
     if (inks6)
      p = SetColorCCMMYK(BLACKOFFSET, p);
     p = PrintRasterGraphics(COMPRESSION, Epson.HdpiMode, Epson.VdpiMode, 1, newWidth, p);
     d = (unsigned char *) &img[ln * bpl];
     /* compress line of data */
     p = compressLine(d, p, RLLBuf, width);
     *p++ = 0xd; /* CR */
    }
   else
    {
     p = TransferRasterImage(BLACKOFFSET, COMPRESSION, 1 /* BPP */, width, 1, p);
     d = (unsigned char *) &img[ln * bpl];
     /* compress line of data */
     p = compressLine(d, p, RLLBuf, width);
     *p++ = 0xd; /* CR */
    }
   p = SetRelYPos(1, p);
   fwrite(Epson.buf, sizeof (char), p - Epson.buf, prnFile);
   if (ferror (prnFile))
     return(PRT_IO_ERROR);
  }
 return(PRT_OK);
}

static int initPrint(int printType, int complexImage, int xpixels,
 int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
  unsigned char *prnBufPtr;
  float *sampleK, *sampleC, *sampleM, *sampleY, *sampleML, *sampleCL;
  /* measured luminance values of equal distribution dither matrix */

  /* STYLUS PRO */
  float samplePro720[17] =
  {93.24, 78.65, 62.64, 49.27, 38.57, 31.18, 26.28, 22.80, 20.60,
   19.37, 18.77, 18.30, 18.07, 17.64, 17.55, 17.34, 17.04};
  float samplePro360[17] =
  {95.12, 88.04, 79.86, 71.98, 64.53, 57.62, 51.80, 46.36, 40.30,
   36.93, 33.52, 31.27, 29.05, 27.72, 26.45, 25.57, 24.52};

  /* STYLUS 800, 850, 600, 1520 */
  float sample720[17] =		/* micro dot values */
  {94.60, 78.02, 63.53, 50.92, 40.66, 33.08, 27.70, 23.02, 19.98,
   18.00, 17.50, 17.27, 17.20, 16.13, 15.63, 15.13, 14.92};
  float cyan720[17] =		/* micro dot values */
  {94.45, 87.00, 80.92, 75.78, 71.74, 68.29, 65.57, 63.01, 60.84,
   58.71, 56.89, 55.16, 53.55, 52.29, 51.30, 50.35, 49.64};
  float magenta720[17] =	/* micro dot values */
  {94.45, 86.17, 79.43, 73.84, 69.34, 65.72, 62.80, 60.36, 58.26,
   56.53, 55.18, 53.82, 52.85, 51.90, 51.25, 50.60, 49.93};
  float yellow720[17] =		/* micro dot values */
  {94.57, 92.41, 90.81, 89.54, 88.54, 87.56, 86.82, 86.17, 85.61,
   85.12, 84.48, 83.86, 83.36, 82.82, 82.52, 82.07, 81.51};

  float sample360[17] =
  {95.90, 86.46, 76.30, 66.13, 57.25, 49.49, 42.18, 34.22, 28.27,
   24.21, 20.30, 19.51, 18.05, 17.67, 17.29, 17.05, 17.01};
  float cyan360[17] =
  {95.90, 90.91, 85.56, 80.78, 76.07, 72.25, 68.43, 64.20, 60.75,
   58.99, 55.80, 54.07, 52.29, 50.97, 50.29, 48.98, 48.46};
  float magenta360[17] =
  {95.90, 91.80, 87.55, 82.99, 79.28, 75.31, 72.63, 70.01, 67.65,
   65.72, 64.45, 62.94, 62.05, 61.38, 60.83, 60.37, 60.23};
  float yellow360[17] =
  {95.90, 95.41, 94.72, 94.16, 93.48, 92.91, 92.45, 92.13, 91.77,
   91.35, 90.98, 90.60, 90.30, 90.05, 89.80, 89.79, 89.63};

  /* STYLUS PHOTO */
  float photoBlack720[17] =
  {94.33, 79.74, 65.42, 53.10, 42.96, 35.46, 29.12, 24.46, 20.82,
   18.66, 17.51, 16.47, 15.91, 15.39, 15.32, 15.31, 15.26};
  float photoCyan720[17] =
  {94.33, 87.37, 81.10, 76.08, 71.69, 68.30, 65.36, 62.89, 60.63,
   58.40, 56.66, 55.04, 53.56, 52.36, 51.34, 50.45, 49.62};
  float photoCyanLight720[17] =
  {94.33, 91.28, 88.46, 86.03, 83.81, 81.77, 79.94, 78.17, 76.64,
   75.07, 73.63, 72.33, 71.11, 70.01, 69.01, 68.09, 67.23};
  float photoMagenta720[17] =
  {94.33, 87.17, 80.19, 74.50, 69.62, 65.87, 62.91, 60.47, 58.30,
   56.49, 55.11, 53.93, 52.90, 52.07, 51.31, 50.64, 50.08};
  float photoMagentaLight720[17] =
  {94.33, 91.21, 88.07, 85.20, 82.67, 80.31, 78.11, 76.08, 74.26,
   72.39, 70.75, 69.20, 67.93, 66.72, 65.60, 64.68, 63.81};
  float photoYellow720[17] =
  {94.33, 92.70, 91.18, 89.96, 88.81, 87.89, 87.18, 86.53, 85.90,
   85.29, 84.62, 84.24, 83.65, 83.15, 82.77, 82.23, 81.88};

  /* STYLUS PHOTO 750-1200 edition */
  float photo2Black720[17] =
  {92.75, 80.32, 69.07, 59.01, 49.82, 43.41, 37.22, 31.99, 28.23,
   25.88, 23.66, 21.53, 20.36, 19.16, 18.20, 17.63, 17.09};
  float photo2Cyan720[17] =
  {92.76, 87.34, 82.72, 78.50, 74.74, 71.95, 69.63, 67.41, 65.26,
   64.04, 62.23, 60.09, 58.58, 57.56, 56.41, 55.03, 53.46};
  float photo2CyanLight720[17] =
  {92.50, 91.00, 89.03, 86.43, 84.81, 82.71, 81.43, 80.17, 79.11,
   77.73, 76.69, 76.11, 74.35, 73.47, 72.56, 71.53, 70.59};
  float photo2Magenta720[17] =
  {92.47, 86.98, 81.67, 77.27, 73.11, 70.18, 66.85, 64.27, 62.21,
   60.65, 59.11, 57.37, 56.15, 55.15, 54.24, 53.17, 52.28};
  float photo2MagentaLight720[17] =
  {92.83, 89.58, 87.01, 84.66, 82.82, 80.81, 79.69, 77.58, 76.23,
   74.79, 73.16, 71.97, 70.16, 68.89, 67.78, 66.66, 65.54};
  float photo2Yellow720[17] =
  {93.06, 91.50, 90.44, 89.33, 88.32, 87.53, 86.60, 85.90, 85.30,
   84.80, 84.60, 84.10, 83.70, 83.20, 82.80, 82.40, 82.00};
/* handicrafted the measurments to fix errors (equal values, non-monotonic, uneven jumps),
  until I got visually acceptable linearity; I guess diff of dL min/max 11 is too small...
  {93.06, 91.01, 90.44, 89.33, 88.32, 87.53, 86.07, 86.07, 85.12,
   84.67, 84.60, 84.33, 83.81, 82.84, 82.91, 82.08, 82.56};
*/

  /* STYLUS PHOTO 870-1270 edition, temporary version single dot size */
  float photo3Black720[17] =
  {94.37, 70.97, 54.78, 42.22, 29.88, 24.81, 20.20, 18.37, 17.06, 
   16.46, 16.34, 16.29, 15.99, 15.94, 15.90, 15.80, 15.77};
  float photo3Cyan720[17] =
  {94.46, 82.85, 74.75, 68.01, 62.77, 58.80, 55.80, 53.47, 51.86,
   50.47, 49.53, 48.83, 48.33, 47.96, 47.77, 47.25, 47.24};
  float photo3CyanLight720[17] =
  {94.50, 89.02, 84.67, 80.87, 77.77, 74.99, 72.81, 70.90, 69.53,
   68.14, 67.13, 66.16, 65.29, 64.56, 63.91, 63.35, 62.81};
  float photo3Magenta720[17] =
  {94.51, 82.19, 73.14, 66.35, 60.83, 57.67, 55.49, 53.65, 52.63,
   51.68, 50.93, 50.45, 49.87, 49.54, 49.14, 48.84, 48.60};
  float photo3MagentaLight720[17] =
  {94.39, 87.30, 81.64, 76.76, 72.67, 69.22, 66.67, 64.56, 63.04,
   61.71, 60.60, 59.83, 59.14, 58.51, 58.01, 57.65, 57.34};
  float photo3Yellow720[17] =
  {94.59, 92.30, 90.50, 89.25, 88.12, 87.00, 86.29, 85.53, 84.87,
   84.21, 83.81, 83.37, 82.94, 82.63, 82.36, 82.09, 81.94};

  Epson.modeType = Epson.currBandType = printType;
  Epson.width = xpixels;
  Epson.height = ypixels;
  Epson.Hdpi = xdpi / 1000;
  Epson.Vdpi = ydpi / 1000;
  Epson.vertPos = 0;
  Epson.inkNo = (inks6) ? 6 : 4;

  sampleK = sampleC = sampleM = sampleY = sampleCL = sampleML = NULL;

  Epson.HdpiMode = Epson.VdpiMode = (Epson.Hdpi==720) ? 5 : 10; /* second case 360 dpi */
  Epson.fname = tempnam ("/tmp", "epsonX");

  if ((Epson.fout = fopen (Epson.fname, "w")) == NULL)
   return (PRT_IO_ERROR);

  /* allocate memory of printer buffer = width*inkNo + bytes for commands; round it to multiple of 512 */
  Epson.bufSize = Epson.inkNo * ((Epson.width + 7) / 8) + 256;
  if (Epson.bufSize % 512)
   Epson.bufSize += 512 - Epson.bufSize % 512;

  /* take 1024 bytes more for compression which makes data bigger instead of smaller */
  Epson.buf = (unsigned char *) calloc (Epson.bufSize+1024, sizeof (char));
  if (Epson.buf == NULL)
   return (PRT_NO_MEMORY);
  RLLBuf = (unsigned char *) calloc (Epson.bufSize / 2+1024, sizeof (char));
  if (RLLBuf == NULL)
   return (PRT_NO_MEMORY);

  prnBufPtr = Epson.buf;
  if (printerClass != CLASS_PHOTO_3)
   {
    prnBufPtr = ResetPrn(prnBufPtr);
    prnBufPtr = ResetPrn(prnBufPtr);
   }
  else /* send all these weird commands hoping that the 720 dpi variable dot mode will then work */
   {
    /* start packet mode exit command */
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x1B;
    *prnBufPtr++=0x01;
    *prnBufPtr++='@';
    *prnBufPtr++='E';
    *prnBufPtr++='J';
    *prnBufPtr++='L';
    *prnBufPtr++=' ';
    *prnBufPtr++='1';
    *prnBufPtr++='2';
    *prnBufPtr++='8';
    *prnBufPtr++='4';
    *prnBufPtr++='.';
    *prnBufPtr++='4';
    *prnBufPtr++=0x0A;
    *prnBufPtr++='@';
    *prnBufPtr++='E';
    *prnBufPtr++='J';
    *prnBufPtr++='L';
    *prnBufPtr++=' ';
    *prnBufPtr++=' ';
    *prnBufPtr++=' ';
    *prnBufPtr++=' ';
    *prnBufPtr++=' ';
    *prnBufPtr++=0x0A;
    /* end packet mode exit command */
    prnBufPtr = ResetPrn(prnBufPtr);
    prnBufPtr = ResetPrn(prnBufPtr);
    /* start enter remote mode command */
    *prnBufPtr++=0x1B;
    *prnBufPtr++='(';
    *prnBufPtr++='R';
    *prnBufPtr++=0x08;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    *prnBufPtr++='R';
    *prnBufPtr++='E';
    *prnBufPtr++='M';
    *prnBufPtr++='O';
    *prnBufPtr++='T';
    *prnBufPtr++='E';
    *prnBufPtr++='1';
    /* remote commands follow */

    *prnBufPtr++='P'; /* undocumented command */
    *prnBufPtr++='M';
    *prnBufPtr++=0x02;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    
    *prnBufPtr++='S'; /* set paper feed sequence */
    *prnBufPtr++='N';
    *prnBufPtr++=0x03;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x07; /* matte paper */

    *prnBufPtr++='F'; /* select left margin */
    *prnBufPtr++='P';
    *prnBufPtr++=0x03;
    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00;

    *prnBufPtr++=0x00;
    *prnBufPtr++=0x00; /* normal left margin, 0.3mm */
    /* last two on 0xB0 and 0xFF set margin on 0 */

    *prnBufPtr++=0x1B;
    *prnBufPtr++=0;
    *prnBufPtr++=0;
    *prnBufPtr++=0;
    /* end of weird commands */
   }

  prnBufPtr = SetGraphicsMode(prnBufPtr);

  if (Epson.Hdpi==720)
   switch (printerClass)
    {
     case CLASS_PRO:
      prnBufPtr = SetDotSize(MICRO_DOT_SIZE, prnBufPtr);
      sampleK = samplePro720;
      break;
     case CLASS_PHOTO:
      /* it should be SINGLE_DOT_SIZE for 720 dpi, but we use 1440 dot */
      prnBufPtr = SetDotSize(MICRO_DOT_SIZE, prnBufPtr);
      prnBufPtr = SelectPrintingSpeed(SPEED_20IPS, prnBufPtr);
      sampleK = photoBlack720;
      sampleC = photoCyan720;
      sampleM = photoMagenta720;
      sampleY = photoYellow720;
      sampleML = photoMagentaLight720;
      sampleCL = photoCyanLight720;
      break;
     case CLASS_PHOTO_2:
      /* this dot size works just nice for 720 dpi */
      prnBufPtr = SetDotSize(MICRO_DOT_SIZE, prnBufPtr);
      prnBufPtr = SelectPrintingSpeed(SPEED_20IPS, prnBufPtr);
      sampleK = photo2Black720;
      sampleC = photo2Cyan720;
      sampleM = photo2Magenta720;
      sampleY = photo2Yellow720;
      sampleML = photo2MagentaLight720;
      sampleCL = photo2CyanLight720;
      break;
     case CLASS_PHOTO_3:
      prnBufPtr = SetDotSize(VARIABLE_DOT_SIZE_2, prnBufPtr);
      prnBufPtr = SetResolutionRasterImage(720, 720, prnBufPtr);
      sampleK = photo3Black720;
      sampleC = photo3Cyan720;
      sampleM = photo3Magenta720;
      sampleY = photo3Yellow720;
      sampleML = photo3MagentaLight720;
      sampleCL = photo3CyanLight720;
      break;
     case CLASS_800:
      /* it should be SINGLE_DOT_SIZE for 720 dpi, but we use 1440 dot */
      prnBufPtr = SetDotSize(MICRO_DOT_SIZE, prnBufPtr);
      prnBufPtr = SelectPrintingSpeed(SPEED_20IPS, prnBufPtr);
      sampleK = sample720;
      sampleC = cyan720;
      sampleM = magenta720;
      sampleY = yellow720;
      break;
    }
   else /* 360 dpi */
    switch (printerClass)
     {
      case CLASS_PRO:
       prnBufPtr = SetDotSize(STANDARD_DOT_SIZE, prnBufPtr);
       sampleK = samplePro360;
       break;
      case CLASS_PHOTO:
       prnBufPtr = SetDotSize(DOUBLE_DOT_SIZE, prnBufPtr);
       /* no measurments so far; use 720 dpi measurments */
       sampleK = photoBlack720;
       sampleC = photoCyan720;
       sampleM = photoMagenta720;
       sampleY = photoYellow720;
       sampleML = photoMagentaLight720;
       sampleCL = photoCyanLight720;
       break;
      case CLASS_PHOTO_2:
       prnBufPtr = SetDotSize (DOUBLE_DOT_SIZE, prnBufPtr);
       /* no measurments so far; use 720 dpi measurments */
       sampleK = photo2Black720;
       sampleC = photo2Cyan720;
       sampleM = photo2Magenta720;
       sampleY = photo2Yellow720;
       sampleML = photo2MagentaLight720;
       sampleCL = photo2CyanLight720;
       break;
     case CLASS_PHOTO_3:
      prnBufPtr = SetDotSize(STANDARD_DOT_SIZE, prnBufPtr);
      prnBufPtr = SetResolutionRasterImage(360, 360, prnBufPtr);
      sampleK = photo3Black720;
      sampleC = photo3Cyan720;
      sampleM = photo3Magenta720;
      sampleY = photo3Yellow720;
      sampleML = photo3MagentaLight720;
      sampleCL = photo3CyanLight720;
      break;
      case CLASS_800:
       prnBufPtr = SetDotSize(DOUBLE_DOT_SIZE, prnBufPtr); /* we want to change that to smaller */
       sampleK = sample360;
       sampleC = cyan360;
       sampleM = magenta360;
       sampleY = yellow360;
       break;
     }
  if (Epson.modeType==IMG_BW)
   {
    prnBufPtr = SetMicroWeave(False, prnBufPtr);
    prnBufPtr = SelectPrintingMode(MONOCHROME_MODE, prnBufPtr);
   }
  else /* IMG_TRUE_COL_DITH */
   {
#ifdef LINEARIZATION
int i;
unsigned short gamma[256];
#endif
    switch (printerClass)
     {
      case CLASS_PRO:
       dithMK = dithM_C = dithM_M = dithM_Y = linearizeDither(sampleK, 1.0);
       break;
      case CLASS_800:
       dithMK =  linearizeDither(sampleK, 1.0);	
       dithM_C = linearizeDither(sampleC, 0.0);
       dithM_M = linearizeDither(sampleM, 0.0);
       dithM_Y = linearizeDither(sampleY, 0.0);
       break;
      case CLASS_PHOTO:
       dithMK =   linearizeDither(sampleK, 3.0);
       dithM_C =  linearizeDither(sampleC, 2.0);
       dithM_M =  linearizeDither(sampleM, 2.0);
       dithM_Y =  linearizeDither(sampleY, 1.0);
       dithM_CL = linearizeDither(sampleCL, 1.0);
       dithM_ML = linearizeDither(sampleML, 1.0);
       break;
      case CLASS_PHOTO_2:
       dithMK =   linearizeDither(sampleK, 0.0);
       dithM_C =  linearizeDither(sampleC, 0.0);
       dithM_M =  linearizeDither(sampleM, 0.0);
       dithM_Y =  linearizeDither(sampleY, 0.0);
       dithM_CL = linearizeDither(sampleCL, 0.0);
       dithM_ML = linearizeDither(sampleML, 0.0);
       break;
      case CLASS_PHOTO_3:
       dithMK =   linearizeDither(sampleK, 5.0);
       dithM_C =  linearizeDither(sampleC, 3.0);
       dithM_M =  linearizeDither(sampleM, 3.0);
       dithM_Y =  linearizeDither(sampleY, 2.0);
       dithM_CL = linearizeDither(sampleCL, 2.0);
       dithM_ML = linearizeDither(sampleML, 2.0);
       break;
     }

#ifdef LINEARIZATION
    /* equal distribution matrix for measurment of non-gamma corrected ink levels */
    gamma[0] = 0;
    for (i = 1; i < 256; i++)
     gamma[i] = 257;
    dithMK = makeDither(gamma);
    gamma[0] = 0;
    for (i = 1; i < 256; i++)
     gamma[i] = 257;
    dithM_C = makeDither(gamma);
    gamma[0] = 0;
    for (i = 1; i < 256; i++)
     gamma[i] = 257;
    dithM_M = makeDither(gamma);
    gamma[0] = 0;
    for (i = 1; i < 256; i++)
     gamma[i] = 257;
    dithM_Y = makeDither(gamma);
    for (i = 1; i < 256; i++)
     gamma[i] = 257;
    dithM_CL = makeDither(gamma);
    for (i = 1; i < 256; i++)
     gamma[i] = 257;
    dithM_ML = makeDither(gamma);
#endif
    if ((!dithMK) || (!dithM_C) || (!dithM_M) || (!dithM_Y))
     {
      printf("dither matrix generation error\n");
      return (PRT_INTERNAL_ERROR);
     }
    /* count the number of dots for each color level and each dithering matrix */
    setDots(&blackDots[0], dithMK);
    setDots(&cyanDots[0], dithM_C);
    setDots(&magentaDots[0], dithM_M);
    setDots(&yellowDots[0], dithM_Y);
    if (inks6)
     {
      int jj, magLStart, cyanLStart, magSwitch, cyanSwitch;
      float factor1, factor2;

      if ((!dithM_ML) || (!dithM_CL))
       {
	printf("dither matrix generation error\n");
	return (PRT_INTERNAL_ERROR);
       }
      setDots (&cyanLightDots[0], dithM_CL);
      setDots (&magentaLightDots[0], dithM_ML);
      if (Epson.Hdpi == 720)
       { cyanLStart = 128; cyanSwitch = 80; magLStart = 131; magSwitch = 100; }
      else // 360 dpi !! not set yet
       { cyanLStart = 112; cyanSwitch = 80; magLStart = 172; magSwitch = 128; }
      /* use same light/dark switch for cyan and magenta; maybe slightly more visible, but less */
      /* color shift variations, on combined use, so more predictable and better for interpolation */
      if (printerClass == CLASS_PHOTO_2)
       { cyanLStart = 128; cyanSwitch = 80; magLStart = 128; magSwitch = 80; }
      if (printerClass == CLASS_PHOTO_3)
       { cyanLStart = 128; cyanSwitch = 80; magLStart = 128; magSwitch = 80; }


      for (jj = 0; jj < 256; jj++)
	cyan2DarkC[jj] = mag2DarkM[jj] = (unsigned char) jj;
      for (jj = 0; jj < 256; jj++)
	cyan2LightC[jj] = mag2LightM[jj] = 0;
      // "smart" overlapping of light and dark magenta/cyan; power to 0.6 worked perceptually
      for (jj = cyanSwitch+OVERLAP; jj >= 0; jj--)
	cyan2LightC[jj] = (cyanLStart * jj) / cyanSwitch;
      for (jj = cyanSwitch; jj >= 0; jj--)
	cyan2DarkC[jj] = 0;
      for (jj = magSwitch+OVERLAP; jj >= 0; jj--)
	mag2LightM[jj] = (magLStart * jj) / magSwitch;
      for (jj = magSwitch; jj >= 0; jj--)
	mag2DarkM[jj] = 0;
      for (jj = OVERLAP; jj>0; jj--)
       {
	factor1 = (OVERLAP-jj*1.0)/OVERLAP;
	factor2 = pow((jj*1.0)/OVERLAP, 0.6);
        cyan2LightC[jj+cyanSwitch] = (unsigned char) ((int)cyan2LightC[jj+cyanSwitch]*factor1);
        mag2LightM[jj+magSwitch] = (unsigned char) ((int)mag2LightM[jj+magSwitch]*factor1);
        cyan2DarkC[jj+cyanSwitch] = (unsigned char) ((int)cyan2DarkC[jj+cyanSwitch]*factor2);
        mag2DarkM[jj+magSwitch] = (unsigned char) ((int)mag2DarkM[jj+magSwitch]*factor2);
       }
#ifdef LINEARIZATION
      for (i = 0; i<256; i++)
       cyan2LightC[i]=cyan2DarkC[i]=mag2LightM[i]=mag2DarkM[i]=i;
#endif
     }
    prnBufPtr = SelectPrintingMode(COLOR_MODE, prnBufPtr);
    /* photo printer does not want to print without microweave! */
    prnBufPtr = SetMicroWeave((inks6) ? True : complexImage, prnBufPtr);
    Epson.firstLine = True;
    /* organization of rows CMYK[0] CMYK[1] CMYK[2] */
    CMYKBlck = (unsigned char *) calloc (xpixels * Epson.inkNo, sizeof (char));
    /* allocate memory for binary output */
    outBlockSize = ((xpixels + 7) / 8) + 1;
    outBlock = (unsigned char *) malloc (outBlockSize * Epson.inkNo * sizeof (char));
    if ((CMYKBlck == NULL) || (outBlock == NULL))
      return (PRT_NO_MEMORY);
    dithColOffsX[CYANOFFSET] = dithColOffsY[CYANOFFSET] = 0;
    dithColOffsX[MAGENTAOFFSET] = dithColOffsY[MAGENTAOFFSET] = 64;
    dithColOffsX[YELLOWOFFSET] = dithColOffsY[YELLOWOFFSET] = 128;
    dithColOffsX[BLACKOFFSET] = dithColOffsY[BLACKOFFSET] = 192;
    dithColOffsX[CYANLIGHTOFFSET] = dithColOffsY[CYANLIGHTOFFSET] = 32;
    dithColOffsX[MAGENTALIGHTOFFSET] = dithColOffsY[MAGENTALIGHTOFFSET] = 96;
   }
  prnBufPtr = SetUnit(Epson.HdpiMode, prnBufPtr);
  prnBufPtr = SetPageLengthU((Epson.Vdpi==720) ? A2HEIGHT_720 :  A2HEIGHT_360, prnBufPtr);
  prnBufPtr = SetPageFormat((Epson.Vdpi==720) ? A2TOP_MARGIN_720 : A2TOP_MARGIN_360, 
   (Epson.Vdpi==720) ? A2BOTTOM_MARGIN_720 : A2BOTTOM_MARGIN_360, prnBufPtr);
  Epson.topM = xpixels;
  Epson.leftM = (xmargin/8)*8;
  prnBufPtr = SetDirection(1, prnBufPtr); /* unidirectional printing */
  prnBufPtr = SetRelYPos(-(Epson.Vdpi / 25.4), prnBufPtr);
  if (ymargin)
    prnBufPtr = SetRelYPos (ymargin, prnBufPtr);
  fwrite (Epson.buf, 1, prnBufPtr - Epson.buf, Epson.fout);
  return (PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
	int lines, char bandType, unsigned char *cmap)
{
 int result;

 result = PRT_INTERNAL_ERROR;
 if (data == NULL)
  return (PRT_INTERNAL_ERROR);
 if (Epson.modeType < bandType)
  return (PRT_WRONG_IMG);

 if (Epson.currBandType != bandType)
  Epson.firstLine = True;

 switch (Epson.modeType)
  {
   case IMG_BW:	/* just BW image possible, anyway */
    result = processImageBlockBW(data, bpl, xpixels, ypos, lines, Epson.leftM, Epson.fout);
    break;
   case IMG_TRUE_COL_DITH:
    if (bandType==IMG_BW)
     result = processImageBlockBW(data, bpl, xpixels, ypos, lines, Epson.leftM, Epson.fout);
    else
     if (cmap)
      result = processImageBlockCmap(data, bpl, xpixels, ypos, lines, cmap, Epson.fout);
     else
      if (bandType==IMG_BW)
       result = processImageBlockTrueColor(data, bpl, xpixels, ypos, lines, Epson.fout);
      else
       result = PRT_INTERNAL_ERROR;
   break;
  }
 if (result == PRT_OK)
   Epson.vertPos += lines;
 Epson.currBandType = bandType;
 return (result);
}

static int endPrint (int printStatus)
{
 unsigned char *prnBufPtr = Epson.buf;

 if (printStatus == PRT_OK)
  {
   prnBufPtr = FormFeed(prnBufPtr);
   prnBufPtr = ResetPrn(prnBufPtr);
   prnBufPtr = SetMicroWeave(False, prnBufPtr);
   fwrite(Epson.buf, 1, prnBufPtr - Epson.buf, Epson.fout);
   fflush(Epson.fout);
   fclose(Epson.fout);
   printStatus = enqueuePrintFile(Epson.fname);
  }
 else
  fclose(Epson.fout);

 /* just cleanup; called even if printing fails */
 if (Epson.modeType==IMG_BW)
  {
   free((char *) Epson.buf);
   free((char *) RLLBuf);
   Epson.bufSize = 0;
   Epson.width = Epson.height = 0;  
  }
 else /* color modes */
  {
   if (dithMK)
    free(dithMK);
   /* we pretend we have these extra matrix on Pro, so do not free them twice */
   if (printerClass!=CLASS_PRO)
    {
     if (dithM_C)
      free(dithM_C);
     if (dithM_M)
      free(dithM_M);
     if (dithM_Y)
      free(dithM_Y);
    }
   if (inks6)
    {
     if (dithM_CL)
      free(dithM_CL);
     if (dithM_ML)
      free(dithM_ML);   
    }
   dithMK = dithM_C = dithM_CL = dithM_M = dithM_ML = dithM_Y = NULL;
   free((char *) Epson.buf);
   free((char *) RLLBuf);
   free((char *) CMYKBlck);
   free((char *) outBlock);
   Epson.bufSize = 0;
   Epson.width = Epson.height = 0;
  }
 remove(Epson.fname);
 free(Epson.fname);
 return(printStatus);
}

static void fillCommon(printerMode *printerMod, int *numModes, printerDensity *prtDensityX,
 int *numDensitX, printerDensity *prtDensityY, int *numDensitY, int *dpiVariable, char **profileName)
{
 printerMod[0].printType = IMG_TRUE_COL_DITH;
 printerMod[1].printType = IMG_BW;
 *numDensitX = *numDensitY = *numModes = 2;
 *dpiVariable = FALSE;
 prtDensityX[0].dpi = prtDensityY[0].dpi = 720000;
 prtDensityX[1].pixels = prtDensityX[0].pixels/2;
 prtDensityX[1].dpi = prtDensityX[0].dpi/2;
 prtDensityY[1].pixels = prtDensityY[0].pixels/2;
 prtDensityY[1].dpi = prtDensityY[0].dpi/2;
 switch (printerClass)
  {
   case CLASS_PRO: *profileName="StylusPro"; inks6=FALSE; break;
   case CLASS_800: *profileName="Stylus800"; inks6=FALSE;  break;
   case CLASS_PHOTO: *profileName="StylusPhoto"; inks6=TRUE;  break;
   case CLASS_PHOTO_2: *profileName="StylusPhoto2"; inks6=TRUE;  break;
   case CLASS_PHOTO_3: *profileName="StylusPhoto3"; inks6=TRUE;  break;
  }
}

/* the pre-historic "original" stylus which was never really supported */
static void getPrtInfo(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PRO;
 prtDensityX[0].pixels = 2880;
 prtDensityY[0].pixels = 3960;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "Stylus";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPro(printerMode * printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PRO;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "StylusPro";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoProXL(printerMode * printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PRO;
 prtDensityX[0].pixels = ((A3EXTWIDTH - A3EXTLEFT_MARGIN - A3EXTRIGHT_MARGIN)*720)/25.4;
 prtDensityY[0].pixels = ((A3EXTHEIGHT - A3EXTTOP_MARGIN - A3EXTBOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A3EXTTOP_MARGIN;
 *nonprintableBottom = A3EXTBOTTOM_MARGIN;
 *nonprintableLeft = A3EXTLEFT_MARGIN;
 *prtName = "StylusProXL";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfo800(printerMode * printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_800;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "Stylus800";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfo850(printerMode * printerMod, int *numModes,
	printerDensity * prtDensityX, int *numDensitX,
	printerDensity * prtDensityY, int *numDensitY,
	int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
	int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_800;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "Stylus850";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfo1520(printerMode * printerMod, int *numModes,
 printerDensity * prtDensityX, int *numDensitX,
 printerDensity * prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_800;
 prtDensityX[0].pixels = ((A2WIDTH - A2LEFT_MARGIN_1520 - A2RIGHT_MARGIN_1520)*720)/25.4;
 prtDensityY[0].pixels = ((A2HEIGHT - A2TOP_MARGIN - A2BOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A2TOP_MARGIN;
 *nonprintableBottom = A2BOTTOM_MARGIN;
 *nonprintableLeft = A2LEFT_MARGIN_1520;
 *prtName = "Stylus1520";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPhoto(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "StylusPhoto";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPhoto700(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "StylusPhoto700";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPhotoEX(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO;
 prtDensityX[0].pixels = ((A3WIDTH - A3LEFT_MARGIN - A3RIGHT_MARGIN)*720)/25.4;
 prtDensityY[0].pixels = ((A3HEIGHT - A3TOP_MARGIN - A3BOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A3TOP_MARGIN;
 *nonprintableBottom = A3BOTTOM_MARGIN;
 *nonprintableLeft = A3LEFT_MARGIN;
 *prtName = "StylusPhotoEX";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfo5000(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO;
 prtDensityX[0].pixels = ((A2WIDTH - A2LEFT_MARGIN_3000 - A2RIGHT_MARGIN_3000)*720)/25.4;
 prtDensityY[0].pixels = ((A2HEIGHT - A2TOP_MARGIN - A2BOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A2TOP_MARGIN;
 *nonprintableBottom = A2BOTTOM_MARGIN;
 *nonprintableLeft = A2LEFT_MARGIN_3000;
 *prtName = "StylusPhoto5000";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfo3000(printerMode * printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_800;
 prtDensityX[0].pixels = ((A2WIDTH - A2LEFT_MARGIN_3000 - A2RIGHT_MARGIN_3000)*720)/25.4;
 prtDensityY[0].pixels = ((A2HEIGHT - A2TOP_MARGIN - A2BOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A2TOP_MARGIN;
 *nonprintableBottom = A2BOTTOM_MARGIN;
 *nonprintableLeft = A2LEFT_MARGIN_3000;
 *prtName = "Stylus3000";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPhoto750(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO_2;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "StylusPhoto750";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPhoto1200(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO_2;
 prtDensityX[0].pixels = ((A3EXTWIDTH - A3EXTLEFT_MARGIN - A3EXTRIGHT_MARGIN)*720)/25.4;
 prtDensityY[0].pixels = ((A3EXTHEIGHT - A3EXTTOP_MARGIN - A3EXTBOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A3EXTTOP_MARGIN;
 *nonprintableBottom = A3EXTBOTTOM_MARGIN;
 *nonprintableLeft = A3EXTLEFT_MARGIN;
 *prtName = "StylusPhoto1200";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

static void getPrtInfoPhoto870(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO_3;
 prtDensityX[0].pixels = 2880*2;
 prtDensityY[0].pixels = 3960*2;
 *nonprintableTop = A4TOP_MARGIN;
 *nonprintableBottom = A4BOTTOM_MARGIN;
 *nonprintableLeft = A4LEFT_MARGIN;
 *prtName = "StylusPhoto870";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}


static void getPrtInfoPhoto1270(printerMode *printerMod, int *numModes,
 printerDensity *prtDensityX, int *numDensitX,
 printerDensity *prtDensityY, int *numDensitY,
 int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
 int *dpiVariable, char **prtName, char **profileName)
{
 printerClass = CLASS_PHOTO_3;
 prtDensityX[0].pixels = ((A3EXTWIDTH - A3EXTLEFT_MARGIN - A3EXTRIGHT_MARGIN)*720)/25.4;
 prtDensityY[0].pixels = ((A3EXTHEIGHT - A3EXTTOP_MARGIN - A3EXTBOTTOM_MARGIN)*720)/25.4;
 *nonprintableTop = A3EXTTOP_MARGIN;
 *nonprintableBottom = A3EXTBOTTOM_MARGIN;
 *nonprintableLeft = A3EXTLEFT_MARGIN;
 *prtName = "StylusPhoto1270";
 fillCommon(printerMod, numModes, prtDensityX, numDensitX,
  prtDensityY, numDensitY, dpiVariable, profileName);
}

void registerStylus()
{ registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint); }

void registerStylusPro()
{ registerPrinter(*getPrtInfoPro, *initPrint, *printBand, *endPrint); }

void registerStylusProXL()
{ registerPrinter(*getPrtInfoProXL, *initPrint, *printBand, *endPrint); }

void registerStylus800()
{ registerPrinter(*getPrtInfo800, *initPrint, *printBand, *endPrint); }

void registerStylus1520()
{ registerPrinter(*getPrtInfo1520, *initPrint, *printBand, *endPrint); }

void registerStylusPhoto()
{ registerPrinter(*getPrtInfoPhoto, *initPrint, *printBand, *endPrint); }

void registerStylusPhoto700()
{ registerPrinter(*getPrtInfoPhoto700, *initPrint, *printBand, *endPrint); }

void registerStylusPhotoEX()
{ registerPrinter(*getPrtInfoPhotoEX, *initPrint, *printBand, *endPrint); }

void registerStylus5000()
{ registerPrinter(*getPrtInfo5000, *initPrint, *printBand, *endPrint); }

void registerStylus850()
{ registerPrinter(*getPrtInfo850, *initPrint, *printBand, *endPrint); }

void registerStylus3000()
{ registerPrinter(*getPrtInfo3000, *initPrint, *printBand, *endPrint); }

void registerStylusPhoto750()
{ registerPrinter(*getPrtInfoPhoto750, *initPrint, *printBand, *endPrint); }

void registerStylusPhoto1200()
{ registerPrinter(*getPrtInfoPhoto1200, *initPrint, *printBand, *endPrint); }

void registerStylusPhoto870()
{ registerPrinter(*getPrtInfoPhoto870, *initPrint, *printBand, *endPrint); }

void registerStylusPhoto1270()
{ registerPrinter(*getPrtInfoPhoto1270, *initPrint, *printBand, *endPrint); }

